/////////////////////////////////////////////////////////////////////////////////////
// This file is part of the PngOptimizer application
// Copyright (C) 2002/2011 Hadrien Nilsson - psydk.org
// For conditions of distribution and use, see copyright notice in PngOptimizer.h
/////////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "PngOptimizer.h"
#include "resource.h"

#include "POTraceCtl.h"
#include "DlgScreenshotsOptions.h"
#include "BmpClipboardDumper.h"
#include "MainWnd.h"
#include "DlgAskForFileName.h"

///////////////////////////////////////////////////////////////////////////////
// Main window title
static const char k_szTitle[] = "PngOptimizer";

// Welcome message
static const char k_szWelcomeMessage[] = "Drop PNG, GIF, BMP or TGA files here";

// Configuration strings
static const char k_szConfigSubDir[]   = "PngOptimizer";
static const char k_szConfigFileName[] = "PngOptimizer.ini";

static const char k_szConfigComment1[] = "PngOptimizer configuration file";
static const char k_szConfigComment2[] = "This file is encoded in UTF-8";

static const char k_szSectionEngine[]         = "Engine";

static const char k_szSectionScreenshots[]      = "Screenshots";
static const char k_szShotUseDefaultDir[]       = "UseDefaultDir";
static const char k_szShotCustomDir[]           = "CustomDir";
static const char k_szShotMaximizeCompression[] = "MaximizeCompression";
static const char k_szShotAskForFileName[]      = "AskForFileName";

static const char k_szSectionWindow[]  = "Window";
static const char k_szWndX[]           = "X";
static const char k_szWndY[]           = "Y";
static const char k_szWndWidth[]       = "Width";
static const char k_szWndHeight[]      = "Height";
static const char k_szWndAlwaysOnTop[] = "AlwaysOnTop";

// Screenshot comments
static const char k_szCreating[] = "Creating ";
static const char k_szCreatingAndOptimizing[] = "Creating & Optimizing ";

///////////////////////////////////////////////////////////////////////////////
POApplication g_app;

POApplication& POApplication::GetInstance()
{
	return g_app;
}

bool POApplication::IsBmpAvailable() const
{
	return m_bmpcd.IsBmpAvailable();
}

POApplication::POApplication()
{
	m_hInstance = NULL;
	m_hWorkingThread = NULL;
	m_hMutexPngOptimizer = NULL;
	m_hAccel = NULL;
}

//////////////////////////////////////////////////////////////
DWORD WINAPI POApplication::ThreadProcStatic(LPVOID pParameter)
{
	POApplication* pApp = (POApplication*) pParameter;
	pApp->ThreadProc();
	return 0;
}

void POApplication::ThreadProc()
{
	m_engine.OptimizeFiles(m_astrFilePaths);

	// Notify the main window that the job is finished

	// A better theorical design would be to send a message to the main thread, independently of the main window.
	// But we cannot use PostThreadMessage because the main thread may not receive the message.
	// This can occur if the main thread is stuck in another message loop than the one we wrote here.
	// For example, when the user is dragging the scrollbar.
	BOOL bPostOk = ::PostMessage(POApplication::GetInstance().m_mainwnd.m_hWnd, WM_APP_THREADJOBDONE, 0, 0);
	bPostOk;
}

bool POApplication::IsJobRunning()
{
	return m_hWorkingThread != 0;
}

void POApplication::StartJob()
{
	if( m_astrFilePaths.GetSize() == 0 )
		return;

	// Job already running ? Should never happen
	ASSERT(m_hWorkingThread == NULL); // So we assert here
	if( m_hWorkingThread != NULL )
	{
		// And we return here for release mode
		return;
	}

	// Bug me not with other file names while working
	m_mainwnd.AllowFileDropping(false);

	DWORD nThreadId = 0;
	m_hWorkingThread = CreateThread(NULL, 0, ThreadProcStatic, this, 0, &nThreadId);
	if( !m_hWorkingThread )
	{
		m_mainwnd.WriteLine(L"Cannot create working thread", POEngine::tcErrorMsg);
	}
}

void POApplication::OnJobDone()
{
	m_hWorkingThread = NULL;
	m_mainwnd.AllowFileDropping(true);
}

//////////////////////////////////////////////////////////////
COLORREF POApplication::CrFromTc(POEngine::TextColor eColor)
{
	COLORREF cr = RGB(0, 0, 0);
	if( eColor == POEngine::tcFilePath )
	{
		// Color for file paths
		cr = RGB(0, 0, 0);
	}
	else if( eColor == POEngine::tcRegularInfo )
	{
		cr = RGB(100, 100, 100);
	}
	else if( eColor == POEngine::tcActionVerb )
	{
		// Creating, Converting, Optimizing...
		cr = RGB(100, 100, 100);
	}
	else if( eColor == POEngine::tcSizeInfo )
	{
		// xxx bytes -> yyy bytes
		cr = RGB(100, 100, 100);
	}
	else if( eColor == POEngine::tcFileEnlarged )
	{
		// (103% of the original size)
		cr = RGB(200, 100, 0);
	}
	else if( eColor == POEngine::tcFileSizeSame )
	{
		// (100% of the original size)
		cr = RGB(100, 100, 100);
	}
	else if( eColor == POEngine::tcFileReduced )
	{
		// (80% of the original size)
		cr = RGB(0, 120, 0);
	}
	else if( eColor == POEngine::tcActionOk )
	{
		// (OK) :-)
		cr = RGB(0, 100, 0);
	}
	else if( eColor == POEngine::tcActionFail )
	{
		// (KO) :-(
		cr = RGB(200, 0, 0);
	}
	else if( eColor == POEngine::tcErrorMsg )
	{
		// could not load bla
		cr = RGB(255, 0, 0);
	}
	else if( eColor == POEngine::tcAnimated )
	{
		// [Animated GIF : converting to APNG]
		// [Animated GIF : ignoring]
		// [APNG]
		cr = RGB(0, 50, 200);
	}
	else if( eColor == POEngine::tcWelcomeMsg )
	{
		// Drop PNG, GIF, BMP or TGA files here
		cr = RGB(0, 0, 200);
	}

	return cr;
}
//////////////////////////////////////////////////////////////

// Handle files given from the command line
void POApplication::ProcessCmdLineArgs()
{
	StringArray astrArgs = Process::CommandLineToArgv( Process::GetCommandLine());
	const int32 nArgCount = astrArgs.GetSize();
	if( nArgCount <= 1 )
		return;

	m_astrFilePaths.SetSize(0);
	m_astrFilePaths.EnsureCapacity(nArgCount - 1);
	for(int32 i = 1; i < nArgCount; ++i)
	{
		m_astrFilePaths.Add(astrArgs[i]);
	}

	StartJob();
}

String POApplication::GetConfigFilePath() const
{
	String strConfigFileName = k_szConfigFileName;

	// A config file in the executable directory ?
	String strExeDir = Process::GetExecutableDirectory();
	String strConfigFilePath = FilePath::Combine(strExeDir, strConfigFileName);
	if( !File::Exists(strConfigFilePath) )
	{
		// Nop, use a user profile location
		String strAppDataDir = System::GetApplicationDataDirectory();
		String strConfigDir = FilePath::Combine(strAppDataDir, k_szConfigSubDir);

		if( !Directory::Exists(strConfigDir) )
		{
			// First time PngOptimizer is run, create profile directory
			Directory::Create(strConfigDir);
		}
		strConfigFilePath = FilePath::Combine(strConfigDir, strConfigFileName);
	}

	return strConfigFilePath;
}

/////////////////////////////////////////////////////////////
bool POApplication::ReadSettings(RECT& rectWnd, bool& bCenterWindow, bool& alwaysOnTop)
{
	const int32 nDefaultWidth = 482;
	const int32 nDefaultHeight = 300;

	RECT rc;
	rc.left = 0;
	rc.top = 0;
	rc.right = rc.left + nDefaultWidth;
	rc.bottom = rc.top + nDefaultHeight;

	rectWnd = rc;
	bCenterWindow = true;
	alwaysOnTop = true; // Default

	String strConfigFilePath = GetConfigFilePath();
	if( !File::Exists(strConfigFilePath) )
	{
		// First run
		return true;
	}

	MemIniFile ini;
	if( !ini.Load(strConfigFilePath) )
	{
		// Cannot read config file :(
		return false;
	}

	ini.SetSection(k_szSectionEngine);
	m_engine.m_settings.LoadFromIni(ini);

	////////////////////////////////////////////////
	ini.SetSection(k_szSectionScreenshots);
	ini.GetBool(k_szShotUseDefaultDir, m_bmpcd.m_bUseDefaultDir);
	ini.GetString(k_szShotCustomDir, m_bmpcd.m_strCustomDir);
	ini.GetBool(k_szShotMaximizeCompression, m_bmpcd.m_bMaximizeCompression);
	ini.GetBool(k_szShotAskForFileName, m_bmpcd.m_bAskForFileName);

	////////////////////////////////////////////////
	ini.SetSection(k_szSectionWindow);

	int nX = 0;
	int nY = 0;
	int nWidth = 0;
	int nHeight = 0;

	if( ini.GetInt(k_szWndX, nX) && ini.GetInt(k_szWndY, nY) )
	{
		// No need to center, the values are ok
		bCenterWindow = false;
	}
	
	ini.GetInt(k_szWndWidth, nWidth);
	ini.GetInt(k_szWndHeight, nHeight);

	// Reajust rect if needed
	if( nX < 0 )
	{
		nX = 0;
	}
	
	if( nY < 0 )
	{
		nY = 0;
	}

	if( nWidth <= 50 )
	{
		nWidth = nDefaultWidth;
	}

	if( nHeight <= 50 )
	{
		nHeight = nDefaultHeight;
	}
	////////////////////////////////////////////////

	rc.left = nX;
	rc.top = nY;
	rc.right = rc.left + nWidth;
	rc.bottom = rc.top + nHeight;
	rectWnd = rc;

	ini.GetBool(k_szWndAlwaysOnTop, alwaysOnTop);

	return true;
}

bool POApplication::WriteSettings(const RECT& rcWnd, bool alwaysOnTop)
{
	int nX = rcWnd.left;
	int nY = rcWnd.top;
	int nWidth = rcWnd.right - rcWnd.left;
	int nHeight = rcWnd.bottom - rcWnd.top;

	MemIniFile ini;

	////////////////////////////////////////////////
	ini.SetSection(k_szSectionEngine);
	m_engine.m_settings.SaveToIni(ini);

	////////////////////////////////////////////////
	ini.SetSection(k_szSectionScreenshots);
	ini.SetBool(k_szShotUseDefaultDir, m_bmpcd.m_bUseDefaultDir);
	ini.SetString(k_szShotCustomDir, m_bmpcd.m_strCustomDir);
	ini.SetBool(k_szShotMaximizeCompression, m_bmpcd.m_bMaximizeCompression);
	ini.SetBool(k_szShotAskForFileName, m_bmpcd.m_bAskForFileName);

	////////////////////////////////////////////////
	ini.SetSection(k_szSectionWindow);
	ini.SetInt(k_szWndX, nX);
	ini.SetInt(k_szWndY, nY);
	ini.SetInt(k_szWndWidth, nWidth);
	ini.SetInt(k_szWndHeight, nHeight);
	ini.SetBool(k_szWndAlwaysOnTop, alwaysOnTop);

	ini.SetCommentLine1(k_szConfigComment1);
	ini.SetCommentLine2(k_szConfigComment2);

	String strConfigFilePath = GetConfigFilePath();
	return ini.Dump(strConfigFilePath, true); // true is for CR-LF, we are kind for notepad users
}

/////////////////////////////////////////////////////////////
// Change a rectangular zone to avoid full overlap
void POApplication::ChangeRectBecauseOfOverlap(RECT& rcWnd)
{
	// We divide the desktop screen into 4 zones, in order to decide
	// the direction of the offset to apply to the new rectangle
	
	int nScreenWidth = GetDeviceCaps(NULL, HORZRES);
	int nScreenHeight = GetDeviceCaps(NULL, VERTRES);
	
	const int nOffset = 16;
	if( rcWnd.left < nScreenWidth / 2 )
	{
		// On the left
		rcWnd.left += nOffset;
		rcWnd.right += nOffset;
	}
	else
	{
		// On the right
		if( (rcWnd.left - nOffset) > 0 )
		{
			rcWnd.left -= nOffset;
			rcWnd.right -= nOffset;
		}
	}

	if( rcWnd.top < nScreenHeight / 2 )
	{
		// Bottom
		rcWnd.top += nOffset;
		rcWnd.bottom += nOffset;
	}
	else
	{
		// Top
		if( (rcWnd.top - nOffset) > 0 )
		{
			rcWnd.top -= nOffset;
			rcWnd.bottom -= nOffset;
		}
	}
}

void POApplication::OnMainWndFilesDropped(const chustd::StringArray& astrFilePaths)
{
	m_astrFilePaths = astrFilePaths;
	StartJob();	
}

void POApplication::OnMainWndDestroying()
{
	HWND hWnd = m_mainwnd.GetHandle();

	WINDOWPLACEMENT wp;
	if( GetWindowPlacement(hWnd, &wp) )
	{
		WriteSettings(wp.rcNormalPosition, m_mainwnd.GetAlwaysOnTop()); // Save settings in the registry
	}
	
	// Suspend the thread so the application won't crash if resources allocated by the main
	// thread are deallocated while still in use by the working thread
	// For allocations made by the working thread, we let Windows do the job
	SuspendThread(m_hWorkingThread);

	// Quit the main thread message loop
	PostQuitMessage(0);
}

void POApplication::DumpScreenshot()
{
	if( !IsBmpAvailable() )
	{
		return;
	}

	if( !m_bmpcd.Dump(&m_engine) )
	{
		String strLastError = m_bmpcd.GetLastError();
		if( !strLastError.IsEmpty() )
		{
			m_mainwnd.WriteLine(L"Screenshot creation failure: " + strLastError, CrFromTc(POEngine::tcErrorMsg));
		}
		return;
	}

	String strCreating = m_bmpcd.m_bMaximizeCompression ? k_szCreatingAndOptimizing : k_szCreating;

	String strScreenshotPath = m_bmpcd.GetFilePath();
	
	// Replace slashs with backslashs because some applications are lost when using something else than backslashs
	strScreenshotPath = strScreenshotPath.ReplaceAll(L"/", L"\\");

	// Stop redrawing to avoid ugly flashs
	m_mainwnd.SetTraceCtlRedraw(false);
	m_mainwnd.ClearLine();
	m_mainwnd.Write(strCreating, CrFromTc(POEngine::tcRegularInfo));
	m_mainwnd.AddLink(m_bmpcd.GetFileName(), strScreenshotPath);
	m_mainwnd.Write(L"... ", CrFromTc(POEngine::tcRegularInfo));
	m_mainwnd.Write(L" (OK)", CrFromTc(POEngine::tcActionOk));
	m_mainwnd.SetTraceCtlRedraw(true);

	// Force redraw here (so outside the SetRedraw(true/false) body)
	m_mainwnd.WriteLine(L"", CrFromTc(POEngine::tcRegularInfo));
}

void POApplication::OnBmpcdStateChanged(BmpClipboardDumper::DumpState eDumpState)
{
	String strCreating = m_bmpcd.m_bMaximizeCompression ? k_szCreatingAndOptimizing : k_szCreating;

	switch(eDumpState)
	{
	case BmpClipboardDumper::dsAskForFileName:
		{
			DlgAskForFileName dlg;
			dlg.m_strFileName = m_bmpcd.GetPreviousFileName();
			if( dlg.m_strFileName.IsEmpty() )
			{
				// Use the computed file name
				dlg.m_strFileName = m_bmpcd.GetFileName();
				dlg.m_bIncrementFileName = false;
			}
			else
			{
				dlg.m_bIncrementFileName = true;
			}

			dlg.m_strScreenshotDir = m_bmpcd.GetDir();

			if( dlg.DoModal( m_mainwnd.GetHandle()) == IDOK )
			{
				m_bmpcd.SetFileName(dlg.m_strFileName);
			}
			else
			{
				m_bmpcd.Abort();
			}
		}
		break;
	
	case BmpClipboardDumper::dsCreating:
		m_mainwnd.Write(strCreating, CrFromTc(POEngine::tcActionVerb));
		m_mainwnd.Write(m_bmpcd.GetFileName(), CrFromTc(POEngine::tcFilePath));
		m_mainwnd.Write(L"... ", CrFromTc(POEngine::tcRegularInfo));
		m_mainwnd.Update();
		break;

	default:
		break;
	}
}

void POApplication::OnEngineProgressing(const String& str, enum POEngine::TextColor eTextColor)
{
	// This is call from the working thread
	m_mainwnd.Write(str, CrFromTc(eTextColor));
	m_mainwnd.Update();
}

void POApplication::OnMainWndListCleared()
{
	m_mainwnd.WriteLine(k_szWelcomeMessage, POApplication::CrFromTc(POEngine::tcWelcomeMsg));
}

void POApplication::DoConnections()
{
	m_mainwnd.m_eventFilesDropped.Handle(this, &POApplication::OnMainWndFilesDropped);
	m_mainwnd.m_eventListCleared.Handle(this, &POApplication::OnMainWndListCleared);
	m_mainwnd.m_eventDestroying.Handle(this, &POApplication::OnMainWndDestroying);

	m_engine.m_eventProgressing.Handle(this, &POApplication::OnEngineProgressing);

	m_bmpcd.m_eventDumpStateChanged.Handle(this, &POApplication::OnBmpcdStateChanged);
}

bool POApplication::Initialize()
{
	// Necessary to enable PngOptimizer as a drag-and-drop source :-p
	OleInitialize(NULL);

	InitCommonControls();

	m_hInstance = ::GetModuleHandle(NULL);

	///////////////////////////////////////////////////////////////////////
	// Get previous instance window rect, in order to avoid a total overlap with the new instance window
	static const wchar k_szMutexName[] = L"PngOptimizer mutex";

	m_hMutexPngOptimizer = CreateMutex(NULL, TRUE, k_szMutexName);
	int nLastError = ::GetLastError();

	bool bOneInstanceAlreadyRunning = false;
	if( m_hMutexPngOptimizer != 0 && nLastError == ERROR_ALREADY_EXISTS )
	{
		bOneInstanceAlreadyRunning = true;
	}
	///////////////////////////////////////////////////////////////////////

	///////////////////////////////////////////////////////////////////////
	// Read application settings from the registry
	RECT rcWnd;
	bool bCenterWindow;
	bool alwaysOnTop;
	ReadSettings(rcWnd, bCenterWindow, alwaysOnTop);

	if( bOneInstanceAlreadyRunning )
	{
		ChangeRectBecauseOfOverlap(rcWnd);
		
		// Write back the settings so the new window rectangle 
		// for the N+1-th instance will be updated
		WriteSettings(rcWnd, alwaysOnTop);
	}
	
	if( !m_mainwnd.Create(String(k_szTitle).GetBuffer(), rcWnd, alwaysOnTop) )
	{
		static const char szPngOptimizerError[] = "PngOptimizer error";
		String strErr = "Cannot create main gui window : " + m_mainwnd.GetLastError();
		MessageBoxW(NULL, strErr.GetBuffer(), String(szPngOptimizerError).GetBuffer(), MB_OK);
		return false;
	}

	if( bCenterWindow )
	{
		m_mainwnd.CenterWindow();
	}

	OnMainWndListCleared();
	m_mainwnd.Show(SW_SHOW);
	m_mainwnd.Update();

	DoConnections();
	
	/////////////////////////////////////////////////////////////
	// Create application shortcuts
	//  - Ctrl+V or Shift+Inser to paste screenshot
	//  - Ctrl+Shift+Delete to clear the list

	const int nAccelCount = 3;
	static ACCEL aAccels[nAccelCount] =
	{
		{ FCONTROL | FVIRTKEY,          'V',       IDM_PASTEFROMCLIPBOARD },
		{ FSHIFT   | FVIRTKEY,          VK_INSERT, IDM_PASTEFROMCLIPBOARD },
		{ FCONTROL | FVIRTKEY,          VK_DELETE, IDM_CLEARLIST }
	};

	m_hAccel = CreateAcceleratorTable(aAccels, nAccelCount);

	/////////////////////////////////////////////////////////////
	// Handle files given through the command line
	ProcessCmdLineArgs();

	return true;
}

// Perform cleanup
POApplication::~POApplication()
{
	if( m_hAccel )
	{
		DestroyAcceleratorTable(m_hAccel);
	}

	if( m_hMutexPngOptimizer )
	{
		CloseHandle(m_hMutexPngOptimizer);
	}

	OleUninitialize();
}

/////////////////////////////////////////////////////////////
// Main application loop

int32 POApplication::Run()
{	
	HWND hMainWnd = m_mainwnd.GetHandle();
	
	MSG msg;
	while( GetMessage(&msg, NULL, 0, 0)) 
	{
		// Keyboard shortcuts are targeted to the main window, so the main window handle is given
		// to TranslateAccelerator 
		if( !TranslateAccelerator(hMainWnd, m_hAccel, &msg) )
		{
			// Ok, no keyboard shortcut, process the message
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}
	return 0;
}

/////////////////////////////////////////////////////////////
int main()
{
	if( !g_app.Initialize() )
	{
		return 1;
	}
	return g_app.Run();
}

