/////////////////////////////////////////////////////////////////////////////////////
// This file is part of the POEngine library, part of the PngOptimizer application
// Copyright (C) 2002/2011 Hadrien Nilsson - psydk.org
// For conditions of distribution and use, see copyright notice in PngOptimizer.h
/////////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "POEngine.h"
#include "PaletteTranslator.h"
#include <crtdbg.h>

using namespace chustd;

///////////////////////////////////////////////////////////////////////////////
const wchar k_szCannotDumpTry[]          = L"Cannot dump internal try, check available memory";
const wchar k_szCannotWriteOnDevice[]    = L"Cannot write on device";
const wchar k_szCannotWriteUncomplete[]  = L"Write uncomplete, device may be full";
const wchar k_szNotEnoughMemoryForInternalConversion[] = L"Not enough memory for internal conversion";
const wchar k_szUnsupportedPixelFormat[] = L"Unsupported pixel format";

const wchar k_szCannotLoadFile[]             = L"Cannot load file";
const wchar k_szFileTooLarge[]               = L"File is too large";
const wchar k_szUnsupportedFileFormat[]      = L"Unsupported file format";
const wchar k_szCannotLoadRawSourceContent[] = L"Cannot load raw source content";
const wchar k_szNotEnoughMemoryToFlattenPaletteImage[] = L"Not enough memory to flatten palette image";
const wchar k_szNotEnoughMemoryToDeflatePaletteImage[] = L"Not enough memory to deflate palette image";
const wchar k_szNotEnoughMemoryToConvertTo24Bits[] = L"Not enough memory to convert to 24 bits";

const wchar k_szPathDoesNotExist[]        = L"Path does not exist";
const wchar k_szCannotGetFileAttributes[] = L"Cannot get file attributes";
const wchar k_szFileIsReadOnly[]          = L"File is read-only";
const wchar k_szCannotPerformBackup[]     = L"Cannot perform backup, previous backup deletion failed";
const wchar k_szCannotPerformBackupRenameFailed[] = L"Cannot perform backup, rename failed";
const wchar k_szNotEnoughMemoryToKeepOriginalFile[] = L"Not enough memory to keep original file";
const wchar k_szCorruptedChunkStructure[] = L"Corrupted chunk structure";
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
DynamicMemoryFile& POEngine::ResultManager::GetCandidate()
{
	// Get the largest, except if 0
	int64 nSize0 = m_dmf0.GetPosition();
	int64 nSize1 = m_dmf1.GetPosition();

	if( nSize0 == 0 )
	{
		return m_dmf0;
	}

	if( nSize1 == 0 )
	{
		return m_dmf1;
	}

	DynamicMemoryFile& dmf = ( nSize1 > nSize0 ) ? m_dmf1 : m_dmf0;

	// Open the dynamic memory file with allocation performed
	
	// The -64 is to be gentle with the memory allocator that needs room for its headers
	const int32 nFirstAlloc = 512 * 1024 - 64;
	dmf.Open(nFirstAlloc);
	dmf.SetPosition(0);

	return dmf;
}

DynamicMemoryFile& POEngine::ResultManager::GetSmallest()
{
	// Get the smallest, except if 0
	int64 nSize0 = m_dmf0.GetPosition();
	int64 nSize1 = m_dmf1.GetPosition();

	if( nSize0 == 0 )
	{
		return m_dmf1;
	}

	if( nSize1 == 0 )
	{
		return m_dmf0;
	}
	DynamicMemoryFile& dmf = ( nSize0 < nSize1 ) ? m_dmf0 : m_dmf1;
	return dmf;
}

void POEngine::ResultManager::Reset()
{
	m_dmf0.SetPosition(0);
	m_dmf1.SetPosition(0);
}

///////////////////////////////////////////////////////////////////////////////////////////////////

POEngine::POEngine()
{
	m_originalFileWriteTime = 0;
}

POEngine::~POEngine()
{
}

// Returns true if a palette is black&white
// pal [in]           Palette to check
// shouldSwap [out]   true if the palette colors 0 and 1 should be swapped
bool POEngine::IsBlackAndWhite(const Palette& pal, bool& shouldSwap)
{
	if( pal.m_count != 2 )
	{
		return false;
	}
	Color32 col0 = pal.m_colors[0];
	Color32 col1 = pal.m_colors[1];
	
	shouldSwap = false;

	if( col0.IsEqualRgb(Color32::black) && col1.IsEqualRgb(Color32::white) )
	{
		return true;
	}
	if( col0.IsEqualRgb(Color32::white) && col1.IsEqualRgb(Color32::black) )
	{
		shouldSwap = true;
		return true;
	}

	if( col0.GetAlpha() == 0 && col1.IsEqualRgb(Color32::white) )
	{
		return true;
	}
	if( col0.GetAlpha() == 0 && col1.IsEqualRgb(Color32::black) )
	{
		shouldSwap = true;
		return true;
	}
	return false;
}
/////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////
bool POEngine::IsGreyPalette(const Palette& pal)
{
	for(int32 i = 0; i < pal.m_count; ++i)
	{
		uint8 r, g, b;
		pal.m_colors[i].ToRgb(r, g, b);

		if( r != g || r != b || g != b )
		{
			return false;
		}
	}
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//! Performs a PNG dump, forcing a background color if necessary.
//!
//! file   Output file to be used for the dump
//! ds     Dump settings. May be changed to honor background option
//!
//! Returns true upon success
///////////////////////////////////////////////////////////////////////////////////////////////////
bool POEngine::CustomPngDump(IFile& file, PngDumpSettings& ds)
{
	ds.useBackgroundColor = false;
	if( ds.pixelFormat == pf32bppRgba )
	{
		// Force or keep background if requested
		if( m_settings.bkgdOption == POChunk_Keep || m_settings.bkgdOption == POChunk_Force )
		{
			ds.useBackgroundColor = true;
		}
	}
	return Png::Dump(file, ds);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////
//! Performs several dump of buffer in memory.
//////////////////////////////////////////////////////////////////////////////////////////////////////
bool POEngine::PerformDumpTries(PngDumpSettings& ds)
{
	// Test 1
	// - No filtering
	// - ZLib default strategy
	// - ZLib normal memory load
	ds.filtering = 0;
	ds.zlibStrategy = PngDumpSettings::zlibStrategyDefault;
	ds.zlibWindowBitsAndMem = 0;
	if( !CustomPngDump(m_resultmgr.GetCandidate(), ds) )
	{
		AddError(k_szCannotDumpTry);
		return false;
	}

	// This can work for low bits per pixel
	int bitsPerPix = ImageFormat::SizeofPixelInBits(ds.pixelFormat);
	if( bitsPerPix <= 8 )
	{
		// Test 2
		// - No filtering
		// - ZLib default strategy
		// - ZLib low memory load (yes, sometimes it can increase compression ! I don't know why... ZLib master, anyone ?
		ds.filtering = 0;
		ds.zlibStrategy = PngDumpSettings::zlibStrategyDefault;
		ds.zlibWindowBitsAndMem = PngDumpSettings::zlibWindowBitsAndMemLow;
		if( !CustomPngDump(m_resultmgr.GetCandidate(), ds) )
		{
			AddError(k_szCannotDumpTry);
			return false;
		}
	}

	// Test 3
	// - Filtering
	// - ZLib filter strategy
	// - ZLib normal memory load
	ds.filtering = 1;
	ds.zlibStrategy = PngDumpSettings::zlibStrategyFilter;
	ds.zlibWindowBitsAndMem = 0;
	if( !CustomPngDump(m_resultmgr.GetCandidate(), ds) )
	{
		AddError(k_szCannotDumpTry);
		return false;
	}

	// Test 4
	// - Filtering
	// - ZLib default strategy
	// - ZLib normal memory load
	ds.filtering = 1;
	ds.zlibStrategy = PngDumpSettings::zlibStrategyDefault;
	ds.zlibWindowBitsAndMem = 0;
	if( !CustomPngDump(m_resultmgr.GetCandidate(), ds) )
	{
		AddError(k_szCannotDumpTry);
		return false;
	}
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Changes and pack pixels if the original 8bpp pixel buffer can be converted to 1bpp black and white
bool POEngine::TryToConvertIndexedToBlackAndWhite(PngDumpSettings& ds)
{
	if( ds.pixelFormat != pf8bppIndexed )
	{
		// Case not managed
		return false;
	}

	// Maybe it's a black&white picture, in that case there is no need for a palette
	bool shouldSwap;
	if( !IsBlackAndWhite(ds.palette, shouldSwap) )
	{
		return false;
	}

	uint8 nAlpha0 = ds.palette.m_colors[0].GetAlpha();
	uint8 nAlpha1 = ds.palette.m_colors[1].GetAlpha(); 
			
	if( (nAlpha0 != 255 || nAlpha1 != 255) && m_settings.avoidGreyWithSimpleTransparency )
	{
		// A black&white picture with transparency, we stay in palette mode
		return false;
	}

	if( shouldSwap )
	{
		Color32 colTmp = ds.palette.m_colors[0];
		ds.palette.m_colors[0] = ds.palette.m_colors[1];
		ds.palette.m_colors[1] = colTmp;

		// Inversion : 0 => 1 & 1 => 0
		PaletteTranslator inverterTranslator;
		inverterTranslator.conv[0] = 1;
		inverterTranslator.conv[1] = 0;
		inverterTranslator.TranslateAll(ds);
	}

	// Check if there was transparency in this paletted image, and if it is
	// compatible with a pure black&white image with no palette

	// In grayscale, only one index can define an alpha of 0
	uint8 a0 = ds.palette.m_colors[0].GetAlpha();
	uint8 a1 = ds.palette.m_colors[1].GetAlpha();
	if( a0 == 255 && a1 == 255 )
	{
		// No transparency at all
		ds.useTransparentColor = false;
		ds.pixelFormat = pf1bppGrayScale;
	}
	else if( a0 == 0 && a1 == 255 )
	{
		// Transparency for index 0
		ds.useTransparentColor = true;
		ds.tRNS.grey = 0;
		ds.pixelFormat = pf1bppGrayScale;
	}
	else if( a0 == 255 && a1 == 0 )
	{
		// Transparency for index 1
		ds.useTransparentColor = true;
		ds.tRNS.grey = 1;
		ds.pixelFormat = pf1bppGrayScale;
	}
	else
	{
		// Another configuration of transparency, we cannot keep the image in black&white
		// We stay in palette mode...
		return false;
	}

	// Pack the pixels according to the new pixel format (1bpp grey)
	const int frameCount = ds.frames.GetSize();

	// Pack the default image if any
	if( ds.hasDefaultImage || frameCount == 0 )
	{
		ImageFormat::PackPixels(ds.pixels, ds.width, ds.height, ds.pixelFormat);
	}

	for(int iFrame = 0; iFrame < frameCount; ++iFrame)
	{
		ApngFrame* pFrame = ds.frames[iFrame];

		const int width = pFrame->GetWidth();
		const int height = pFrame->GetHeight();
		ImageFormat::PackPixels(pFrame->m_pixels, width, height, ds.pixelFormat);
	}
	return true;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////
bool POEngine::DumpBestResultToFile(const String& strNewFileName)
{
	DynamicMemoryFile& dmf = m_resultmgr.GetSmallest();
	int32 nSizeToDump = (int32) dmf.GetPosition();

	File file;
	if( !file.Open(strNewFileName, File::modeWrite) )
	{
		AddError(k_szCannotWriteOnDevice);
		return false;
	}

	if( m_settings.keepFileDate )
	{
		file.SetLastWriteTime(m_originalFileWriteTime);
	}

	int32 nWritten = file.Write(dmf.GetContent().GetReadPtr(), nSizeToDump);
	if( nWritten != nSizeToDump )
	{
		AddError(k_szCannotWriteUncomplete);
		return false;
	}
	file.Close();
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// For 8bpp only
void AddFrameColorCounts(const ApngFrame* pFrame, uint32* pCounts)
{
	const int width = pFrame->GetWidth();
	const int height = pFrame->GetHeight();
	const int pixelCount = width * height;
	const uint8* pPixels = pFrame->m_pixels.GetReadPtr();
	for(int i = 0; i < pixelCount; ++i)
	{
		pCounts[ pPixels[i] ]++;
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////////
void CountColors(const PngDumpSettings& ds, uint32* pCounts)
{
	ASSERT( ds.pixelFormat == pf8bppIndexed );
	
	Memory::Zero32(pCounts, 256);
	
	const int frameCount = ds.frames.GetSize();

	if( ds.hasDefaultImage || frameCount == 0 )
	{
		const int pixelCount = ds.width * ds.height;
		const uint8* pPixels = ds.pixels.GetReadPtr();
		for(int i = 0; i < pixelCount; ++i)
		{
			pCounts[ pPixels[i] ]++;
		}
	}
	
	for(int iFrame = 0; iFrame < frameCount; ++iFrame)
	{
		const ApngFrame* pFrame = ds.frames[iFrame];
		AddFrameColorCounts(pFrame, pCounts);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////
// ds : input image information
bool POEngine::OptimizePaletteMode(PngDumpSettings& ds)
{
	ASSERT( pf1bppIndexed <= ds.pixelFormat && ds.pixelFormat <= pf8bppIndexed);

	UnpackPixelFrames(ds);
	
	// Count the number of times a color appears
	uint32 colCounts[256];
	CountColors(ds, colCounts);
	
	// Remove holes in the palette and create the conversion table old index -> new index
	PaletteTranslator noHolesTranslator;
	noHolesTranslator.BuildUnusedColors(ds.palette, colCounts);

	// Remove duplicated colors
	PaletteTranslator dupTranslator;
	dupTranslator.BuildDuplicatedColors(ds.palette, colCounts);

	// Sort the palette according to an increasing alpha
	PaletteTranslator sortedAlphaTranslator;
	sortedAlphaTranslator.BuildSortAlpha(ds.palette, colCounts);

	// Change the pixel buffer so every index points on the new palette
	PaletteTranslator tmpTranslator = PaletteTranslator::Combine(noHolesTranslator, dupTranslator);
	PaletteTranslator finalTranslator = PaletteTranslator::Combine(tmpTranslator, sortedAlphaTranslator);
	finalTranslator.TranslateAll(ds);

	// Check if we can convert to Black&White
	if( TryToConvertIndexedToBlackAndWhite(ds) )
	{
		return PerformDumpTries(ds);
	}

	// Some tries with different orders
	// To get a stable optimization, sort by luminance before sorting by population

	//////////////////////////
	PaletteTranslator luminanceTranslator;
	luminanceTranslator.BuildSortLuminance(ds.palette, colCounts);
	luminanceTranslator.TranslateAll(ds);
	PackPixelFrames(ds);
	if( !PerformDumpTries(ds) )
	{
		return false;
	}

	// Save CPU or Memory ? :-p
	UnpackPixelFrames(ds);

	//////////////////////////
	// Sometimes we get some best result by sorting twice
	PaletteTranslator populationTranslator;
	populationTranslator.BuildSortPopulation(ds.palette, colCounts);

	PaletteTranslator populationTranslator2;
	populationTranslator2.BuildSortPopulation(ds.palette, colCounts);

	PaletteTranslator finalPopulationTranslator = PaletteTranslator::Combine(populationTranslator, populationTranslator2);
	finalPopulationTranslator.TranslateAll(ds);

	PackPixelFrames(ds);
	if( !PerformDumpTries(ds) )
	{
		return false;
	}
	// We're done
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Find an unused color among the pixels
// Needs memory allocation, thus can fail
bool POEngine::FindUnusedColorHardcoreMethod(const uint8* pRgba, int32 nPixelCount, 
                                             uint8& nRed, uint8& nGreen, uint8& nBlue)
{
	ByteArray aSorted;
	if( !aSorted.SetSize(nPixelCount * 4) )
	{
		return false;
	}

	ByteArray aSorted2;
	if( !aSorted2.SetSize(nPixelCount * 4) )
	{
		return false;
	}

	uint32* pRgba32 = (uint32*) pRgba;

	// ByteSort function that works the same on both endian mode
	chustd::Sort::ByteSortLittleEndian( pRgba32, (uint32*)aSorted.GetPtr(), (uint32*)aSorted2.GetPtr(), nPixelCount);

	// This first entry is pure black with alpha set to 0,
	// If we are in this function it's because it is not suitable so we start at 1

	if( k_ePlatformByteOrder == boBigEndian )
	{
		// Convert uint32 to little endian
		IFile::Swap32( (uint32*) aSorted.GetPtr(), nPixelCount);
	}

	uint32* pSorted32 = (uint32*) aSorted.GetPtr();

	uint32 nA = 0x00000000; // ABGR in register
	for(int32 i = 1; i < nPixelCount; ++i)
	{
		uint32 nB = (0x00ffffff & pSorted32[i]);
		uint32 nDif = nB - nA;
		if( nDif > 1 )
		{
			// Found a useable range of colors to be used as transparent
			break;
		}
		nA = nB; // Next entry
	}

	if( nA == 0x00ffffff )
	{
		// This can only occurs if all 16777216 colors are present in the picture
		// Like a 4096*4096 picture
		return false;
	}
	
	uint32 nColor = nA + 1;

	nRed = uint8(nColor & 0x000000ff);
	nGreen = uint8((nColor & 0x0000ff00) >> 8);
	nBlue = uint8((nColor & 0x00ff0000) >> 16);
	
	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////
// Find an unused color among the pixels
// rbRgb [in] Pixel buffer
// nRed, nGreen, nBlue [out] : color not present in the pixel buffer
// Returns true upon success
bool POEngine::FindUnusedColor(const Buffer& aRgb, uint8& nRed, uint8& nGreen, uint8& nBlue)
{
	// Check with some candidates
	FixArray<Color32, 7> aCandidates;
	aCandidates[0].SetRgb(1, 0, 0);
	aCandidates[1].SetRgb(1, 1, 0);
	aCandidates[2].SetRgb(1, 1, 1);
	aCandidates[3].SetRgb(0, 1, 1);
	aCandidates[4].SetRgb(0, 0, 1);
	aCandidates[5].SetRgb(1, 0, 1);
	aCandidates[6].SetRgb(0, 1, 0);
	
	const int32 nPixelCount = aRgb.GetSize() / 3;
	
	foreach(aCandidates, iCandidate)
	{
		bool bGoodCandidate = true;

		uint8 cr, cg, cb;
		aCandidates[iCandidate].ToRgb(cr, cg, cb);

		const uint8* pPixelsRgb = aRgb.GetReadPtr();
		for(int32 i = 0; i < nPixelCount; ++i)
		{
			uint8 r = pPixelsRgb[0];
			uint8 g = pPixelsRgb[1];
			uint8 b = pPixelsRgb[2];

			if( r == cr && g == cg && b == cb )
			{
				// Found, not a good candidate
				bGoodCandidate = false;
				break;
			}

			pPixelsRgb += 3;
		}

		if( bGoodCandidate )
		{
			nRed = cr;
			nGreen = cg;
			nBlue = cb;
			return true;
		}
	}
	return false;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
bool POEngine::Optimize32BitsMode(PngDumpSettings& ds)
{
	ASSERT(ds.pixelFormat == pf32bppRgba);

	const int32 nWidth = ds.width;
	const int32 nHeight = ds.height;
	const int32 nPixelCount = nWidth * nHeight;

	///////////////////////////////////////////////////////////////////////////////////////////////
	// First step : set to 0 every color which alpha is 0
	
	// It will allow a potential optimisation 32 bits --> palette mode
	// We set the fully transparent color to 0 as 4 bytes to 0 are nicely compressed

	uint8* pPixels8 = ds.pixels.GetWritePtr();

	for(int32 iPixel = 0; iPixel < nPixelCount; ++iPixel)
	{
		// In memory : RGBA, RGBA, RGBA
		//             0123  0123  0123
		uint8 nAlpha = pPixels8[3];

		if( nAlpha == 0 )
		{
			// Alpha set to 0, set everything to 0
			uint32* pPixels32 = (uint32*) pPixels8;
			pPixels32[0] = 0x00000000;
		}

		pPixels8 += 4;
	}

	// Test 0
	// We save the result as-is, sometimes a small 32 bits image can be smaller than the same one in 24 or 8 bits
	if( nPixelCount <= (64 * 64) )
	{
		// Our definition of "small" is 64*64
		if( !CustomPngDump(m_resultmgr.GetCandidate(), ds) )
		{
			AddError(k_szCannotDumpTry);
			return false;
		}
	}

	///////////////////////////////////////////////////////////////////////////////////////////////
	// Verify the alpha channel is really necessary
	const uint8* pBuffer = ds.pixels.GetReadPtr();
	const uint8* pSrcBuffer = pBuffer;

	Buffer rbNewRgb;
	if( !rbNewRgb.SetSize(nPixelCount * 3) )
	{
		AddError(k_szNotEnoughMemoryToConvertTo24Bits);
		return false;
	}

	uint8* pNewBuffer = rbNewRgb.GetWritePtr();

	// To count how many times the same alpha appears in the image
	int32 anAlphas[256];
	Memory::Zero32(anAlphas, 256);

	for(int32 i = 0; i < nPixelCount; ++i)
	{
		uint8 r = pSrcBuffer[0];
		uint8 g = pSrcBuffer[1];
		uint8 b = pSrcBuffer[2];
		uint8 nAlpha = pSrcBuffer[3];

		pNewBuffer[0] = r;
		pNewBuffer[1] = g;
		pNewBuffer[2] = b;

		anAlphas[nAlpha] += 1;
		
		pSrcBuffer += 4;
		pNewBuffer += 3;
	}

	// All pixels are opaque, no need for an alpha channel, use the 24 bits optimizer
	if( anAlphas[255] == nPixelCount )
	{
		ds.pixels = rbNewRgb;
		ds.pixelFormat = pf24bppRgb;
	
		return Optimize24BitsMode(ds);
	}

	// Ok, maybe we can keep the 24 bits buffer if every alpha is set to 255 except for one color

	if( (anAlphas[0] + anAlphas[255]) == nPixelCount )
	{
		// Yes we can, but the color to be used to mean a pixel is totaly transparent must not appear
		// elsewhere the picture as being an opaque pixel.

		// It means we must choose a color not used in the picture, that's a bit tricky

		////////////////////////////////////////////////////
		// First, check if black is a used color
		
		bool bBlackAsTransparentColor = true;
		uint8* pPixelsRgba = ds.pixels.GetWritePtr();

		for(int32 i = 0; i < nPixelCount; ++i)
		{
			uint8 r = pPixelsRgba[0];
			uint8 g = pPixelsRgba[1];
			uint8 b = pPixelsRgba[2];
			uint8 nAlpha = pPixelsRgba[3];
			if( nAlpha == 255 )
			{
				if( r == 0 && g == 0 && b == 0 )
				{
					// Yes, black is a used color, we cannot use it as a the transparent color
					bBlackAsTransparentColor = false;
					break;
				}
			}
			
			pPixelsRgba += 4;
		}
		////////////////////////////////////////////////////
		
		uint8 nTransRed = 0;
		uint8 nTransGreen = 0;
		uint8 nTransBlue = 0;

		bool bContinueIn24Bits = bBlackAsTransparentColor;

		if( !bBlackAsTransparentColor )
		{
			// Find an alternative transparent color
			if(  FindUnusedColor(rbNewRgb, nTransRed, nTransGreen, nTransBlue)
			  || FindUnusedColorHardcoreMethod(ds.pixels.GetReadPtr(), nPixelCount, nTransRed, nTransGreen, nTransBlue) )
			{
				bContinueIn24Bits = true;

				// Ok, we have our color to be use for transparency :)
				// Convert every pixel to that color

				uint8* pPixelsRgb = rbNewRgb.GetWritePtr();
				const uint8* pPixelsRgba = ds.pixels.GetReadPtr();

				for(int32 i = 0; i < nPixelCount; ++i)
				{
					uint8 nAlpha = pPixelsRgba[3];
					if( nAlpha == 0 )
					{
						// Put the transparent color
						pPixelsRgb[0] = nTransRed;
						pPixelsRgb[1] = nTransGreen;
						pPixelsRgb[2] = nTransBlue;
					}
					
					pPixelsRgba += 4;
					pPixelsRgb += 3;
				}
			}
		}
		
		if( bContinueIn24Bits )
		{
			// Now we have our 24 bits image + one color for transparency, continue with 24 bits optimization...
			ds.pixels = rbNewRgb;
			ds.pixelFormat = pf24bppRgb;
			ds.useTransparentColor = true;
			ds.tRNS.red = nTransRed;
			ds.tRNS.green = nTransGreen;
			ds.tRNS.blue = nTransBlue;

			return Optimize24BitsMode(ds);
		}
	}

	///////////////////////////////////////////////////////////////////
	// Dump to memory

	//ds.pBuffer = pBuffer;
	ds.pixelFormat = pf32bppRgba;
	return PerformDumpTries(ds);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////
bool POEngine::Optimize24BitsMode(PngDumpSettings& ds)
{
	ASSERT(ds.pixelFormat == pf24bppRgb);

	const uint8* pBuffer = ds.pixels.GetReadPtr();
	const int32 nWidth = ds.width;
	const int32 nHeight = ds.height;

	////////////////////////////////////////////////////////////////
	// Some test to check if we can switch to palette mode
	Palette palTest;
	palTest.m_count = 0;

	const uint8* pSrcBuffer = pBuffer;
	const int32 nPixelCount = nWidth * nHeight;

	Buffer rbNew;
	if( !rbNew.SetSize(nPixelCount) )
	{
		return false;
	}
	uint8* pNewBuffer = rbNew.GetWritePtr();

	bool bTooMuchColors = false;
	for(int32 i = 0; i < nPixelCount; ++i)
	{
		uint8 r = pSrcBuffer[0];
		uint8 g = pSrcBuffer[1];
		uint8 b = pSrcBuffer[2];
		
		// Find the color in the palette
		int32 iCol = 0;
		for(; iCol < palTest.m_count; ++iCol)
		{
			const Color32& col = palTest.m_colors[iCol];
			uint8 rp, gp, bp;
			col.ToRgb(rp, gp, bp);

			if( r == rp && g == gp && b == bp )
			{
				// Found !
				break;
			}
		}

		pNewBuffer[0] = uint8(iCol); // Fill the new buffer

		if( iCol == palTest.m_count )
		{
			// The 24 bit color was not in the palette

			// Add a new entry in the palette
			if( palTest.m_count == 256 )
			{
				// The palette is full
				bTooMuchColors = true;
				break;
			}

			// New entry in the palette
			palTest.m_colors[ palTest.m_count].SetRgb(r, g, b);
			palTest.m_count++;
		}

		pSrcBuffer += 3;
		pNewBuffer += 1;
	}

	// If the picture is very small, we may enlarge it if we switch to palette mode
	const int32 nMaxSizeOrigin = nPixelCount * 3;
	const int32 nMaxSizeNew = nPixelCount + palTest.m_count * 3 + 12; // 12 = min size chunk
	if( nMaxSizeOrigin < nMaxSizeNew )
	{
		bTooMuchColors = true;
	}
	
	///////////////////////////////////////////////////////////////////
	//ds.pBuffer = pBuffer;
	ds.pixelFormat = pf24bppRgb;
	if( !PerformDumpTries(ds) )
	{
		return false;
	}

	///////////////////////////////////////////////////////////////////

	if( bTooMuchColors )
	{
		// Cannot convert to palette mode, we stop here
		return true;
	}

	// Add transparency to the palette
	if( ds.useTransparentColor )
	{
		// Find the transparent color
		int32 iCol = 0;
		for(; iCol < palTest.m_count; ++iCol)
		{
			const Color32& col = palTest.m_colors[iCol];
			uint8 rp, gp, bp;
			col.ToRgb(rp, gp, bp);

			if( ds.tRNS.red == rp && ds.tRNS.green == gp && ds.tRNS.blue == bp )
			{
				// Aha ! :)
				palTest.m_colors[iCol].SetAlpha(0);
				break;
			}
		}
	}

	// Now we have a 8 bits indexed image, we can continue with the palette mode optimizer
	ds.pixels = rbNew;
	ds.palette = palTest;
	ds.pixelFormat = pf8bppIndexed;
	return OptimizePaletteMode(ds);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////
bool POEngine::OptimizeGrayScale(PngDumpSettings& ds)
{
	ASSERT( pf1bppGrayScale <= ds.pixelFormat && ds.pixelFormat <= pf16bppGrayScale );

	return PerformDumpTries(ds);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
void POEngine::BgraToRgba(PngDumpSettings& ds)
{
	uint8* pBuffer = ds.pixels.GetWritePtr();
	int32 nPixelCount = ds.width * ds.height;

	for(int32 i = 0; i < nPixelCount; ++i)
	{
		uint8 b = pBuffer[0];
		uint8 r = pBuffer[2];
		pBuffer[0] = r;
		pBuffer[2] = b;

		pBuffer += 4;
	}

	ds.pixelFormat = pf32bppRgba;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
void POEngine::BgrToRgb(PngDumpSettings& ds)
{
	uint8* pBuffer = ds.pixels.GetWritePtr();
	int32 nPixelCount = ds.width * ds.height;

	for(int32 i = 0; i < nPixelCount; ++i)
	{
		uint8 b = pBuffer[0];
		uint8 r = pBuffer[2];
		pBuffer[0] = r;
		pBuffer[2] = b;

		pBuffer += 3;
	}

	ds.pixelFormat = pf24bppRgb;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
bool POEngine::Rgb16ToRgb24(PngDumpSettings& ds)
{
	const uint8* pSrc = ds.pixels.GetReadPtr();
	int32 nPixelCount = ds.width * ds.height;

	Buffer newBuffer;
	if( !newBuffer.SetSize(nPixelCount * 3) )
	{
		return false;
	}

	uint8* pDst = newBuffer.GetWritePtr();

	if( ds.pixelFormat == pf16bppRgb555 )
	{
		for(int32 i = 0; i < nPixelCount; i++)
		{
			uint16 nPixel = uint16( (pSrc[0] << 8) | pSrc[1]);
			pSrc += 2;

			uint8 r = uint8((nPixel & 0x7c00) >> 10);
			uint8 g = uint8((nPixel & 0x03e0) >> 5);
			uint8 b = uint8((nPixel & 0x001f));

			r = uint8( (r << 3) + ((r >> 2) & 0x07));
			g = uint8( (g << 3) + ((g >> 2) & 0x07));
			b = uint8( (b << 3) + ((b >> 2) & 0x07));

			pDst[0] = uint8(r);
			pDst[1] = uint8(g);
			pDst[2] = uint8(b);
			pDst += 3;
		}
	}
	else if( ds.pixelFormat == pf16bppRgb565 )
	{
		for(int32 i = 0; i < nPixelCount; i++)
		{
			uint16 nPixel = uint16( (pSrc[0] << 8) | pSrc[1]);
			pSrc += 2;

			uint8 r = uint8( (nPixel & 0xf800) >> 11 );
			uint8 g = uint8( (nPixel & 0x07e0) >> 5 );
			uint8 b = uint8( (nPixel & 0x001f) );

			r = uint8( (r << 3) + ((r >> 2) & 0x07));
			g = uint8( (g << 2) + ((g >> 1) & 0x03));
			b = uint8( (b << 3) + ((b >> 2) & 0x07));

			pDst[0] = uint8(r);
			pDst[1] = uint8(g);
			pDst[2] = uint8(b);
			pDst += 3;
		}
	}
	else
	{
		return false;
	}

	ds.pixelFormat = pf24bppRgb;
	ds.pixels = newBuffer;
	return true;
}

/////////////////////////////////////////////////////////////////////////////////////////////////
// Public version of the optimizing function
bool POEngine::OptimizeExternalBuffer(PngDumpSettings& ds, const String& strNewFileName)
{
	m_astrErrors.SetSize(0);
	m_resultmgr.Reset();
	m_originalFileWriteTime = 0;

	return Optimize(ds, strNewFileName);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
bool POEngine::Optimize(PngDumpSettings& ds, const String& strNewFileName)
{
	if( ds.pixelFormat == pf32bppBgra )
	{
		// Yuk ! Change this
		BgraToRgba(ds);
	}
	else if( ds.pixelFormat == pf24bppBgr )
	{
		// Yuk ! Change this aswell
		BgrToRgb(ds);
	}
	else if( ds.pixelFormat == pf16bppRgb555 || ds.pixelFormat == pf16bppRgb565 )
	{
		// Change this, but harder to do this time because we need to reallocate
		if( !Rgb16ToRgb24(ds) )
		{
			AddError(k_szNotEnoughMemoryForInternalConversion);
			return false;
		}
	}

	ds.zlibCompressionLevel = 9; // Best compression

	PixelFormat pf = ds.pixelFormat;

	bool bOptimizeOk = false;

	if( pf1bppIndexed <= pf && pf <= pf8bppIndexed )
	{
		bOptimizeOk = OptimizePaletteMode(ds);
	}
	else if( pf == pf32bppRgba )
	{
		bOptimizeOk = Optimize32BitsMode(ds);
	}
	else if( pf == pf24bppRgb )
	{
		bOptimizeOk = Optimize24BitsMode(ds);
	}
	else if( pf1bppGrayScale <= pf && pf <= pf16bppGrayScale )
	{
		bOptimizeOk = OptimizeGrayScale(ds);
	}
	else if( pf == pf48bppRgb || pf == pf64bppRgba
	         || pf == pf16bppGrayScaleAlpha || pf == pf32bppGrayScaleAlpha )
	{
		// Other uncommon pixel formats
		
		// Test 1 : do a regular dump
		bOptimizeOk = CustomPngDump(m_resultmgr.GetCandidate(), ds);

		if( bOptimizeOk )
		{
			// Test 2
			//  - ZLib strategy : filter
			ds.zlibStrategy = PngDumpSettings::zlibStrategyFilter;

			bOptimizeOk = CustomPngDump(m_resultmgr.GetCandidate(), ds);
		}
	}
	else
	{
		AddError(k_szUnsupportedPixelFormat);
		return false;
	}

	if( !bOptimizeOk )
	{
		return false;
	}

	return DumpBestResultToFile(strNewFileName);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
void POEngine::PrintText(const String& strText, enum TextColor eColor)
{
	m_eventProgressing.Fire(strText, eColor);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Builds a new PNG from another one (given by file) excluding unwanted chunks
//
// file: Original file
//
// Returns true upon success
///////////////////////////////////////////////////////////////////////////////////////////////////
bool POEngine::InsertCleanOriginalPngAsResult(IFile& file)
{
	if( !Png::IsPng(file) )
	{
		// Not a PNG
		ASSERT(0); // Programming error
		return false;
	}

	const int64 fileSize = file.GetSize();
	const int32 fileSize32 = int32(fileSize);

	DynamicMemoryFile& dmf = m_resultmgr.GetCandidate();
	dmf.SetByteOrder(boBigEndian);
	
	// Ensure capacity
	if( !dmf.EnsureCapacity(fileSize32) )
	{
		AddError(k_szNotEnoughMemoryToKeepOriginalFile);
		return false;
	}
	dmf.SetPosition(0);

	Png::WriteSignature(dmf);

	ChunkedFile chfIn(file);

	bool bOkHandled = true;
	bool bIENDFound = false;
	bool forcedChunksWritten = false;
	PngChunk_IHDR pngIHDR; // We need to get some information here

	while( !bIENDFound)
	{
		bool bChunkPresent = chfIn.BeginChunkRead();
		if( !bChunkPresent )
		{
			break;
		}

		if( chfIn.m_nChunk == Png::chunk_IHDR )
		{
			// Read this one to get some info
			int64 posBefore = file.GetPosition();
			if( Png::Read_IHDR(file, chfIn.m_nSizeof, pngIHDR) != 0 )
			{
				AddError(k_szCannotLoadFile);
				return false;
			}
			file.SetPosition(posBefore);
		}

		// If we remove or force a chunk, we remove the original one(s)
		bool keepChunk = false;
		if( chfIn.m_nChunk == Png::chunk_bKGD )
		{
			keepChunk = (m_settings.bkgdOption == POChunk_Keep);
		}
		else if( chfIn.m_nChunk == Png::chunk_tEXt )
		{
			keepChunk = (m_settings.textOption == POChunk_Keep);
		}
		else
		{
			switch( chfIn.m_nChunk )
			{
			case Png::chunk_IEND:
				bIENDFound = true;
			case Png::chunk_IHDR:
			case Png::chunk_PLTE:
			case Png::chunk_IDAT:
			case Png::chunk_tRNS:
			case Png::chunk_acTL:
			case Png::chunk_fcTL:
			case Png::chunk_fdAT:
				keepChunk = true;
				break;

			default:
				break;
			}
		}

		// Write forced chunks if any, before acTL if found, or IDAT
		if( !forcedChunksWritten && (chfIn.m_nChunk == Png::chunk_acTL || chfIn.m_nChunk == Png::chunk_IDAT) )
		{
			if( m_settings.bkgdOption == POChunk_Force )
			{
				PixelFormat pf = Png::GetPixelFormat(pngIHDR);
				if( pf == pf32bppRgba )
				{
					// Only for this mode
					PngChunk_bkGD bkGD;
					bkGD.red = m_settings.bkgdColor.r;
					bkGD.green = m_settings.bkgdColor.g;
					bkGD.blue = m_settings.bkgdColor.b;

					ChunkedFile cf(dmf);
					if( !Png::WriteChunk_bkGD(cf, pngIHDR.colorType, bkGD) )
					{
						return false;
					}
				}
			}

			if( m_settings.textOption == POChunk_Force )
			{
				PngChunk_tEXt text;
				text.keyword = m_settings.textKeyword;
				text.data = m_settings.textData;

				ChunkedFile cf(dmf);
				if( !Png::WriteChunk_tEXt(cf, text) )
				{
					return false;
				}
			}

			forcedChunksWritten = true;
		}

		if( keepChunk )
		{
			// We keep that chunk
			dmf.Write32( chfIn.m_nSizeof);
			dmf.Write32( chfIn.m_nChunk);
			if( dmf.WriteFromFile(chfIn, chfIn.m_nSizeof) != chfIn.m_nSizeof )
			{
				// The source file is invalid
				bOkHandled = false;
				break;
			}

			if( !chfIn.EndChunkRead() )
			{
				// The source file is invalid
				bOkHandled = false;
				break;
			}
			else
			{
				// We bypassed the CRC computation, so we just write the one we read
				dmf.Write32(chfIn.m_nCrc);
			}
		}
		else
		{
			if( !chfIn.EndChunkRead() )
			{
				// Unexpected end of stream, CRC not present or chunk mishandled (read too much)
				bOkHandled = false;
				break;
			}
		}
	}

	if( !(bOkHandled && bIENDFound) )
	{
		AddError(k_szCorruptedChunkStructure);
		return false;
	}
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
bool POEngine::OptimizeFileNoBackup(const String& strFileName, const String& strNewFileName)
{
	m_astrErrors.SetSize(0);
	m_resultmgr.Reset();
	m_originalFileWriteTime = 0;

	File fileImage;
	if( !fileImage.Open(strFileName, File::modeRead) )
	{
		AddError(k_szCannotLoadFile);
		return false;
	}

	if( m_settings.keepFileDate )
	{
		m_originalFileWriteTime = fileImage.GetLastWriteTime();
	}

	const int64 fileSize = fileImage.GetSize();
	if( fileSize > MAX_INT32 )
	{
		AddError(k_szFileTooLarge);
		return false;
	}

	///////////////////////////////////////////////////////////////////////////
	ImageLoader imgloader;
	if( !imgloader.InstanciateLosslessFormat(fileImage) )
	{
		AddError(k_szUnsupportedFileFormat);
		return false;
	}

	ImageFormat& img = *imgloader.m_pImageType;
	bool bLoadOk = false;

	if( imgloader.m_nType == ImageLoader::typePng )
	{
		// Load as-is in memory
		DynamicMemoryFile dmfAsIs;
		if( !dmfAsIs.Open(int32(fileSize)) )
		{
			AddError(k_szCannotLoadFile);
			return false;
		}
		if( dmfAsIs.WriteFromFile(fileImage, int32(fileSize)) != fileSize )
		{
			AddError(k_szCannotLoadFile);
			return false;
		}
		fileImage.Close(); // No need for the file anymore

		dmfAsIs.SetPosition(0);
		bLoadOk = img.LoadFromFile(dmfAsIs);

		// Insert a clean version of the source PNG
		// "clean" means the same PNG expect some unwanted chunks (like the gamma chunk)
		
		// As we insert a copy of the source file, the source file becomes a candidate for the best result,
		// thus if we cannot achieve a better compression than the original file, we just dump the original file
		dmfAsIs.SetPosition(0);
		if( !InsertCleanOriginalPngAsResult(dmfAsIs) )
		{
			return false;
		}
	}
	else
	{
		bLoadOk = img.LoadFromFile(fileImage);
		fileImage.Close();
	}

	if( !bLoadOk )
	{
		String strImgErr = img.GetLastErrorString();
		
		String strErr = L"Cannot load image: ";
		strErr = strErr + strImgErr;
		
		AddError(strErr);
		return false;
	}

	const int32 nWidth = img.GetWidth();
	const int32 nHeight = img.GetHeight();
	const Buffer& pixels = img.GetPixels();
	const Palette& palette = img.GetPalette();
	PixelFormat pf = img.GetPixelFormat();
	
	//////////////////////////////////////////////////////////////
	PngDumpSettings ds; // Final dump structure
	ds.width = nWidth;
	ds.height = nHeight;
	ds.zlibCompressionLevel = 9; // Best compression
	ds.pixels = pixels;
	ds.palette = palette;
	ds.pixelFormat = pf;

	if( imgloader.m_nType == ImageLoader::typePng )
	{
		const Png& png = (const Png&) img;
		ds.useTransparentColor = png.IsSimpleTransparencyUsed();
		ds.tRNS.grey = png.GetGreyTransIndex();
		png.GetTransIndexes(ds.tRNS.red, ds.tRNS.green, ds.tRNS.blue);

		if( m_settings.keepInterlacing )
		{
			if( png.IsInterlaced() )
			{
				ds.interlaced = true;
			}
		}

		if( m_settings.bkgdOption == POChunk_Keep )
		{
			// We only keep the background color for non-palette images
			// Why ?
			// 1. Tracking the correct background index through a palette reorganization is very tricky
			// 2. Keeping background color is used so 32 bits images are not blended with gray
			// in Internet Explorer 6
			//
			// So let's wait before a real need comes

			if( pf1bppIndexed <= pf && pf <= pf8bppIndexed )
			{
				// Give up
			}
			else
			{
				ds.useBackgroundColor = png.HasBackgroundColor();
				ds.bkGD.grey = png.GetBkGrey();
				ds.bkGD.red = png.GetBkRed();
				ds.bkGD.green = png.GetBkGreen();
				ds.bkGD.blue = png.GetBkBlue();
			}
		}

		if( m_settings.textOption == POChunk_Keep )
		{
			ds.textInfos = png.GetTexts();
		}

		if( png.IsAnimated() )
		{
			PrintText(L"[APNG] ", tcAnimated);
		}
	}
	else if( imgloader.m_nType == ImageLoader::typeGif )
	{
		const Gif& gif = (const Gif&) img;
		if( gif.IsAnimated() )
		{
			if( m_settings.ignoreAnimatedGifs )
			{
				PrintText(L"[Animated GIF: ignored]\n", tcAnimated);
				return false;
			}
			else
			{
				PrintText(L"[Animated GIF: converting to APNG] ", tcAnimated);
			}
		}
	}

	if( m_settings.bkgdOption == POChunk_Force )
	{
		ds.useBackgroundColor = true;
		ds.bkGD.Clear();
		ds.bkGD.red = m_settings.bkgdColor.r;
		ds.bkGD.green = m_settings.bkgdColor.g;
		ds.bkGD.blue = m_settings.bkgdColor.b;
	}
	if( m_settings.textOption == POChunk_Force )
	{
		PngChunk_tEXt text;
		text.keyword = m_settings.textKeyword;
		text.data = m_settings.textData;
		ds.textInfos.SetSize(1);
		ds.textInfos[0] = text;
	}

	if( img.IsAnimated() )
	{
		return OptimizeAnimated(img, ds, strNewFileName);
	}
	return Optimize(ds, strNewFileName);
}


///////////////////////////////////////////////////////////////////////////////////////////////////
void POEngine::OptimizeFileOrDir(const String& strFilePath, const String& strDisplayDir)
{
	m_astrErrors.SetSize(0);

	bool bSrcIsDir = false;
	bool bSrcIsReadOnly = false;
	if( !File::GetFileAttributes(strFilePath, bSrcIsDir, bSrcIsReadOnly) )
	{
		// We cannot even read file attributes, so we just stop there for this file
		AddError(k_szPathDoesNotExist);
		return;
	}

	// We need the name alone
	String strDirOnly, strNameOnly;
	FilePath::Split(strFilePath, strDirOnly, strNameOnly);

	if( bSrcIsDir )
	{
		// A directory, we optimize it before we continue
		StringArray astrSubFileNames = Directory::GetFileNames(strFilePath, L"*.*");

		// Set the directory to display
		String strNewDisplayDir;
		if( !strDisplayDir.IsEmpty() )
		{
			strNewDisplayDir = strDisplayDir + L"/";
		}
		strNewDisplayDir = strNewDisplayDir + strNameOnly;

		OptimizeFiles(strFilePath, astrSubFileNames, strNewDisplayDir);
		return;
	}

	// Filter by extension
	String strExt = File::GetExtension(strFilePath);
	strExt = strExt.ToLowerCase();

	bool bExtOk = (strExt == L"png" || strExt == L"gif" || strExt == L"bmp" || strExt == L"tga");
	if( !bExtOk )
	{
		return;
	}
			
	///////////////////////////////////////////////////////////////////////
	// Write the waiting message during the optimization
	if( strExt == L"png" )
	{
		PrintText(L"Optimizing ", tcActionVerb);
	}
	else
	{
		PrintText(L"Converting ", tcActionVerb);
	}
	///////////////////////////////////////////////////////////////////////
	
	if( !strDisplayDir.IsEmpty() )
	{
		// Display the sub-directory
		PrintText(strDisplayDir + L"/", tcFilePath);
	}
	PrintText(strNameOnly, tcFilePath);
	
	PrintText(L"... ", tcRegularInfo);

	// Get the source file size
	uint32 nOldSize = (uint32) File::GetSize(strFilePath);

	strDirOnly = FilePath::AddSeparator(strDirOnly);

	String strOldFilePath;
	String strNewFilePath;
	
	// If the source file is a PNG, we rename that source file with a "_" at the beginning of its name
	if( strExt == L"png" )
	{
		if( bSrcIsReadOnly )
		{
			// File is read only, we won't be able to modify it
			AddError(k_szFileIsReadOnly);
			return;
		}

		if( m_settings.backupOldPngFiles )
		{
			String strBackupFilePath = strDirOnly + L"_" + strNameOnly;

			// Delete a possible previous backup file
			if( File::Exists(strBackupFilePath) )
			{
				if( !File::Delete(strBackupFilePath) )
				{
					// We cannot delete the old backup file, we stop here
					AddError(k_szCannotPerformBackup);
					return;
				}
			}

			if( File::Rename(strFilePath, strBackupFilePath) )
			{
				// Rename possible
				strOldFilePath = strBackupFilePath;
				strNewFilePath = strFilePath;
			}
			else
			{
				// We cannot rename for unknown reason
				AddError(k_szCannotPerformBackupRenameFailed);
				return;
			}
		}
		else
		{
			strOldFilePath = strFilePath;
			strNewFilePath = strFilePath;
		}
	}
	else
	{
		// Not a PNG file
		strOldFilePath = strFilePath;
		strNewFilePath = strDirOnly + FilePath::RemoveExtension(strNameOnly) + L".png";
	}

	// TMP DEBUG : to compare size before and after, uncomment the line above
	//strNewFilePath = File::AddFileNamePrefix(strNewFilePath, L"optimized-");

	if( OptimizeFileNoBackup(strOldFilePath, strNewFilePath) )
	{
		PrintText(L" (OK) ", tcActionOk);

		uint32 nNewSize = (uint32) File::GetSize(strNewFilePath);

		if( nOldSize > 0 && nNewSize > 0 )
		{
			int64  nRatio = nNewSize;
			nRatio *= 100;
			nRatio /= nOldSize;
			
			String strOldSize = String::FromInt(nOldSize);
			String strNewSize = String::FromInt(nNewSize);
			String strRatio = String::FromInt(uint32(nRatio));

			StringBuilder sb;
			sb += strOldSize;
			sb += L" bytes ";
			sb += L"-> "; // Arrow to the right
			sb += strNewSize;
			sb += L" bytes ";

			PrintText(sb.ToString(), tcSizeInfo);

			sb.Empty();

			sb += '(';
			sb += strRatio;
			sb += L"% of the original size)\n"; // End of line here
			
			if( nNewSize > nOldSize )
			{
				PrintText(sb.ToString(), tcFileEnlarged);
			}
			else if( nNewSize == nOldSize )
			{
				PrintText(sb.ToString(), tcFileSizeSame);
			}
			else
			{
				PrintText(sb.ToString(), tcFileReduced);
			}
		}
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// astrFilePaths : absolute file paths of files to optimize
///////////////////////////////////////////////////////////////////////////////////////////////////
void POEngine::OptimizeFiles(const StringArray& astrFilePaths)
{
	OptimizeFiles(L"", astrFilePaths, L"");
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// strDir        : directory name containing the files
// astrFileNames : file paths relative to base dir
///////////////////////////////////////////////////////////////////////////////////////////////////
void POEngine::OptimizeFiles(const String& strDir, const StringArray& astrFileNames)
{
	OptimizeFiles(strDir, astrFileNames, L"");
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// strBaseDir :    directory name containing the files
// astrFileNames : file path relative to base dir
// strDisplayDir : directory name to display
///////////////////////////////////////////////////////////////////////////////////////////////////
void POEngine::OptimizeFiles(const String& strBaseDir, const StringArray& astrFileNames, const String& strDisplayDir)
{
	const int32 nFileCount = astrFileNames.GetSize();
	for(int32 iFile = 0; iFile < nFileCount; ++iFile)
	{
		String strFilePath = FilePath::Combine(strBaseDir, astrFileNames[iFile]);

		OptimizeFileOrDir(strFilePath, strDisplayDir);
			
		String strLastError = GetLastErrorString();
		if( !strLastError.IsEmpty() )
		{
			PrintText(L" (KO) ", tcActionFail);
			PrintText( strLastError + L"\n", tcErrorMsg);

			m_astrErrors.Clear();
		}
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////////
void POEngine::AddError(const String& str)
{
	m_astrErrors.Add(str);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
String POEngine::GetLastErrorString() const
{
	const int32 nErrCount = m_astrErrors.GetSize();
	if( nErrCount == 0 )
	{
		return String();
	}

	String str = m_astrErrors[0];
	for(int32 i = 1; i < nErrCount; ++i)
	{
		str = str + L"\n" + m_astrErrors[i];
	}
	return str;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//! Blits a block of pixels to a specific position.
///////////////////////////////////////////////////////////////////////////////////////////////////
void BasicBlit(int bytesPerPixel, const uint8* pSrc, int srcWidth, int srcHeight,
                                        uint8* pDst, int dstWidth, int /*dstHeight*/, int posX, int posY)
{
	// Blit at the right place
	pDst += ((posY * dstWidth) + posX) * bytesPerPixel;

	for(int y = 0; y < srcHeight; ++y)
	{
		for(int i = 0; i < srcWidth * bytesPerPixel; ++i)
		{
			pDst[i] = pSrc[i];
		}
		pSrc += srcWidth * bytesPerPixel;
		pDst += dstWidth * bytesPerPixel;
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Unpacks the pixels of each animation frame so one pixel is one byte.
///////////////////////////////////////////////////////////////////////////////////////////////////
void POEngine::UnpackPixelFrames(PngDumpSettings& ds)
{
	if( !ImageFormat::IsIndexed(ds.pixelFormat ) )
	{
		// Case not managed
		return;
	}
	if( ds.pixelFormat == pf8bppIndexed )
	{
		// Already ok
		return;
	}

	const int frameCount = ds.frames.GetSize();

	// Unpack the default image if any
	if( ds.hasDefaultImage || frameCount == 0 )
	{
		ImageFormat::UnpackPixels(ds.pixels, ds.width, ds.height, ds.pixelFormat);
	}

	for(int iFrame = 0; iFrame < frameCount; ++iFrame)
	{
		ApngFrame* pFrame = ds.frames[iFrame];

		const int width = pFrame->GetWidth();
		const int height = pFrame->GetHeight();
		ImageFormat::UnpackPixels(pFrame->m_pixels, width, height, ds.pixelFormat);
	}
	ds.pixelFormat = pf8bppIndexed;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Packs the pixels of each animation frame according to the pixel format.
// Example: if the pixel format is 8 bits but the palette contains 16 entries, the pixels can be
// repacked to 4 bits.
///////////////////////////////////////////////////////////////////////////////////////////////////
void POEngine::PackPixelFrames(PngDumpSettings& ds)
{
	if( ds.pixelFormat != pf8bppIndexed )
	{
		// Case not managed
		ASSERT(0);
		return;
	}
	if( ds.palette.m_count > 16 )
	{
		// Cannot pack
		return;
	}
	if( ds.palette.m_count > 4 )
	{
		ds.pixelFormat = pf4bppIndexed;
	}
	else if( ds.palette.m_count > 2 )
	{
		ds.pixelFormat = pf2bppIndexed;
	}
	else
	{
		ds.pixelFormat = pf1bppIndexed;
	}

	const int frameCount = ds.frames.GetSize();

	// Pack the default image if any
	if( ds.hasDefaultImage || frameCount == 0 )
	{
		ImageFormat::PackPixels(ds.pixels, ds.width, ds.height, ds.pixelFormat);
	}

	for(int iFrame = 0; iFrame < frameCount; ++iFrame)
	{
		ApngFrame* pFrame = ds.frames[iFrame];

		const int width = pFrame->GetWidth();
		const int height = pFrame->GetHeight();
		ImageFormat::PackPixels(pFrame->m_pixels, width, height, ds.pixelFormat);
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Fills the DumpSettings with all information from the image
void PrepareAnimatedDumpSettings(const ImageFormat& img, PngDumpSettings& ds)
{
	// Move image information to dump information, including animation frames
	ds.pixels = img.GetPixels();
	ds.width  = img.GetWidth();
	ds.height = img.GetHeight();
	ds.pixelFormat = img.GetPixelFormat();
	ds.palette    = img.GetPalette();
	ds.hasDefaultImage = img.HasDefaultImage();
	ds.loopCount   = img.GetLoopCount();

	ds.frames.Clear();

	const int frameCount = img.GetFrameCount();
	for(int i = 0; i < frameCount; ++i)
	{
		const AnimFrame* pSrc = img.GetAnimFrame(i);

		ApngFrame* pFrame = new ApngFrame(null);

		pFrame->m_fctl.width = pSrc->GetWidth();
		pFrame->m_fctl.height = pSrc->GetHeight();
		pFrame->m_fctl.offsetX = pSrc->GetOffsetX();
		pFrame->m_fctl.offsetY = pSrc->GetOffsetY();
		pFrame->m_fctl.blending = pSrc->GetBlending();
		pFrame->m_fctl.delayFracDenominator = (uint16) pSrc->GetDelayFracDenominator();
		pFrame->m_fctl.delayFracNumerator = (uint16) pSrc->GetDelayFracNumerator();
		pFrame->m_fctl.disposal = pSrc->GetDisposal();
		if( i == 0 )
		{
			pFrame->m_fctl.blending = AnimFrame::blendSource;
		}
		else
		{
			pFrame->m_fctl.blending = pSrc->GetBlending();
		}
		pFrame->m_pixels = pSrc->GetPixels();
		ds.frames.Add(pFrame);
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////////
struct FrameNewPal
{
	Palette           palette;
	PaletteTranslator translator;
};

///////////////////////////////////////////////////////////////////////////////////////////////////
// Merge all palettes in one single one, converts to true color if the merged palette exceeds 256 colors
void MergePalettes(const ImageFormat& img, PngDumpSettings& ds)
{
	if( ds.pixelFormat != pf8bppIndexed )
	{
		// Case not managed
		return;
	}

	Array<const Palette*> apPalettes;
	const int frameCount = img.GetFrameCount();
	for(int i = 0; i < frameCount; ++i)
	{
		const AnimFrame* pSrc = img.GetAnimFrame(i);
		if( img.IsIndexed() )
		{
			const Palette* pPal = &(pSrc->GetPalette());
			if( pPal->m_count > 0 )
			{
				apPalettes.AddUnique( pPal );
			}
		}
	}

	Palette superPalette; // In case a merging of local palettes is needed (GIF)

	bool switchToTrueColors = false;
	bool switchToRgba = false;

	// Check if we we should expand the first frame to be used as the default image
	bool expandFirstFrameAndMakeItTheDefaultImage = false;
	if( !ds.hasDefaultImage )
	{
		bool bOk1 = ds.width == ds.frames[0]->m_fctl.width;
		bool bOk2 = ds.height == ds.frames[0]->m_fctl.height;
		bool bOk3 = ds.frames[0]->m_fctl.offsetX == 0;
		bool bOk4 = ds.frames[0]->m_fctl.offsetY == 0;
		expandFirstFrameAndMakeItTheDefaultImage = !(bOk1 && bOk2 && bOk3 && bOk4);
	}
	else
	{
		// A default image is already provided, keep it
	}

	if( !expandFirstFrameAndMakeItTheDefaultImage && apPalettes.GetSize() <= 1 )
	{
		// No trick needed with the default image, one single palette, can stop here
		return;
	}


	// Build new local palettes and the super palette
	bool hasTransparency = false; // Switches to true if one of the color is transparent
	Array<FrameNewPal> frameNewPals;
	frameNewPals.SetSize(frameCount);

	for(int iFrame = 0; iFrame < frameCount; ++iFrame)
	{
		ApngFrame* pFrame = ds.frames[iFrame];

		uint32 colCounts[256];
		Memory::Zero32(colCounts, 256);
		AddFrameColorCounts(pFrame, colCounts);

		// Create the no-hole palette
		FrameNewPal fnp;
		fnp.palette = img.GetAnimFrame(iFrame)->GetPalette();
		fnp.translator.BuildUnusedColors(fnp.palette, colCounts);

		if( !hasTransparency && fnp.palette.HasNonOpaqueColor() )
		{
			hasTransparency = true;
		}

		if( !switchToTrueColors )
		{
			PaletteTranslator superTranslator;
			if( !superTranslator.MergePalette(superPalette, fnp.palette) )
			{
				// Too many colors, have to switch to RGB mode
				// Continue however to compute no holes palettes
				switchToTrueColors = true;
			}
			fnp.translator = PaletteTranslator::Combine(fnp.translator, superTranslator);
		}
		frameNewPals[iFrame] = fnp;
	}
		
	////////////////////////////////////////////////////////////////////
	// Choose between RGB and RGBA when switching to true colors
	if( switchToTrueColors && hasTransparency )
	{
		// Alpha is needed
		switchToRgba = true;
	}
	////////////////////////////////////////////////////////////////////

	////////////////////////////////////////////////////////////////////
	if( expandFirstFrameAndMakeItTheDefaultImage )
	{
		// Frame0 expansion is needed, we need to find or create a color to fill the blank.
		// This color has to be transparent.
		if( !switchToTrueColors )
		{
			// Palette mode, look in the super palette
			if( !hasTransparency )
			{
				// No transparent color to be used to create the default image. 
				// Try to add a new entry in the palette
				if( superPalette.m_count < 255 )
				{
					superPalette.m_colors[superPalette.m_count].SetRgba(0, 0, 0, 0);
					superPalette.m_count++;
				}
				else
				{
					// The palette is full, we have to switch to true colors with alpha
					switchToTrueColors = true;
					switchToRgba = true;
				}
			}
		}
		else
		{
			// Alpha is needed for transparency
			switchToRgba = true;
		}
	}
	////////////////////////////////////////////////////////////////////

	if( !switchToTrueColors )
	{
		// Use the super palette when saving
		ds.palette = superPalette;

		// Convert all frame indexes to match the super palette
		for(int iFrame = 0; iFrame < frameCount; ++iFrame)
		{
			ApngFrame* pFrame = ds.frames[iFrame];

			const FrameNewPal& fnp = frameNewPals[iFrame];

			int pixelCount = pFrame->m_fctl.width * pFrame->m_fctl.height;
			uint8* pPixels = pFrame->m_pixels.GetWritePtr();
			ASSERT(pFrame->m_pixels.GetSize() == pixelCount);
					
			fnp.translator.Translate(pPixels, pixelCount);
		}
	}
	else
	{
		// Convert to true-colors
		for(int iFrame = 0; iFrame < frameCount; ++iFrame)
		{
			const AnimFrame* pSrc = img.GetAnimFrame(iFrame);
			ApngFrame* pDst = ds.frames[iFrame];

			int dstPixelSize = switchToRgba ? 4 : 3;
			const int32 nPixCount = pDst->m_fctl.width * pDst->m_fctl.height;
			pDst->m_pixels.SetSize(nPixCount * dstPixelSize);
						
			const uint8* pSrcPixels = pSrc->GetPixels().GetReadPtr();
			uint8* pDstPixels = pDst->m_pixels.GetWritePtr();
			const Palette* pPalette = &(pSrc->GetPalette());

			if( switchToRgba )
			{
				// Palette to RGBA
				ds.pixelFormat = pf32bppRgba;
				for(int iPix = 0; iPix < nPixCount; ++iPix)
				{
					uint8 r, g, b, a;
					pPalette->m_colors[ pSrcPixels[iPix] ].ToRgba(r, g, b, a);
							
					pDstPixels[0] = r;
					pDstPixels[1] = g;
					pDstPixels[2] = b;
					pDstPixels[3] = a;
					pDstPixels += dstPixelSize;
				}
			}
			else
			{
				// Palette to RGB
				ds.pixelFormat = pf24bppRgb;
				for(int iPix = 0; iPix < nPixCount; ++iPix)
				{
					uint8 r, g, b, a;
					pPalette->m_colors[ pSrcPixels[iPix] ].ToRgba(r, g, b, a);
							
					pDstPixels[0] = r;
					pDstPixels[1] = g;
					pDstPixels[2] = b;
					pDstPixels += dstPixelSize;
				}
			}
		}
	}
	if( expandFirstFrameAndMakeItTheDefaultImage )
	{
		Buffer oldPixels = ds.frames[0]->m_pixels;
		ds.frames[0]->m_pixels.Clear();
		uint32 oldWidth = ds.frames[0]->m_fctl.width;
		uint32 oldHeight = ds.frames[0]->m_fctl.height;
		uint32 oldOffsetX = ds.frames[0]->m_fctl.offsetX;
		uint32 oldOffsetY = ds.frames[0]->m_fctl.offsetY;

		ds.frames[0]->m_fctl.width = ds.width;
		ds.frames[0]->m_fctl.height = ds.height;
		ds.frames[0]->m_fctl.offsetX = 0;
		ds.frames[0]->m_fctl.offsetY = 0;

		ds.frames[0]->m_pixels.SetSize(ds.height * ImageFormat::ComputeByteWidth(ds.pixelFormat, ds.width));

		const uint8* pSrc = oldPixels.GetReadPtr();
		uint8* pDst = ds.frames[0]->m_pixels.GetWritePtr();
			
		int bytesPerPixel = ImageFormat::SizeofPixelInBits(ds.pixelFormat) / 8;
			
		// Fill with transparent color
		if( bytesPerPixel == 1 )
		{
			// Use the transparent index
			int transparentColorIndex = superPalette.GetFirstFullyTransparentColor();
			ASSERT(transparentColorIndex >= 0 );
			ds.frames[0]->m_pixels.Fill(uint8(transparentColorIndex));
		}
		else
		{
			// Use transparent black in true colors
			ds.frames[0]->m_pixels.Fill(0);
		}

		BasicBlit(bytesPerPixel, pSrc, oldWidth, oldHeight, pDst, ds.width, ds.height, oldOffsetX, oldOffsetY);
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Converts a loaded animated GIF into an APNG file or convert an existing APNG.
// As GIF has no concept of default image, the IDAT is included in the animation.
//
// Returns true upon success.
///////////////////////////////////////////////////////////////////////////////////////////////////
bool POEngine::OptimizeAnimated(const ImageFormat& img, PngDumpSettings& ds, const String& filePath)
{
	if( !img.IsAnimated() )
	{
		return false;
	}
	
	// Fills the dump settings
	PrepareAnimatedDumpSettings(img, ds);
	
	// Unpack pixels to 8 bits per pixel if indexed to make treatment easier
	UnpackPixelFrames(ds);
	
	// Merge palettes if frame local palettes are found (GIF)
	// The pixel format can switch to RGB or RGBA
	MergePalettes(img, ds);
	
	ds.interlaced = false; // Always to false with animation
	ds.filtering = 0;
	ds.zlibCompressionLevel = 9;

	// If there is one frame only, use it as the main image and discard any animation information
	if( ds.pixels.IsEmpty() && ds.frames.GetSize() == 1 )
	{
		ApngFrame* pFrame = ds.frames[0];
		ds.width = pFrame->GetWidth();
		ds.height = pFrame->GetHeight();
		ds.pixels = pFrame->m_pixels;
		ds.frames.Clear();
	}

	bool optiOk = false;
	if( ds.pixelFormat == pf8bppIndexed )
	{
		// Continue with specialized optimizer
		optiOk = OptimizePaletteMode(ds);
	}
	else
	{
		// Other pixel formats
		optiOk = PerformDumpTries(ds);
	}

	bool dumpOk = false;
	if( optiOk )
	{
		dumpOk = DumpBestResultToFile(filePath);
	}
	m_resultmgr.Reset();
	return optiOk && dumpOk;
}
