/////////////////////////////////////////////////////////////////////////////////////
// This file is part of the POEngine library, part of the PngOptimizer application
// Copyright (C) 2002/2011 Hadrien Nilsson - psydk.org
// For conditions of distribution and use, see copyright notice in PngOptimizer.h
/////////////////////////////////////////////////////////////////////////////////////
#ifndef POENGINE_POENGINE_H
#define POENGINE_POENGINE_H

#include "../chustd/chustd.h"
using namespace chustd;

#include "POEngineSettings.h"

///////////////////////////////////////////////////////////////////////////////////////////////////
// PNG optimizing engine class
class POEngine
{
public:
	POEngineSettings m_settings;

public:
	// Optimizes several image files
	// astrFilePaths : array of absolute file paths
	// m_eventProgressing is fired on a regular basis with the text to be printed
	void OptimizeFiles(const StringArray& astrFilePaths);

	// Optimizes several image files
	// strDir : base directory name
	// astrFileNames : array of relative file paths
	// m_eventProgressing is fired on a regular basis with the text to be printed
	void OptimizeFiles(const String& strDir, const StringArray& astrFileNames);

	// Optimizes a buffer in memory
	// ds : source buffer description
	// strFileName : target file
	// Returns true upon success
	bool OptimizeExternalBuffer(PngDumpSettings& ds, const String& strNewFileName);

	// Optimizes an image file
	// This function ignores m_bBackupOldPngFiles
	// Return true upon success
	bool OptimizeFileNoBackup(const String& strFileName, const String& strNewFileName);

	// Gets error explanation when OptimizeFile fails
	String GetLastErrorString() const;

	POEngine();
	virtual ~POEngine();

public:
	// Color used when notifying progress
	enum TextColor
	{
		tcFilePath,      // black
		tcRegularInfo,   // gray
		tcActionVerb,    // gray :  Creating, Converting, Optimizing...
		tcSizeInfo,      // gray :  xxx bytes -> yyy bytes
		tcFileEnlarged,  // amber : (103% of the original size)
		tcFileSizeSame,  // gray :  (100% of the original size)
		tcFileReduced,   // green : (80% of the original size)
		tcActionOk,      // green : (OK)
		tcActionFail,    // red :   (KO)
		tcErrorMsg,      // red :   could not load image
		tcAnimated,      // blue :  [Animated GIF : ...]
		tcWelcomeMsg,    // blue :  Drop PNG, GIF, BMP or TGA files here
		tcLast
	};

	// Event fired during the optimization process
	Event2<const String&, enum TextColor> m_eventProgressing;

private:
	// This class allows the storing of the different tries done by POEngine
	// When dumping, we keep the best result of all
	class ResultManager
	{
	public:
		DynamicMemoryFile& GetCandidate(); // Gets the slot to test a new compression flavour
		DynamicMemoryFile& GetSmallest();  // Gets the best slot of all
		
		void Reset();

	private:
		DynamicMemoryFile m_dmf0;
		DynamicMemoryFile m_dmf1;
	};
	
	ResultManager m_resultmgr;
	
	// Last errors
	StringArray m_astrErrors;
	uint64 m_originalFileWriteTime;

private:
	void OptimizeFileOrDir(const String& strFilePath, const String& strDisplayDir);
	void OptimizeFiles(const String& strBaseDir, const StringArray& astrFileNames, const String& strDisplayDir);
	bool Optimize(PngDumpSettings& ds, const String& strNewFileName);
	
	// Those functions fill the DynamicMemoryFiles of m_resultmgr
	bool OptimizePaletteMode(PngDumpSettings& ds);
	bool Optimize24BitsMode(PngDumpSettings& ds);
	bool Optimize32BitsMode(PngDumpSettings& ds);
	bool OptimizeGrayScale(PngDumpSettings& ds);

	bool IsBlackAndWhite(const Palette& pal, bool& bShouldSwap);
	bool IsGreyPalette(const Palette& pal);
	bool TryToConvertIndexedToBlackAndWhite(PngDumpSettings& ds);
	bool FindUnusedColor(const Buffer& aRgb, uint8& nRed, uint8& nGreen, uint8& nBlue);
	bool FindUnusedColorHardcoreMethod(const uint8* pRgba, int32 nPixelCount, uint8& nRed, uint8& nGreen, uint8& nBlue);
	bool DumpBestResultToFile(const String& strNewFileName);

	bool InsertCleanOriginalPngAsResult(IFile& file);

	void AddError(const String& str);

	void PrintText(const String& strText, enum TextColor eColor);
	
	static void BgrToRgb(PngDumpSettings& ds);
	static void BgraToRgba(PngDumpSettings& ds);
	static bool Rgb16ToRgb24(PngDumpSettings& ds);

public:
	// public for unit testing
	bool OptimizeAnimated(const ImageFormat& img, PngDumpSettings& ds, const String& filePath);
private:
	bool PerformDumpTries(PngDumpSettings& ds);

	bool CustomPngDump(IFile& file, PngDumpSettings& ds);
	static void UnpackPixelFrames(PngDumpSettings& ds);
	static void PackPixelFrames(PngDumpSettings& ds);
};

#endif
