/////////////////////////////////////////////////////////////////////////////////////
// This file is part of the POEngine library, part of the PngOptimizer application
// Copyright (C) 2002/2011 Hadrien Nilsson - psydk.org
// For conditions of distribution and use, see copyright notice in PngOptimizer.h
/////////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "POEngineSettings.h"

using namespace chustd;

static const char k_szBackupOldPngFiles[]     = "BackupOldPngFiles";
static const char k_szKeepInterlacing[]       = "KeepInterlacing";
static const char k_szAvoidGreyWithSimpleTransparency[] = "AvoidGreyWithSimpleTransparency";
static const char k_szIgnoreAnimatedGifs[]    = "IgnoreAnimatedGifs";
static const char k_szKeepFileDate[]          = "KeepFileDate";

static const char k_szKeepBackgroundColor[]   = "KeepBackgroundColor";
static const char k_szForcedBackgroundColor[] = "ForcedBackgroundColor";

static const char k_szKeepTextualData[]       = "KeepTextualData";
static const char k_szForcedTextKeyword[]     = "ForcedTextKeyword";
static const char k_szForcedTextData[]        = "ForcedTextData";

///////////////////////////////////////////////////////////////////////////////////////////////////
POEngineSettings::POEngineSettings()
{
	backupOldPngFiles = true;
	keepInterlacing = false;
	avoidGreyWithSimpleTransparency = false;
	ignoreAnimatedGifs = false;
	keepFileDate = false;

	bkgdOption = POChunk_Remove;
	textOption = POChunk_Remove;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Load settings from an INI file. The section should be set.
///////////////////////////////////////////////////////////////////////////////////////////////////
void POEngineSettings::LoadFromIni(const MemIniFile& ini)
{
	ini.GetBool(k_szBackupOldPngFiles, backupOldPngFiles);
	ini.GetBool(k_szKeepInterlacing, keepInterlacing);
	ini.GetBool(k_szAvoidGreyWithSimpleTransparency, avoidGreyWithSimpleTransparency);
	ini.GetBool(k_szIgnoreAnimatedGifs, ignoreAnimatedGifs);
	ini.GetBool(k_szKeepFileDate, keepFileDate);

	ini.GetInt(k_szKeepBackgroundColor, (int&)bkgdOption);
	String strColBk;
	int32 nColBk = 0;
	ini.GetString(k_szForcedBackgroundColor, strColBk);
	if( strColBk.ToInt(nColBk, 'x') )
	{
		uint8 r = (nColBk >> 16) & 0x00ff;
		uint8 g = (nColBk >> 8) & 0x00ff;
		uint8 b = (nColBk) & 0x00ff;
		bkgdColor.SetRgb(r, g, b);
	}
	else
	{
		bkgdColor = Color32::black;
	}

	ini.GetInt(k_szKeepTextualData, (int&)textOption);
	ini.GetString(k_szForcedTextKeyword, textKeyword);
	ini.GetString(k_szForcedTextData, textData);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
void POEngineSettings::SaveToIni(MemIniFile& ini)
{
	ini.SetBool(k_szBackupOldPngFiles,  backupOldPngFiles);
	ini.SetBool(k_szKeepInterlacing,    keepInterlacing);
	ini.SetBool(k_szAvoidGreyWithSimpleTransparency, avoidGreyWithSimpleTransparency);
	ini.SetBool(k_szIgnoreAnimatedGifs, ignoreAnimatedGifs);
	ini.SetBool(k_szKeepFileDate,       keepFileDate);

	ini.SetInt(k_szKeepBackgroundColor, bkgdOption);
	uint8 r, g, b;
	bkgdColor.ToRgb(r, g, b);
	int32 nColBk = (r << 16) | (g <<8) | b;
	String strColBk = String::FromInt(nColBk, 'x', 6, '0').Right(6);
	ini.SetString(k_szForcedBackgroundColor, strColBk);

	ini.SetInt(k_szKeepTextualData, textOption);
	ini.SetString(k_szForcedTextKeyword, textKeyword);
	ini.SetString(k_szForcedTextData, textData);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
POChunkOption GetChunkOption(const ArgvParser& ap, const String& optionName)
{
	POChunkOption chunkOption = POChunk_Remove;
	if( ap.GetFlag(optionName) )
	{
		chunkOption = POChunk_Keep;
		
		String strKbcOption = ap.GetFlagString(optionName);
		if( strKbcOption == "0" || strKbcOption == "R" )
		{
			chunkOption = POChunk_Remove;
		}
		else if( strKbcOption == "2" || strKbcOption == "F" )
		{
			chunkOption = POChunk_Force;
		}
	}
	return chunkOption;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Load settings from a command line parsed with ArgvParser.
///////////////////////////////////////////////////////////////////////////////////////////////////
void POEngineSettings::LoadFromArgv(const ArgvParser& ap)
{
	backupOldPngFiles = ap.GetFlag(k_szBackupOldPngFiles);
	keepInterlacing = ap.GetFlag(k_szKeepInterlacing);
	avoidGreyWithSimpleTransparency = ap.GetFlag(k_szAvoidGreyWithSimpleTransparency);
	ignoreAnimatedGifs = ap.GetFlag(k_szIgnoreAnimatedGifs);
	keepFileDate = ap.GetFlag(k_szKeepFileDate);

	bkgdOption = GetChunkOption(ap, k_szKeepBackgroundColor);

	Color32 bkColor;
	if( ap.GetFlag(k_szForcedBackgroundColor) )
	{
		String strBkColor = ap.GetFlagString(k_szForcedBackgroundColor);
		if( strBkColor.GetLength() == 6 )
		{
			int bkR = 0;
			int bkG = 0;
			int bkB = 0;
			strBkColor.Mid(0, 2).ToInt(bkR, 'x');
			strBkColor.Mid(2, 2).ToInt(bkG, 'x');
			strBkColor.Mid(4, 2).ToInt(bkB, 'x');
			bkgdColor.SetRgb(uint8(bkR), uint8(bkG), uint8(bkB));
		}
	}
	
	textOption = GetChunkOption(ap, k_szKeepTextualData);
	textKeyword = ap.GetFlagString(k_szForcedTextKeyword);
	textData = ap.GetFlagString(k_szForcedTextData);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Writes usage when parsing from argv.
///////////////////////////////////////////////////////////////////////////////////////////////////
void POEngineSettings::WriteArgvUsage(const String& indent)
{
	Console::WriteLine(indent + "[-" + String(k_szBackupOldPngFiles) + "]");
	Console::WriteLine(indent + "[-" + String(k_szKeepInterlacing) + "]");
	Console::WriteLine(indent + "[-" + String(k_szAvoidGreyWithSimpleTransparency) + "]");
	Console::WriteLine(indent + "[-" + String(k_szIgnoreAnimatedGifs) + "]");
	Console::WriteLine(indent + "[-" + String(k_szKeepFileDate) + "]");
	Console::WriteLine(indent + "[-" + String(k_szKeepBackgroundColor) + "][:R|K|F] [-" + String(k_szForcedBackgroundColor) + ":RRGGBB]");
	Console::WriteLine(indent + "[-" + String(k_szKeepTextualData) + "][:R|K|F]     [-" + String(k_szForcedTextKeyword) + ":Foo] [-" 
	                                                                                    + String(k_szForcedTextData) + ":Bar]");
}
