///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2011 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////


#include "stdafx.h"
#include "ArgvParser.h"
#include "StringBuilder.h"

namespace chustd {

ArgvParser::ArgvParser(const String& strCmdLine) : m_strCmdLine(strCmdLine)
{
	Parse();
}

ArgvParser::~ArgvParser()
{

}

const wchar* ArgvParser::JumpOverExeName(const wchar* pszCmdLine)
{
	const int nLength = String::SZLength(pszCmdLine);
	for(int i = 0; i < nLength; ++i)
	{
		wchar c = pszCmdLine[i];
		if( i == 0 && c != '"')
		{
			// No double quote found
			break;
		}

		if( i > 0 && c == '"')
		{
			// Closing double quote found
			return pszCmdLine + i + 1;
		}
	}
	return pszCmdLine;
}


int32 ArgvParser::GetFlagIndex(const String& strFlagName) const
{
	String strWanted = strFlagName.ToLowerCase();

	const int32 nArgCount = m_aArgs.GetSize();
	for(int32 i = 0; i < nArgCount; ++i)
	{
		if( m_aArgs[i].m_bFlag)
		{
			String strName = m_aArgs[i].m_strName.ToLowerCase();

			if( strName == strWanted)
				return i;
		}
	}

	return -1;
}

bool ArgvParser::GetFlag(const String& strFlagName) const
{
	if( GetFlagIndex(strFlagName) < 0)
	{
		return false;
	}
	return true;
}

String ArgvParser::GetFlagString(const String& strFlagName) const
{
	const int32 nIndex = GetFlagIndex(strFlagName);
	if( nIndex < 0)
	{
		return String();
	}

	return m_aArgs[nIndex].m_strFlagString;
}

void ArgvParser::Parse()
{
	const int32 length = m_strCmdLine.GetLength();
	const wchar* psz = m_strCmdLine.GetBuffer();
	
	JumpOverExeName(psz);

	StringArray astrArgs;
	FirstPassSplit(psz, length, astrArgs);

	ConvertToArgs(astrArgs);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Build a string array by splitting the arguments according to space and double quotes
void ArgvParser::FirstPassSplit(const wchar* psz, int length, StringArray& astrArgs)
{
	bool bInsideArg = false;
	bool bInsideDQ = false;

	StringBuilder sb;

	for(int i = 0; i < length; ++i)
	{
		wchar c = psz[i];
		if( !bInsideArg)
		{
			if( c != ' ' && c != '\t')
			{
				if( c == '"')
				{
					if( bInsideDQ)
					{
						bInsideDQ = false;
					}
					else
					{
						bInsideDQ = true;
					}
				}

				// New arguments
				sb.Empty();
				sb += c;
				bInsideArg = true;
			}
		}
		else
		{
			// We are inside an argument
			if( c != ' ' && c != '\t')
			{
				if( c == '"')
				{
					if( bInsideDQ)
					{
						bInsideDQ = false;
					}
					else
					{
						bInsideDQ = true;
					}
				}

				sb += c;
			}
			else
			{
				// A separator, on insert it in the argument
				// if we are inside double quotes
				if( bInsideDQ)
				{
					sb += c;
				}
				else
				{
					// End of argument
					astrArgs.Add(sb.ToString());
					bInsideArg = false;
				}
			}
		}
	}
	
	// Add the last string
	astrArgs.Add(sb.ToString());
}

void ArgvParser::ConvertToArgs(const StringArray& astrArgs)
{
	const int32 nArgCount = astrArgs.GetSize();
	for(int32 iArg = 0; iArg < nArgCount; ++iArg)
	{
		const String& strArg = astrArgs[iArg];
		const int32 nArgLength = strArg.GetLength();

		if( nArgLength > 0)
		{
			int32 nIndex = m_aArgs.Add();
			Arg& arg = m_aArgs[nIndex];

			wchar cFirst = strArg.GetAt(0);
			if( cFirst == '/' || cFirst == '-')
			{
				StringBuilder sbName;
				StringBuilder sbFlagString;

				// We've got an argument
				arg.m_bFlag = true;

				// Cut the argument
				enum SearchMode { searchFlagName, searchPreFlagString, searchFlagStringNoDQ, searchFlagStringDQ };
				SearchMode searchMode = searchFlagName;

				for(int i = 1; i < nArgLength; ++i)
				{
					wchar c = strArg.GetAt(i);

					if( searchMode == searchFlagName)
					{
						if( c == ':')
						{
							searchMode = searchPreFlagString;
						}
						else if( c == '"')
						{
							searchMode = searchFlagStringDQ;
						}
						else
						{
							sbName += c;
						}
					}
					else
					{
						if( searchMode == searchFlagStringNoDQ)
						{
							sbFlagString += c;
						}
						else if( searchMode == searchFlagStringDQ)
						{
							if( c == '"')
							{
								// We reached the end of this argument
								break;
							}
							else
							{
								sbFlagString += c;
							}
						}
						else if( searchMode == searchPreFlagString)
						{
							if( c == '"')
							{
								searchMode = searchFlagStringDQ;
							}
							else
							{
								// The flag string is not in double-quotes
								sbFlagString += c;
								searchMode = searchFlagStringNoDQ;
							}
						}
					}
				}
				arg.m_strName = sbName.ToString();
				arg.m_strFlagString = sbFlagString.ToString();
			}
			else
			{
				arg.m_bFlag = false;
				arg.m_strName = strArg;
			}
		}
	}
}

/////////////////////
} // namespace chustd
