///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2011 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ChunkedFile.h"
#include "IFile.h"

using namespace chustd;
//////////////////////////////////////////////////////////////////////
ChunkedFile::ChunkedFile(IFile& file)
{
	m_pFile = &file;
	
	m_nChunk = 0;
	m_nSizeof = 0;
	m_nCrc = 0;

	m_nStartAt = -1;
	m_bInsideChunk = false;
}

ChunkedFile::~ChunkedFile()
{
	if( m_bInsideChunk )
	{
		EndChunkWrite();
	}
}

//////////////////////////////////////////////////////////////////////
// Write Chunk : Begin
bool ChunkedFile::BeginChunkWrite(uint32 chunkName)
{
	// We cannot have a chunk inside another chunk
	if( m_bInsideChunk )
	{
		// A call to EndChunk is missing
		return false;
	}
	
	// Checker flag
	m_bInsideChunk = true;

	// Get the chunk position in the stream
	m_nStartAt = GetPosition();

	// At this very moment the sizeof is not valid
	const uint32 chunkSizeof = 'szof';
	if( !Write32(chunkSizeof) )
		return false;

	// Crc for this new chunk, updated for each call to Write
	// The CRC does not includes the  sizeof  field, and anyway we could not include this sizeof field
	// now as we don't know its value yet. Thank you the PNG standard.
	InitCrc(m_nCrc);

	if( !Write32(chunkName) )
		return false;

	return true;
}

// Write Chunk : End
bool ChunkedFile::EndChunkWrite()
{
	// BeginChunk should be called before calling this function
	if( !m_bInsideChunk )
	{
		// A call to BeginChunk is missing
		return false;
	}

	///////////////////////////////////////////////////////////////
	// Chunk sizeof computation

	const int64 curPos = GetPosition();
	
	// The sizeof dos not include the chunkname nor the sizeof nor the crc
	// The crc has not been written yet, so - 8 instead of - 12
	const uint32 chunkSizeof = uint32( (curPos - m_nStartAt) - 8);
	///////////////////////////////////////////////////////////////

	// CRC ending computation
	FinalizeCrc(m_nCrc);

	// Write CRC
	if( !Write32(m_nCrc) )
	{
		return false;
	}

	// Go to the position where the sizeof must be written
	SetPosition(m_nStartAt);
			
	// Write sizeof
	if( !Write32(chunkSizeof) )
	{
		return false;
	}

	// Jump at the end
	SetPosition(curPos + 4);
	
	// Checker flag
	m_bInsideChunk = false;
	return true;
}

bool ChunkedFile::BeginChunkRead()
{
	m_nStartAt = (uint32) GetPosition();
	
	m_nSizeof = 0;
	m_nChunk = 0;

	if( !Read32(m_nSizeof) )
		return false;

	if( m_nSizeof < 0 )
		return false;

	if( !Read32(m_nChunk) )
		return false;

	return true;
}

bool ChunkedFile::EndChunkRead()
{
	// BeginChunkRead should be called before calling this function
	if( m_nStartAt < 0 )
	{
		ASSERT(0); // This is to help the developer
		return false;
	}

	// Read the CRC of the chunk
	uint32 chunkCrc = 0;
	if( !Read32(chunkCrc) )
	{
		return false;
	}

	// Get the number of bytes read so far
	const uint32 read = uint32( GetPosition() - m_nStartAt);
	
	// m_nSizeof does not includes the chunkname nor the sizeof nor the crc
	const uint32 chunkSizeof = m_nSizeof + 12;

	if( read > chunkSizeof )
	{
		// The application has read too much data
		return false;
	}
	
	if( read < chunkSizeof )
	{
		// The application is allowed to read less bytes than the number contained in the chunk
		SetPosition(m_nStartAt + chunkSizeof);
	}
	
	// Reset begin chunk info
	m_nSizeof = 0;
	m_nChunk = 0;

	// Update end chunk info
	m_nCrc = chunkCrc;

	return true;
}

int32 ChunkedFile::GetRemainingLength() const
{
	const uint32 read = uint32( GetPosition() - m_nStartAt);
	// - 8 so we do not include the SIZEOF and NAME
	const int32 remains = (m_nSizeof - (read - 8));

	return remains;
}

// Compute a CRC from a byte buffer
void ChunkedFile::InitCrc(uint32& crc)
{
	crc = 0xffffffff;
}

void ChunkedFile::FinalizeCrc(uint32& crc)
{
	crc ^= 0xffffffff;
}

void ChunkedFile::UpdateCrc(uint32& pendingCrc, const void* pBuffer, int32 bufferSize)
{
	// CRC look-up table. Computed only once
	static uint32 aCrcLUT[256];
	static bool bCrcLUTComputed = false;

	// CRC look-up table computation
	if( !bCrcLUTComputed )
	{
		bCrcLUTComputed = true;
		for( int32 n = 0; n < 256; ++n)
		{
			uint32 c = n;
			for( int32 k = 0; k < 8; ++k)
			{
				if( c & 1 )
					c = 0xEDB88320 ^ (c >> 1);
				else
					c = c >> 1;
			}
			aCrcLUT[n] = c;	
		}
	}

	// CRC computation
	uint32 tmpCrc = pendingCrc;
	const uint8* pByteBuffer = (uint8*) pBuffer;

	for( int32 i = 0; i < bufferSize; ++i)
	{
		const uint8 byte = pByteBuffer[i];
		const int32 nLUTIndex = (tmpCrc ^ byte) & 0x000000ff;
		tmpCrc = aCrcLUT[ nLUTIndex ] ^ (tmpCrc >> 8);
	}
	pendingCrc = tmpCrc;
}

bool ChunkedFile::SetPosition(int64 offset, Whence eWhence)
{
	return m_pFile->SetPosition(offset, eWhence);
}

int64 ChunkedFile::GetPosition() const
{
	return m_pFile->GetPosition();
}

int64 ChunkedFile::GetSize()
{
	return m_pFile->GetSize();
}

int32 ChunkedFile::Read(void* pBuffer, int size)
{
	return m_pFile->Read(pBuffer, size);
}

int32 ChunkedFile::Write(const void* pBuffer, int size)
{
	UpdateCrc(m_nCrc, pBuffer, size);
	return m_pFile->Write(pBuffer, size);
}

ByteOrder ChunkedFile::GetByteOrder() const
{
	return m_pFile->GetByteOrder();
}

void ChunkedFile::SetByteOrder(ByteOrder eByteOrder)
{
	m_pFile->SetByteOrder(eByteOrder);
}
