///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2011 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////
#ifndef CHUSTD_DATETIME_H
#define CHUSTD_DATETIME_H

namespace chustd {

// Manages a duration between two dates
class Duration
{
	friend class DateTime;
public:
	
	void Split(int32& days, int8& hours, int8& minutes, int8& seconds, int16& milliseconds) const;
	int64 GetStampInSec() const;

	Duration();
	Duration(int64 chuTimeStamp);
	Duration(int32 days, int8 hours, int8 minutes, int8 seconds, int16 milliseconds);

	void FormatFrench(class String& str) const;
	void FormatSmall(class String& str) const;

	static Duration FromSeconds(int64 seconds);

protected:
	int64 m_nStamp;
};

// Used by the DateTime::Format method
enum DateFormat
{
	dfExpanded,  // YYYY-MM-DD hh:mm:ss
	dfExpandedT, // YYYY-MM-DDThh:mm:ss
	dfCompact    // YYYYMMDDThhmmss
};

// Manages a date+time
class DateTime
{
public:
	uint16 GetYear() const { return m_nYear; }
	uint8 GetMonth() const { return m_nMonth; }
	uint8 GetDay() const { return m_nDay; }
	uint8 GetHour() const { return m_nHour; }
	uint8 GetMinute() const { return m_nMinute; }
	uint8 GetSecond() const { return m_nSecond; }
	
	int64 GetStampInSec() const;
	int64 GetChuTimeStamp() const;

	void AddDays(int count);
	void SetTime(uint8 hour, uint8 minute, uint8 second);

	static DateTime FromChuTimeStamp(int64 stamp);

public:
	const Duration operator - (const DateTime& t2) const;
	bool operator > (const DateTime& t2) const;
	bool operator < (const DateTime& t2) const;
	bool operator >= (const DateTime& t2) const;
	bool operator <= (const DateTime& t2) const;
	
	const DateTime& operator += (Duration dur);

	static DateTime GetNow();
	
	static DateTime FromUnixTimeStamp(int64 time);
	int64 ToUnixTimeStamp() const;
	
	// Formats date and hour as YYYY-MM-DD hh:mm:ss (ISO 8601)
	String Format(DateFormat eFormat = dfExpanded) const;

	// Formats date only as YYYY-MM-DD
	String FormatDate(DateFormat eFormat = dfExpanded) const;

	// Formats time only as hh:mm:ss
	String FormatTime(DateFormat eFormat = dfExpanded) const;

	// Gets date as YYYY-MM-DD hh:mm:ss
	// Works if last pars are missing, for instance : YYYY-MM 
	// Returns the number of read members
	int32 Scan(const String& strDate);

	void Set(int16 year, uint8 month = 1, uint8 day = 1, uint8 hour = 0, uint8 minute = 0, uint8 second = 0);

	DateTime();

protected:
	int64 m_nStamp;

	uint16 m_nYear;
	uint8 m_nMonth;
	uint8 m_nDay;

	uint8 m_nHour;
	uint8 m_nMinute;
	uint8 m_nSecond;

	uint8 m_nDummy;

protected:
	// Updates m_nYear, m_nMonth, m_nDay, m_nHour, m_nMinute, m_nSecond
	void UpdateFieldsFromStamp();

	// Updates m_nStamp
	void UpdateStampFromFields();
};

} // namespace chustd

#endif // ndef CHUSTD_DATE_H
