///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2011 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "File.h"
#include "StringBuilder.h"
#include "TextEncoding.h"

#ifdef __APPLE__
#include <CoreFoundation/CoreFoundation.h>
#endif


namespace chustd {
	class FileImpl
	{
	public:
#ifdef CHUOS
		IFileSystem::Handle m_handle;
#else
		void* m_handle;
#endif
		uint32 m_nOpenMode;

	public:
		FileImpl()
		{
			m_handle = 0;
		}
	};
}

//////////////////////////////////////////////////////////////////////
using namespace chustd;
//////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////
// Mac OS X uses UTF-8 for POSIX, we need this class to convert 
// wchar to UTF-8 in order to use plain old fopen() with wchar
#ifdef __APPLE__
class Utf8
	{
	public:
		Utf8(const String& s): m_utf8(NULL)
		{
			const wchar_t* wsz = s.GetBuffer();
			const int bytes = wcslen(wsz) * sizeof(wchar_t);
			// OS X uses 32-bit wchar
			CFStringEncoding encoding = ( k_ePlatformByteOrder == boLittleEndian ) ? kCFStringEncodingUTF32LE 
			                                                                       : kCFStringEncodingUTF32BE;
			CFStringRef str = CFStringCreateWithBytesNoCopy(NULL, (const UInt8*)wsz, bytes, encoding, false, kCFAllocatorNull);
			const int bytesUtf8 = CFStringGetMaximumSizeOfFileSystemRepresentation(str);
			m_utf8 = new char[bytesUtf8];
			CFStringGetFileSystemRepresentation(str, m_utf8, bytesUtf8);
			CFRelease(str);
		}	
		
		~Utf8() 
		{ 
			if( m_utf8 )
			{
				delete[] m_utf8;
			}
		}
		
	public:
		operator const char*() const { return m_utf8; }
		
	private:
		char* m_utf8;
	};
#endif


File::File()
{
	m_pImpl = new FileImpl;
}

File::~File()
{
	Close();
	delete m_pImpl;
}

bool File::Open(const String& strFilePath, uint32 mode/*=modeRead*/)
{
	if( strFilePath.IsEmpty() )
		return false;

#ifdef WIN32
	int access = 0;
	int disposition = 0;
	int shareMode = FILE_SHARE_READ;

	bool bOpenRead = (mode & modeRead) != 0;
	bool bOpenWrite = (mode & modeWrite) != 0;
	bool bAppend = (mode & modeAppend) != 0;

	if( !bAppend )
	{
		if( bOpenRead && !bOpenWrite )
		{
			// Read only on an existing file
			access = GENERIC_READ;
			disposition = OPEN_EXISTING;
		}
		else if( !bOpenRead && bOpenWrite )
		{
			// Write only on a unexisting or existing file
			access = GENERIC_WRITE;
			disposition = CREATE_ALWAYS;
		}
		else if( bOpenRead && bOpenWrite )
		{
			// If the file exists, open it and set at the start of it
			// If the file does not exist, create it
			access = GENERIC_READ | GENERIC_WRITE;
			disposition = OPEN_ALWAYS;
		}
		else
		{
			// Bad open mode
			return false;
		}
	}
	else
	{
		// Note : the append mode in chustd has a different behaviour in comparison with the Libc:
		// A call to SetPosition followed by a call to Write, writes data at the position specified by SetPosition.
		// The Libc seeks to the end of the file before writing.

		// In append mode, the file is implicitely open in write mode
		mode |= modeWrite;
		access = GENERIC_WRITE;

		if( bOpenRead )
		{
			access |= GENERIC_READ;
		}
		disposition = OPEN_ALWAYS;
	}

	HANDLE handle = CreateFileW(strFilePath.GetBuffer(), 
		access, shareMode, NULL, disposition, FILE_ATTRIBUTE_NORMAL, NULL);
	if( handle == INVALID_HANDLE_VALUE )
	{
		return false;
	}

	if( bAppend )
	{
		DWORD ret = SetFilePointer(handle, 0, 0, FILE_END);
		ret;
	}
	
#elif CHUOS
	/*
	IDrive* pDrive = Kernel::GetDriveFromPath(strFilePath);
	String strAbsoluteFilePath = pDrive->GetAbsolutePath(strFilePath);
	*/
	IFileSystem* pFileSystem = null;
	IFileSystem::Handle handle = pFileSystem->OpenFile(strFilePath, mode);
	
#elif __APPLE__
	void* handle = NULL;
	
	switch( mode & (modeRead | modeWrite | modeAppend) )
	{
	case modeRead:
		handle = fopen(Utf8(strFilePath), "rb");
		break;
	case modeWrite:
		handle = fopen(Utf8(strFilePath), "wb");
		break;
	case modeAppend:
		handle = fopen(Utf8(strFilePath), "ab");
		break;
	case modeReadWrite:
		handle = fopen(Utf8(strFilePath), "r+b");
		if( handle == null )
		{
			// Could not open file, it doesn't exist
			handle = fopen(Utf8(strFilePath), "w+b");
		}
		break;
	default:
		ASSERT(0);
		break;
	}
#endif

	m_pImpl->m_handle = handle;
	m_pImpl->m_nOpenMode = mode;
	return true;
}

void File::Close()
{
	if( !m_pImpl )
		return;

	if( m_pImpl->m_handle== null )
		return;

#ifdef WIN32
	CloseHandle(m_pImpl->m_handle);
#elif CHUOS
	IFileSystem* pFileSystem = null;
	pFileSystem->CloseFile(m_pImpl->m_handle);
#endif

	m_pImpl->m_handle = 0;
}

ByteOrder File::GetByteOrder() const
{
	if( !m_pImpl )
	{
		return boBigEndian;
	}

	if( (m_pImpl->m_nOpenMode & modeLittleEndian) != 0 )
	{
		return boLittleEndian;
	}
	return boBigEndian;
}

void File::SetByteOrder(ByteOrder eByteOrder)
{
	if( !m_pImpl )
		return;

	if( eByteOrder == boLittleEndian )
	{
		m_pImpl->m_nOpenMode |= modeLittleEndian;
	}
	else if( eByteOrder == boBigEndian )
	{
		m_pImpl->m_nOpenMode &= !modeLittleEndian;
	}
	else
	{
		ASSERT(0);	
	}
}


int32 File::Read(void* pBuffer, int size)
{
	if( !m_pImpl )
		return -1;

	if( m_pImpl->m_handle == null )
		return -1;

	if( size < 0 )
		return -2;

	if( size == 0 )
	{
		return 0; // Not an error
	}
#ifdef WIN32
	DWORD read = 0;
	BOOL bRet = ReadFile(m_pImpl->m_handle, pBuffer, size, &read, NULL);
	if( !bRet )
	{
		return -3;
	}
#elif CHUOS
	IFileSystem* pFileSystem = null;
	int32 read = pFileSystem->ReadFile(m_pImpl->m_handle, pBuffer, size);
#else
	int32 read = fread(pBuffer, size, size, (FILE*)m_pImpl->m_handle);
#endif
	return int32(read);
}

int32 File::Write(const void* pBuffer, int size)
{
	if( !m_pImpl )
		return -1;

	if( m_pImpl->m_handle == null )
		return -1;

	if( size <= 0 )
		return -2;

#ifdef WIN32
	DWORD written = 0;
	BOOL bRet = WriteFile(m_pImpl->m_handle, pBuffer, size, &written, NULL);
	if( !bRet )
	{
		return -3;
	}
#elif CHUOS
	IFileSystem* pFileSystem = null;
	int32 written = pFileSystem->WriteFile(m_pImpl->m_handle, pBuffer, size);
#else
	int32 written = fwrite(pBuffer, size, size, (FILE*)m_pImpl->m_handle);
#endif

	return int32(written);
}

bool File::SetPosition(int64 offset, Whence eWhence)
{
	if( !m_pImpl )
		return false;

	if( m_pImpl->m_handle == null )
		return false;

#ifdef WIN32
	int32 nCFrom = 0;
	switch(eWhence)
	{
	case posBegin:
		nCFrom = FILE_BEGIN;
		break;
	case posCurrent:
		nCFrom = FILE_CURRENT;
		break;
	case posEnd:
		nCFrom = FILE_END;
		break;
	default:
		ASSERT(0); // Bad value for seek
		break;
	}
	
	LONG high = LONG(offset >> 32);
	DWORD low = DWORD(offset & 0x00000000ffffffff);
	DWORD ret = SetFilePointer(m_pImpl->m_handle, low, &high, nCFrom);

	DWORD error = 0;
	if( ret == INVALID_SET_FILE_POINTER && (error = GetLastError()) != NO_ERROR )
	{
		return false;
	}
	return true;

#elif CHUOS
	IFileSystem* pFileSystem = null;
	pFileSystem->SetFilePosition(m_pImpl->m_handle, offset, eWhence);
	return true;
	
#else
	int whence = 0;
	switch(eWhence)
	{
	case posBegin:   whence = SEEK_SET; break;
	case posCurrent: whence = SEEK_CUR; break;
	case posEnd:     whence = SEEK_END; break;
	default:
		ASSERT(0); // Bad value for seek
		break;
	}
	
	const int ret = fseeko((FILE*)m_pImpl->m_handle, offset, whence);
	return ( ret == 0 );
#endif
}

int64 File::GetPosition() const
{
	if( !m_pImpl )
		return 0;

	if( m_pImpl->m_handle == null )
		return 0;

#ifdef WIN32
	LONG high = 0;
	DWORD ret = SetFilePointer(m_pImpl->m_handle, 0, &high, FILE_CURRENT);

	DWORD error = 0;
	if( ret == INVALID_SET_FILE_POINTER && (error = GetLastError()) != NO_ERROR )
	{
		return 0;
	}

	int64 position = high;
	position <<= 32;
	position |= ret;

	return position;

#elif CHUOS
	IFileSystem* pFileSystem = null;
	return pFileSystem->GetFilePosition(m_pImpl->m_handle);
	
#else
	return ftello((FILE*)m_pImpl->m_handle);
#endif
}

int64 File::GetSize()
{
	if( !m_pImpl )
		return 0;

	if( !m_pImpl->m_handle )
		return 0;

#ifdef WIN32
	DWORD sizeHigh = 0;
	DWORD sizeLow = GetFileSize(m_pImpl->m_handle, &sizeHigh);

	int64 length = sizeHigh;
	length <<= 32;
	length |= sizeLow;

	return length;

#elif CHUOS
	IFileSystem* pFileSystem = null;
	return pFileSystem->GetFileSize(m_pImpl->m_handle);
	
#else
	SetPosition(0, posEnd);
	return GetPosition();
#endif
}

//////////////////////////////////////////////////////////////////////////
bool File::Rename(const String& strOldName, const String& strNewName)
{
#ifdef WIN32
	BOOL bRet = MoveFileW(strOldName.GetBuffer(), strNewName.GetBuffer());
	return bRet != FALSE;
#elif CHUOS
	// Get the file names alone
	String strDir, strDir2;
	String strOldName2, strNewName2;

	File::SplitPath(strOldName, strDir, strOldName2);
	File::SplitPath(strNewName, strDir2, strNewName2);

	IFileSystem* pFileSystem = null;
	return pFileSystem->RenameFile(strDir, strOldName2, strNewName2);
	
#elif __APPLE__
	return rename(Utf8(strOldName), Utf8(strNewName));
	
#endif
}

String File::GetDrive(const String& strPath)
{
	const int32 length = strPath.GetLength();
	if( length == 0 )
	{
		return String();
	}
	
	uint16 c = 0;

	int32 iSeparator = 0;
	for(; iSeparator < length; ++iSeparator)
	{
		c = strPath.GetAt(iSeparator);
		if( c == '\\' || c == '/' || c == ':' )
		{
			break;
		}
	}

	if( iSeparator < 0 )
	{
		// No separator found
		return String();
	}

	if( c == ':' )
	{
		// Drive found
		return strPath.Left(iSeparator);
	}
	
	return String();
}

// Cuts string to parent's directory
// Does not leave the final separator
String File::CutToParent(const String& strPath)
{
	int32 length = strPath.GetLength();
	
	const wchar* pszPath = strPath.GetBuffer();

	// Is last character a separator ?
	wchar c = pszPath[length - 1];
	if( c == L':' )
		return strPath; // A partition, do not modify the string

	if( c == L'\\' || c == L'/' )
		length--; // Yep, do not count the separator

	int32 i = length - 1;
	for(;i >= 0; i--)
	{
		if( pszPath[i] == L':' || pszPath[i] == L'\\' || pszPath[i] == L'/' )
			break;
	}
	
	return strPath.Left(i + 1);
}

// Returns true if a path name is absolute

// plop/onk.txt returns false
// ../plop/onk.txt returns false

// file://plop/onk.txt returns true
// d:/plop/onk.txt returns true
// d:onk.txt returns true
// /plop/onk.txt returns true
bool File::IsAbsolutePath(const String& strPath)
{
	const int32 length = strPath.GetLength();
	if( length == 0 )
		return false;

	const wchar* pszPath = strPath.GetBuffer();

	bool bSeparatorFound = false;
	int32 i = 0;
	for(; i < length; i++)
	{
		if( pszPath[i] == '/'
			|| pszPath[i] == '\\'
			|| pszPath[i] == ':')
		{
			bSeparatorFound = true;
			break;
		}
	}
	
	if( !bSeparatorFound )
		return false;

	if( pszPath[i] == ':' )
		return true;

	if( pszPath[0] == '/' || pszPath[0] == '\\' )
		return true;

	return false;
}

String File::GetExtension(const String& strFilePath)
{
	const int32 length = strFilePath.GetLength();
	String strExt;

	for(int i = length - 1; i >= 0; i--)
	{
		wchar c = strFilePath.GetAt(i);
		if( c == L'.' )
		{
			const int count = length - 1 - i;
			strExt = strFilePath.Right(count);
			break;
		}
	}
	return strExt;
}

// Non static
// 2 Go Max
ByteArray File::GetContent()
{
	const int64 currentPosition = GetPosition();
	int64 fileLength = GetSize();
	
	if( fileLength > 0x000000007fffffff )
	{
		fileLength = 0x000000007fffffff;
	}
	
	ByteArray aContent;
	aContent.SetSize( int32(fileLength));

	Read(aContent.GetPtr(), int32(fileLength));
	
	SetPosition(currentPosition);
	
	aContent.SetAsReturnValue();
	return aContent;
}

// Static
ByteArray File::GetContent(const String& strFilePath)
{
	File file;
	if( !file.Open(strFilePath, File::modeRead) )
	{
		return ByteArray();
	}

	ByteArray aContent = file.GetContent();

	aContent.SetAsReturnValue();
	return aContent;
}

// Non static
String File::GetTextContent(const TextEncoding& te)
{
	ByteArray aContent = GetContent();
	return String::FromBytes(aContent, te);
}

// Static
String File::GetTextContent(const String& strFilePath, const TextEncoding& te)
{
	ByteArray aContent = GetContent(strFilePath);
	return String::FromBytes(aContent, te);
}

// Non static
StringArray File::GetLines(const TextEncoding& te)
{
	String strContent = GetTextContent(te);
	StringArray astrLines = strContent.SplitByEndlines();
	
	astrLines.SetAsReturnValue();
	return astrLines;
}

// Static
StringArray File::GetLines(const String& strFilePath, const TextEncoding& te)
{
	File file;
	if( !file.Open(strFilePath, File::modeRead) )
	{
		return StringArray();
	}

	StringArray astrLines = file.GetLines(te);

	astrLines.SetAsReturnValue();
	return astrLines;
}


bool File::Exists(const String& strPath)
{
	bool bDirectory, bReadOnly;
	return GetFileAttributes(strPath, bDirectory, bReadOnly) && !bDirectory;
}

int64 File::GetSize(const String& strPath)
{
	File file;
	if( !file.Open(strPath) )
		return 0;

	return file.GetSize();
}

uint64 File::GetLastWriteTime()
{
	uint64 lastWriteTime = 0;
#ifdef WIN32
	FILETIME timeLastWrite = {0};
	BOOL bGetTime = ::GetFileTime(m_pImpl->m_handle, NULL, NULL, &timeLastWrite);
	if( bGetTime )
	{
		lastWriteTime = *((uint64*)&timeLastWrite);
	}
#endif
	return lastWriteTime;
}

bool File::SetLastWriteTime(uint64 time)
{
	if( m_pImpl == NULL || m_pImpl->m_handle == NULL )
	{
		return false;
	}
#ifdef WIN32
	FILETIME timeLastWrite = *((FILETIME*)&time);
	if( !::SetFileTime(m_pImpl->m_handle, NULL, NULL, &timeLastWrite) )
	{
		return false;
	}
#endif
	return true;
}

String File::GetThisProgramName()
{
#ifdef WIN32
	HINSTANCE hInstance = GetModuleHandle(NULL);

	wchar szBuffer[MAX_PATH];
	::GetModuleFileName(hInstance, szBuffer, MAX_PATH);

	String strProgram(szBuffer);
	return strProgram;

#elif CHUOS
	return String();

#else
	return String();
	
#endif
}


String File::GetAbsolutePath(const String& strCurrentDir, const String& strRelativePath, wchar cModel)
{
	wchar szModel[2] = {cModel, 0};

	ASSERT(strCurrentDir.StartsWith(szModel));

	String strPath;

	if( strRelativePath.StartsWith(szModel) )
	{
		// Starts with the root dir, no need to cat with strCurrentDir
		strPath = strRelativePath;
	}
	else
	{
		strPath = FilePath::AddSeparator(strCurrentDir);
		strPath = strPath + strRelativePath;
	}

	StringArray astrFragments = strPath.Split(cModel);
	StringArray astrStack;

	foreach(astrFragments, i)
	{
		String& strFrag = astrFragments[i];
		if( strFrag == L".." )
		{
			// Before : plop/abc/..
			// After:   plop/
			astrStack.RemoveLast();
		}
		else if( strFrag == L"." )
		{
			// Before : plop/abc/.
			// After:   plop/abc/
		}
		else
		{
			if( !strFrag.IsEmpty() )
			{
				astrStack.Add(strFrag);
			}
		}
	}

	if( astrStack.GetSize() == 0 )
		return String(szModel, 1);

	StringBuilder sbResult;
	foreach(astrStack, i)
	{
		sbResult += cModel;
		sbResult += astrStack[i];
	}
	return sbResult.ToString();
}


String File::GetAbsolutePath(const String& strFilePath)
{
#ifdef WIN32
	const int length = ::GetFullPathName(strFilePath.GetBuffer(), 0, NULL, NULL);

	wchar* pBuffer = new wchar[length];
	wchar* pFilePart;
	::GetFullPathName(strFilePath.GetBuffer(), length, pBuffer, &pFilePart);

	String strFullPathName(pBuffer);

	delete[] pBuffer;

	return strFullPathName;

#elif CHUOS
	return strFilePath;

#else
	return strFilePath;
	
#endif
}

bool File::Copy(const String& strPathSrc, const String& strPathDst)
{
#ifdef WIN32
	BOOL bRet = ::CopyFileW(strPathSrc.GetBuffer(), strPathDst.GetBuffer(), FALSE);
	return bRet != 0;
	
#elif CHUOS
	IFileSystem* pFileSystem = null;
	return pFileSystem->CopyFile(strPathSrc, strPathDst);
	
#else
	// TODO
	ASSERT(0);
	return false;
	
#endif
}


bool File::GetFileAttributes(const String& strPath, bool& bDirectory, bool& bReadOnly)
{
#ifdef WIN32
	DWORD dwAttributs = ::GetFileAttributesW(strPath.GetBuffer());
	if( dwAttributs == INVALID_FILE_ATTRIBUTES )
		return false;

	bDirectory = (dwAttributs & FILE_ATTRIBUTE_DIRECTORY) != 0;
	bReadOnly = (dwAttributs & FILE_ATTRIBUTE_READONLY) != 0;

	return true;

#elif CHUOS
	/*
	IFileSystem* pFileSystem = null;
	if( !pFileSystem->GetFileAttributes(strPath, bDirectory, bReadOnly) )
		return false;
	*/
	return true;

#else
	// TODO
	ASSERT(FALSE);
	return false;
	
#endif
}


bool File::Delete(const String& strPath)
{
#ifdef WIN32
	return ::DeleteFileW(strPath.GetBuffer()) != FALSE;

#elif CHUOS

	IFileSystem* pFileSystem = null;
	return pFileSystem->DeleteFile(strPath);

#else
	// TODO
	ASSERT(0);
	return false;
	
#endif
}

bool File::WriteTextUtf8(const String& path, const String& content)
{
	File file;
	if( !file.Open(path, modeWrite) )
	{
		return false;
	}
	static const uint8 bom[] = { 0xEF, 0xBB, 0xBF };
	if( file.Write(bom, sizeof(bom)) != sizeof(bom) )
	{
		return false;
	}
	if( !file.WriteString(content, TextEncoding::Utf8()) )
	{
		return false;
	}
	return true;
}








