///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2011 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#ifndef CHUSTD_FILE_H
#define CHUSTD_FILE_H

#include "IFile.h"
#include "String.h"
#include "Array.h"
#include "FilePath.h"

namespace chustd {

class FileImpl;

class File : public IFile
{
public:
	///////////////////////////////////////////////////////////////////////
	virtual bool  SetPosition(int64 offset, Whence eWhence = posBegin);
	virtual int64 GetPosition() const;
	virtual int64 GetSize();
	virtual int32 Read(void* pBuffer, int size);
	virtual int32 Write(const void* pBuffer, int size);
	virtual ByteOrder GetByteOrder() const;
	virtual void SetByteOrder(ByteOrder eByteOrder);
	///////////////////////////////////////////////////////////////////////

	// Open a file. Default flags are :
	// Read mode + Binary mode + Big Endian mode
	bool Open(const String& strFilePath, uint32 mode = modeRead);
	
	// Close the file
	void Close();

	// Copy the whole file content to a buffer
	ByteArray GetContent();

	// Copy the whole file content to a string
	String GetTextContent(const TextEncoding& te);

	// Copy all the lines of the file to a string array
	StringArray GetLines(const TextEncoding& te);

	uint64 GetLastWriteTime();
	bool   SetLastWriteTime(uint64 lastWriteTime);

	File();
	virtual ~File();

public:
	// Does a copy the whole file content of a file to a buffer
	static ByteArray GetContent(const String& strFilePath);

	// Does a copy the whole file content of a file to a string
	static String GetTextContent(const String& strFilePath, const TextEncoding& te);

	// Does a copy all the lines of the into a string array
	static StringArray GetLines(const String& strFilePath, const TextEncoding& te);

	// Gets the extension of a file, in lower case and without the dot
	static String GetExtension(const String& strFilePath);

	// Returns true if a character is a directory separator
	static bool IsSeparator(const wchar c) { return ( c == '\\' || c == '/'); }

	// Returns true if a path is absolute
	static bool IsAbsolutePath(const String& strPath);

	// Removes the last level of a path
	static String CutToParent(const String& strPath);

	// Deletes a file
	static bool Delete(const String& strPath);
	
	// Renames a file, returns true if the rename occured
	static bool Rename(const String& strOldName, const String& strNewName);
	
	// Performs a file copy, returns true if the copy occured
	static bool Copy(const String& strPathSrc, const String& strPathDst);

	// Gets the drive part of a path (the result can be empty if no drive name is in the path)
	static String GetDrive(const String& strPath);

	// Returns the absolute file name using the current directory of the system
	static String GetAbsolutePath(const String& strFilePath);
	static String GetAbsolutePath(const String& strCurrentDir, const String& strFilePath, wchar cModel = '/');

	// Tests if a file exists
	// If the path describes a directory, this method returns false
	static bool Exists(const String& strPath);

	// Gets the size of a file in bytes
	static int64 GetSize(const String& strPath);

	// Gets this program name
	static String GetThisProgramName();

	// Gets a file attributes
	static bool GetFileAttributes(const String& strPath, bool& bDirectory, bool& bReadOnly);

	static bool WriteTextUtf8(const String& path, const String& content);

private:
	FileImpl* m_pImpl;
};

//////////////////////////////////////////////////////////////////////////////////////////////////////

} // namespace chustd

#endif // ndef CHUSTD_FILE_H
