///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2011 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ImageFormat.h"
#include "File.h"

//////////////////////////////////////////////////////////////////////
using namespace chustd;

//////////////////////////////////////////////////////////////////////
Color32 Color32::black(0, 0, 0);
Color32 Color32::white(255, 255, 255);
Color32 Color32::red(255, 0, 0);
Color32 Color32::green(0, 255, 0);
Color32 Color32::blue(0, 0, 255);
Color32 Color32::yellow(255, 255, 0);
Color32 Color32::cyan(0, 255, 255);
Color32 Color32::magenta(255, 0, 255);
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
PixelFormat AnimFrame::GetPixelFormat() const
{
	if( m_pOwner != null )
	{ 
		return m_pOwner->GetPixelFormat();
	}
	return pfUnknown;
}

bool AnimFrame::IsSimpleTransparencyUsed() const
{
	if( m_pOwner != null )
	{ 
		return m_pOwner->IsSimpleTransparencyUsed();
	}
	return false;
}

uint16 AnimFrame::GetGreyTransIndex() const
{
	if( m_pOwner != null )
	{
		return m_pOwner->GetGreyTransIndex();
	}
	return 0;
}

void AnimFrame::GetTransIndexes(uint16& red, uint16& green, uint16& blue) const
{
	if( m_pOwner != null )
	{
		return m_pOwner->GetTransIndexes(red, green, blue);
	}
	red = green = blue = 0;
}

const Palette& AnimFrame::GetPalette() const
{
	if( m_pOwner != null )
	{
		return m_pOwner->GetPalette();
	}
	return Palette::Null;
}

//////////////////////////////////////////////////////////////////////
const Palette Palette::Null;

//////////////////////////////////////////////////////////////////////
// Returns true if one of the color has a alpha not equal to 255
bool Palette::HasNonOpaqueColor() const
{
	for(int i = 0; i < m_count; ++i) 
	{
		if( m_colors[i].GetAlpha() != 255 )
		{
			return true;
		}
	}
	return false;
}

bool Palette::AllAlphasAreOpaque() const
{
	for(int i = 0; i < m_count; ++i) 
	{
		if( m_colors[i].GetAlpha() != 0 )
			return false;
	}
	return true;
}

void Palette::SetAlphaFullOpaque()
{
	for(int i = 0; i < m_count; ++i) 
	{
		m_colors[i].SetAlpha(255);
	}
}

// Returns -1 if not found
int Palette::FindColor(Color32 col) const
{
	for(int i = 0; i < m_count; ++i) 
	{
		if( m_colors[i] == col )
		{
			return i;
		}
	}
	return -1;
}

// Returns -1 if not found
int Palette::GetFirstFullyTransparentColor() const
{
	for(int i = 0; i < m_count; ++i) 
	{
		if( m_colors[i].a == 0 )
		{
			return i;
		}
	}
	return -1;
}

//////////////////////////////////////////////////////////////////////
ImageFormat::ImageFormat()
{
	m_nWidth = 0;
	m_nHeight = 0;
	m_nLastError = 0;
}

ImageFormat::~ImageFormat()
{

}

////////////////////////////////////////////////////
bool ImageFormat::IsIndexed(PixelFormat pf)
{
	return pf == pf1bppIndexed || pf == pf2bppIndexed || pf == pf4bppIndexed || pf == pf8bppIndexed;
}

int32 ImageFormat::SizeofPixelInBits(PixelFormat ePixelFormat)
{
	switch(ePixelFormat)
	{
	case pfUnknown:
		return 0;
		
	case pf1bppGrayScale:
		return 1;
	case pf2bppGrayScale:
		return 2;
	case pf4bppGrayScale:
		return 4;
	case pf8bppGrayScale:
		return 8;
	case pf16bppGrayScale:
		return 16;
		
	case pf16bppGrayScaleAlpha:
		return 16;
	case pf32bppGrayScaleAlpha:
		return 32;

	case pf1bppIndexed:
		return 1;
	case pf2bppIndexed:
		return 2;
	case pf4bppIndexed:
		return 4;
	case pf8bppIndexed:
		return 8;
		
	case pf16bppArgb1555:
	case pf16bppArgb4444:
	case pf16bppRgb555:
	case pf16bppRgb565:
		return 16;

	case pf24bppRgb:
	case pf24bppBgr:
		return 24;

	case pf32bppRgba:
	case pf32bppBgra:
		return 32;

	case pf48bppRgb:
		return 48;

	case pf64bppRgba:
		return 64;

	}
	ASSERT(0);
	return 0;
}

int32 ImageFormat::ComputeByteWidth(PixelFormat epf, int32 width)
{
	const int32 sizeofPixelInBits = ImageFormat::SizeofPixelInBits(epf);

	// Size of a row in bits
	const int32 bitWidth = width * sizeofPixelInBits;
	
	// Round up for size of a row in bytes
	const int32 addWidth = ((bitWidth & 7) != 0) ? 1 : 0;
	const int32 byteWidth = bitWidth / 8 + addWidth;

	return byteWidth;
}
//////////////////////////////////////////////////////////////////////////////

String ImageFormat::GetLastErrorString() const
{
	switch(m_nLastError)
	{
	case ioErr:
		return L"Error accessing to the file";
	case badFileFormat:
		return L"Bad File Format";
	case notEnoughMemory:
		return L"Not enough memory";
	case uncompleteFile:
		return L"Uncomplete file";
	}
	return L"Unknown error";
}

bool ImageFormat::Load(const String& strFileName)
{
	File file;
	if( !file.Open(strFileName) )
	{
		m_nLastError = ioErr;
		return false;
	}
	return LoadFromFile(file);
}

int32 ImageFormat::GetWidth() const
{
	return m_nWidth;
}

int32 ImageFormat::GetHeight() const
{
	return m_nHeight;
}

const Buffer& ImageFormat::GetPixels() const
{
	return m_pixels;
}

void ImageFormat::FreeBuffer()
{
	m_pixels.Clear();
}

void ImageFormat::FlipVertical()
{
	PixelFormat epf = GetPixelFormat();
	if( epf == pfUnknown )
	{
		return;
	}
	const int32 bytesPerRow = ImageFormat::ComputeByteWidth(epf, m_nWidth);
	uint8* pComponents = m_pixels.GetWritePtr();
	const int32 heightDiv2 = m_nHeight / 2;

	for(int iY = 0; iY < heightDiv2; ++iY)
	{
		uint8* pSrc = pComponents + (iY * bytesPerRow);
		uint8* pDst = pComponents + ((m_nHeight - 1 - iY) * bytesPerRow);

		for(int iByte = 0; iByte < bytesPerRow; ++iByte)
		{
			uint8 swap = pDst[iByte];
			pDst[iByte] = pSrc[iByte];
			pSrc[iByte] = swap;
		}
	}
}

void ImageFormat::SetAlphaFullOpaque()
{
	PixelFormat epf = GetPixelFormat();
	if( epf == pfUnknown )
	{
		return;
	}

	const int32 pixelCount = m_nWidth * m_nHeight;
	if( epf == pf32bppRgba || epf == pf32bppBgra )
	{
		uint8* pDst = m_pixels.GetWritePtr();
		const uint8 alpha = 255;
		for(int i = 0; i < pixelCount; ++i)
		{
			pDst[3] = alpha;
			pDst += 4;
		}

	}
}

int32 ImageFormat::GetLastError() const
{ 
	return m_nLastError; 
}

bool ImageFormat::IsIndexed() const 
{ 
	return IsIndexed(GetPixelFormat()); 
}


bool ImageFormat::IsAnimated() const
{
	return false;
}

bool ImageFormat::HasDefaultImage() const
{
	return false;
}

int32 ImageFormat::GetFrameCount() const
{
	return 0;
}

const AnimFrame* ImageFormat::GetAnimFrame(int /*index*/) const
{
	return null;
}

int32 ImageFormat::GetLoopCount() const
{
	return 0;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Packs single-pixel bytes into multiple-pixels bytes.
//
// pixels [in,out]  Pixels in: one pixel per byte, out: multiple pixels per byte
// width  [in]      Width in pixels
// height [in]      Height in pixels
// pixelFormat [in] Target pixel format
//
// Returns true upon success
///////////////////////////////////////////////////////////////////////////////////////////////////
bool ImageFormat::PackPixels(Buffer& pixels, int width, int height, PixelFormat pixelFormat)
{
	if( width < 0 || height < 0 )
	{
		return false; // Bad arg
	}
	int bitsPerPix = ImageFormat::SizeofPixelInBits(pixelFormat);
	if( bitsPerPix > 8 )
	{
		return false; // Bad arg
	}
	if( bitsPerPix == 8 )
	{
		// Nothing to do
		return true;
	}
	if( pixels.GetSize() < (width * height) )
	{
		ASSERT(0);
		return false;
	}
	const int32 bytesPerRow = ImageFormat::ComputeByteWidth(pixelFormat, width);
	uint8* pIn = pixels.GetWritePtr();
	uint8* pOut = pIn;
	
	// Note: there is padding at the end of each row
	uint8 currentByte = 0; // Current byte value
	int usedBitCount = 0;  // Number of bits used inside the byte

	for(int iRow = 0; iRow < height; iRow++)
	{
		currentByte = 0;
		usedBitCount = 0;
		
		for(int iCol = 0; iCol < width; iCol++)
		{
			uint8 nVal = pIn[iRow * width + iCol];
			
			usedBitCount += bitsPerPix;
			currentByte |= (nVal << (8 - usedBitCount));
			
			if( usedBitCount == 8 )
			{
				pOut[0] = currentByte;
				++pOut;
				
				currentByte = 0;
				usedBitCount = 0;
			}
		}
		
		if( usedBitCount != 0 )
		{
			pOut[0] = currentByte;
			++pOut;
		}
	}
	int newSize = bytesPerRow * height;
	pixels.SetSize(newSize);
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Unpacks multiple pixels per byte to get one pixel per byte.
//
// [in,out] pixels      Pixels in: one pixel per bytes, out: multiple pixels per byte
// [in]     width       Width in pixels
// [in]     height      Height in pixels
// [in]     pixelFormat Source pixel format
//
// Returns true upon success.
///////////////////////////////////////////////////////////////////////////////////////////////////
bool ImageFormat::UnpackPixels(Buffer& pixels, int width, int height, PixelFormat pixelFormat)
{
	int bitsPerPix = ImageFormat::SizeofPixelInBits(pixelFormat);
	if( bitsPerPix >= 8 )
	{
		// Nothing to do
		return true;
	}
	Buffer newBuffer;
	if( !newBuffer.SetSize(width * height) )
	{
		return false;
	}

	const uint8* pSrc = pixels.GetReadPtr();
	uint8* pDst = newBuffer.GetWritePtr();

	int32 mask = ~((~0L) << bitsPerPix);
	int32 shift = 8 - bitsPerPix;
	
	int srcIndex = 0;
	int dstIndex = 0;

	for(int i = 0; i < height; ++i)
	{
		for(int j = 0; j < width; ++j)
		{
			uint8 currentByte = pSrc[srcIndex];

			uint8 colorIndex = uint8( (currentByte >> shift) & mask);
			
			shift -= bitsPerPix;
			if( shift < 0 )
			{
				shift = 8 - bitsPerPix;
				srcIndex++;
			}
			
			pDst[dstIndex] = colorIndex;
			dstIndex++;
		}
		// Some padding bits may remain
		if( shift != 8 - bitsPerPix )
		{
			shift = 8 - bitsPerPix;
			srcIndex++;
		}
	}
	pixels = newBuffer;
	return true;
}
