///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2011 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Png.h"
#include "File.h"
#include "Math.h"
#include "TextEncoding.h"

//////////////////////////////////////////////////////////////////////
using namespace chustd;

static const uint8 k_PngSignature[8] = {0x89,'P','N','G','\r','\n',0x1a,'\n'};

static const int k_nChunkDataSize_IHDR = 13;
static const int k_nChunkDataSize_fcTL = 26;

//////////////////////////////////////////////////////////////////////
void Png::ImageDataInfo::Clear()
{
	pPixels = null;
	width = 0;
	height = 0;
	byteWidth = 0;
	uncompressedDataSize = 0;
}

//////////////////////////////////////////////////////////////////////
Png::Png()
{
	Initialize();
}

Png::~Png()
{

}

// Initialize members
// The initialization is made outside the constructor so we can reset the
// instance in order to load several images with the same instance
void Png::Initialize()
{
	m_IHDR.Clear();

	m_nWidth = 0;
	m_nHeight = 0;
	m_nLastError = 0;

	m_PLTE.m_count = 0;

	m_nSizeofPixel = 0;
	m_nSizeofPixelInBits = 0;

	m_idiCurrent.Clear();

	m_nCounter_IDAT = 0;
	m_nCounter_fdAT = 0;
	m_nOutputOffset = 0;

	m_nGamma = 100000;
	m_tRNS.Clear();
	m_bkGD.Clear();
	m_tEXts.Clear();

	m_acTL.frameCount = 0;
	m_acTL.loopCount = 0;

	m_apFrames.Clear();
	m_currentSequenceNumber = 0;

	m_nExpectedChunks = cf_IHDR;
	m_nLastUnexpectedChunk = 0;
	m_foundChunks = 0;
}

bool Png::IsPng(IFile& file)
{
	const int32 signSize = sizeof(k_PngSignature);
	uint8 aRead[signSize];
	
	if( file.Read(aRead, signSize) != signSize )
	{
		return false;
	}

	if( !chustd::Memory::Equals(k_PngSignature, aRead, signSize) )
	{
		return false;
	}
	return true;
}

bool Png::WriteSignature(IFile& file)
{
	int32 written = file.Write(k_PngSignature, 8);
	return written == 8;
}

// Loads a png picture from an open file
// Returns false if the loading failed
bool Png::LoadFromFile(IFile& file)
{
	typedef bool(Png::*PFN_CHUNKHANDLER)(IFile&, int32);
	struct ChunkInfo
	{
		uint32 name;
		uint32 flag;
		PFN_CHUNKHANDLER pfnHandler;
		uint32 unwantedAfter;
	};

	static const ChunkInfo aChunkInfo[] =
	{
		{ chunk_IHDR, cf_IHDR, &Png::Handle_IHDR, cf_IHDR  },
		{ chunk_gAMA, cf_gAMA, &Png::Handle_gAMA, cf_IHDR|cf_gAMA  },
		{ chunk_PLTE, cf_PLTE, &Png::Handle_PLTE, cf_IHDR|cf_gAMA|cf_PLTE  },
		{ chunk_tRNS, cf_tRNS, &Png::Handle_tRNS, cf_IHDR|cf_gAMA|cf_PLTE  },
		{ chunk_IDAT, cf_IDAT, &Png::Handle_IDAT, cf_IHDR|cf_gAMA|cf_PLTE|cf_tRNS|cf_bKGD|cf_acTL  },
		{ chunk_bKGD, cf_bKGD, &Png::Handle_bKGD, cf_IHDR|cf_gAMA|cf_PLTE  },
		{ chunk_tEXt, cf_tEXt, &Png::Handle_tEXt, cf_IEND  },
		{ chunk_acTL, cf_acTL, &Png::Handle_acTL, cf_IHDR  },
		{ chunk_fcTL, cf_fcTL, &Png::Handle_fcTL, cf_IHDR|cf_acTL|cf_fcTL  },
		{ chunk_fdAT, cf_fdAT, &Png::Handle_fdAT, cf_IHDR|cf_acTL|cf_IDAT  }
	};

	// Reset the object
	FreeBuffer();
	Initialize();


	//////////////////////////////////////////////////////
	// Read the header first
	if( !IsPng(file) )
	{
		m_nLastError = errNotAPngFile;
		return false;
	}

	ChunkedFile cd(file);

	// Current position is the start of a chunk collection

	bool bOkHandled = true;
	bool bIENDFound = false;

	while( !bIENDFound)
	{
		bool bChunkPresent = cd.BeginChunkRead();
		if( !bChunkPresent )
		{
			// No more chunk
			break;
		}

		if( cd.m_nChunk == chunk_IEND )
		{
			bIENDFound = true;
			break;
		}

		const int handlerCount = sizeof(aChunkInfo) / sizeof(aChunkInfo[0]);
		for(int i = 0; i < handlerCount; ++i)
		{
			const ChunkInfo& chunkinfo = aChunkInfo[i];
			if( chunkinfo.name == cd.m_nChunk )
			{
				// Handler found
				if( (m_nExpectedChunks & uint32(chunkinfo.flag)) == 0 )
				{
					// Error, unexpected chunk
					m_nLastError = errUnexpectedChunk;
					m_nLastUnexpectedChunk = cd.m_nChunk;
					bOkHandled = false;
				}
				else
				{
					// Ok to decode
					bOkHandled = (this->*chunkinfo.pfnHandler)(file, cd.m_nSizeof);
					m_foundChunks |= chunkinfo.flag;
					m_nExpectedChunks = cf_ALL & ~(chunkinfo.unwantedAfter);
				}
				break;
			}
		}

		if( !bOkHandled )
		{
			// Error while reading a chunk content in one of the Handle_* function
			// m_nLastError is correctly set by one of those functions
			break;
		}

		if( !cd.EndChunkRead() )
		{
			// Unexpected end of stream, CRC not present or chunk mishandled (read too much)
			m_nLastError = uncompleteFile;
			bOkHandled = false;
			break;
		}
	}

	m_CompressedBuffer.Clear();

	if( bOkHandled && !bIENDFound )
	{
		// End of stream with no damaged chunk but IEND not found
		m_nLastError = errNoIENDFound;
		
		// We set bOkHandled to false so we perform the needed on-error cleanup
		bOkHandled = false;
	}
	
	if( !bOkHandled )
	{
		// An error occurred, we clean the object except the lasterror field
		m_ZStream.InflateEnd();

		int32 lastError = m_nLastError;
		FreeBuffer();
		m_nLastError = lastError;
		return false;
	}

	if( !IsChunkFound(cf_IDAT) )
	{
		// No error yet, but no mandatory IDAT chunk found
		FreeBuffer();
		m_nLastError = errIDATNotFound;
		return false;
	}

	bool bTerminateOk = EndImageDataProcessing();
	if( !bTerminateOk )
	{
		// An error occurred, we clean the object except the lasterror field
		// (TerminateJob fills m_nLastError correctly)

		int32 lastError = m_nLastError;
		FreeBuffer();
		m_nLastError = lastError;
		return false;
	}

	const int32 frameCount = m_apFrames.GetSize();

	// Check default image
	if( m_pixels.IsEmpty() )
	{
		// APNG with no default image ?
		if( frameCount >= 0 )
		{
			// Yes, the IDAT is part of the animation. But to be valid, its assigned fcTL must have
			// values compatible with those in the IHDR.
			ApngFrame* pFrame = m_apFrames[0];
			bool bOk1 = (pFrame->m_fctl.width == m_IHDR.width);
			bool bOk2 = (pFrame->m_fctl.height == m_IHDR.height);
			bool bOk3 = (pFrame->m_fctl.offsetX == 0);
			bool bOk4 = (pFrame->m_fctl.offsetY == 0);

			if( !(bOk1 && bOk2 && bOk3 && bOk4) )
			{
				// Error
				FreeBuffer();
				m_nLastError = errIDATfcTLNotConsistentWithIHDR;
				return false;
			}
		}
	}

	// Check acTL missing
	if( frameCount >= 1 )
	{
		if( !IsChunkFound(cf_acTL) )
		{
			m_nLastError = erracTLMissing;
			return false;
		}
	}

	// Check frame count
	if( IsChunkFound(cf_acTL) )
	{
		if( m_acTL.frameCount != frameCount )
		{
			m_nLastError = errFrameCountMismatch;
			return false;
		}
	}
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
bool Png::IsValidBitDepth(uint8 bitDepth)
{
	return bitDepth == 1 || bitDepth == 2 || bitDepth == 4 || bitDepth == 8 || bitDepth == 16;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
bool Png::IsValidColorType(uint8 colorType)
{
	return colorType == 0 || colorType == 2 || colorType == 3 || colorType == 4 || colorType == 6;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
int Png::Read_IHDR(IFile& file, int32 dataSizeof, PngChunk_IHDR& out)
{
	if( dataSizeof != k_nChunkDataSize_IHDR )
	{
		return errBadHeaderSize;
	}

	ASSERT( sizeof(PngChunk_IHDR) >= k_nChunkDataSize_IHDR );

	if( file.Read(&out, dataSizeof) != int32(dataSizeof) )
	{
		return uncompleteFile;
	}
	
	if( file.ShouldSwapBytes() )
	{
		out.SwapBytes();
	}

	// Maximum size is 2**25, this lets us 7 bits to do computations on width and height
	const int32 maxWidth = 1 << 25;
	if(    !(0 <= out.width && out.width <= maxWidth)
		|| !(0 <= out.height && out.height <= maxWidth) )
	{
		// We limit the picture size to a sensible maximum size so we can detect
		// bad data in the file
		return errBadPicSize;
	}

	if( !IsValidBitDepth(out.bitDepth) )
	{
		return errBadBitDepth;
	}

	if( !IsValidColorType(out.colorType) )
	{
		return errBadColorType;
	}

	// Check combination
	if( GetPixelFormat(out.colorType, out.bitDepth) == pfUnknown )
	{
		return errBadPixelFormat;
	}

	if( out.compressionMethod != 0 )
	{
		return errBadCompressionMethod;
	}

	if( out.filterMethod != 0 )
	{
		// The only filtering method supported is "Adaptative"
		// (none, sub, up, average, paeth)
		return errBadFilterMethod;
	}

	if( out.interlaceMethod > 1 )
	{
		return errBadInterlaceMethod;
	}
	return noErr;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
bool Png::Handle_IHDR(IFile& file, int32 dataSizeof)
{
	int ret = Read_IHDR(file, dataSizeof, m_IHDR);
	if( ret != 0 )
	{
		m_nLastError = ret;
		return false;
	}

	// Fill parent class attributes
	m_nWidth = m_IHDR.width;
	m_nHeight = m_IHDR.height;

	PixelFormat epf = GetPixelFormat();
	m_nSizeofPixelInBits = ImageFormat::SizeofPixelInBits(epf);

	// Computes the size of a pixel in bytes
	m_nSizeofPixel = (m_nSizeofPixelInBits + 7 ) / 8;
	return true;
}

// Allocates the image buffer
// bInterlacedBuffer: true if the buffer will hold interlaced data
// Returns false if allocation failed
bool Png::AllocateImageBuffer(bool bInterlacedBuffer)
{
	// One extra byte is used in order to store the filtering sub-code
	const int32 rowByteCount = m_idiCurrent.byteWidth + 1;

	if( bInterlacedBuffer )
	{
		// The raw data size is bigger is interlacing is used
		m_idiCurrent.uncompressedDataSize = ComputeInterlacedSize(m_idiCurrent.width, m_idiCurrent.height, m_nSizeofPixelInBits);
	}
	else
	{
		m_idiCurrent.uncompressedDataSize = rowByteCount * m_idiCurrent.height;
	}

	if( !m_idiCurrent.pPixels->SetSize(m_idiCurrent.uncompressedDataSize) )
	{
		m_nLastError = notEnoughMemory;
		return false;
	}
	return true;
}

// Gets gamma information
bool Png::Handle_gAMA(IFile& file, int32 dataSizeof)
{
	if( dataSizeof < 4 )
	{
		m_nLastError = errNotEnoughDataInGammaChunk;
		return false;
	}
	
	if( !file.Read32(m_nGamma) )
	{
		m_nLastError = uncompleteFile;
		return false;
	}

	return true;
}

// Gets bKGD information
bool Png::Handle_bKGD(IFile& file, int32 dataSizeof)
{
	if( m_IHDR.colorType == 0x03 )
	{
		// Indexed color
		if( dataSizeof < 1 )
		{
			m_nLastError = errNotEnoughDataInBkChunk;
			return false;
		}

		if( !file.Read8(m_bkGD.index) )
		{
			m_nLastError = uncompleteFile;
			return false;
		}
		return true;
	}
	
	if( m_IHDR.colorType == 0x00 || m_IHDR.colorType == 0x04 )
	{
		// Grey or Grey+Alpha
		if( dataSizeof < 2 )
		{
			m_nLastError = errNotEnoughDataInBkChunk;
			return false;
		}

		if( !file.Read16(m_bkGD.grey) )
		{
			m_nLastError = uncompleteFile;
			return false;
		}
		return true;
	}

	if( m_IHDR.colorType == 0x02 || m_IHDR.colorType == 0x06 )
	{
		if( dataSizeof < 6 )
		{
			m_nLastError = errNotEnoughDataInBkChunk;
			return false;
		}
		
		// TrueColor or TrueColor+Alpha
		if( !(file.Read16(m_bkGD.red) && file.Read16(m_bkGD.green) && file.Read16(m_bkGD.blue)) )
		{
			m_nLastError = uncompleteFile;
			return false;
		}
		return true;
	}

	// Other type should not exist and be detected when handling the IHDR chunk
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
bool Png::Handle_tEXt(IFile& file, int32 dataSizeof)
{
	PngChunk_tEXt content;
	ByteArray bytes;
	if( !bytes.SetSize(dataSizeof) )
	{
		m_nLastError = notEnoughMemory;
		return false;
	}
	if( file.Read(bytes.GetPtr(), dataSizeof) != dataSizeof )
	{
		m_nLastError = uncompleteFile;
		return false;
	}
	// Find the keyword/data separator
	int zeroAt = -1;
	foreach(bytes, i)
	{
		if( bytes[i] == 0 )
		{
			zeroAt = i;
			break;
		}
	}
	if( zeroAt < 0 )
	{
		m_nLastError = errTextMissingSeparator;
		return false;
	}
	if( !(1 <= zeroAt && zeroAt <= 79) )
	{
		m_nLastError = errTextBadKeywordLength;
		return false;
	}
	ByteArray kwBytes;
	kwBytes.Set(bytes.GetPtr(), zeroAt);
	ByteArray dataBytes;
	dataBytes.Set(bytes.GetPtr() + zeroAt + 1, bytes.GetSize() - zeroAt - 1);
	content.keyword = TextEncoding::Iso8859_1().BytesToString(kwBytes);
	content.data = TextEncoding::Iso8859_1().BytesToString(dataBytes);
	m_tEXts.Add(content);
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
bool Png::Handle_tRNS(IFile& file, int32 dataSizeof)
{
	if( m_IHDR.colorType == 3 )
	{
		// Indexed color image
		// Note : there can be less alpha values than palette entries
		// In this case, the remaining entries use 255 for their alpha value
		uint8 aAlphaValues[256];

		int32 colorCount = dataSizeof;
		if( colorCount > 256 )
		{
			colorCount = 256;
		}
		else if( colorCount < 0 )
		{
			colorCount = 0;
		}

		file.Read(aAlphaValues, colorCount);

		for(int i = 0; i < colorCount; ++i)
		{
			m_PLTE.m_colors[i].SetAlpha( aAlphaValues[i]);
		}
	}
	else if( m_IHDR.colorType == 0 )
	{
		// Greyscale
		if( !file.Read16(m_tRNS.grey) )
		{
			m_nLastError = uncompleteFile;
			return false;
		}
	}
	else if( m_IHDR.colorType == 2 )
	{
		// True color without alpha
		if( !(file.Read16(m_tRNS.red) && file.Read16(m_tRNS.green) && file.Read16(m_tRNS.blue)) )
		{
			m_nLastError = uncompleteFile;
			return false;
		}
	}
	return true;
}

// Gets palette information
bool Png::Handle_PLTE(IFile& file, int32 dataSizeof)
{
	int32 colorCount = dataSizeof / 3;
	if( colorCount > 256 )
	{
		colorCount = 256;
	}
	m_PLTE.m_count = colorCount;

	uint8 aTmpPalette[256 * 3];
	file.Read( aTmpPalette, colorCount * 3);

	int32 iTmpColor = 0;

	for(int i = 0; i < m_PLTE.m_count; ++i)
	{
		const uint8 r = aTmpPalette[iTmpColor + 0];
		const uint8 g = aTmpPalette[iTmpColor + 1];
		const uint8 b = aTmpPalette[iTmpColor + 2];
		iTmpColor += 3;

		m_PLTE[i].SetRgb(r, g, b);
	}
	return true;
}

bool Png::BeginImageDataProcessing()
{
	if( m_apFrames.GetSize() == 0 )
	{
		// The default image is not part of the animation, use the image pixel buffer
		m_idiCurrent.pPixels = &m_pixels;
		m_idiCurrent.width = m_IHDR.width;
		m_idiCurrent.height = m_IHDR.height;
	}
	else
	{
		// Uncompressing a frame
		ApngFrame* pFrame = m_apFrames.GetLast();
		
		m_idiCurrent.pPixels = &(pFrame->m_pixels);
		m_idiCurrent.width = pFrame->m_fctl.width;
		m_idiCurrent.height = pFrame->m_fctl.height;
	}

	// Compute the width of the image in bytes
	PixelFormat epf = GetPixelFormat();
	m_idiCurrent.byteWidth = ImageFormat::ComputeByteWidth(epf, m_idiCurrent.width);

	m_idiCurrent.uncompressedDataSize = 0; // Filled by AllocateImageBuffer

	// Allocate the image buffer
	if( !AllocateImageBuffer(m_IHDR.interlaceMethod == 1) )
	{
		return false;
	}

	// The first time we meet an IDAT chunk, we initialize the zstream structure
	m_ZStream.SetBuffers(null, 0, null, 0);
	
	const int nZErr = m_ZStream.InflateInit();
	if( nZErr != ZStream::Z_OK )
	{
		m_nLastError = errInflateErr;
		return false;
	}
	m_nOutputOffset = 0;
	return true;
}

// Gets image data information
bool Png::Handle_IDAT(IFile& file, int32 dataSizeof)
{
	m_nCounter_IDAT++;
	if( m_nCounter_IDAT == 1 )
	{
		// Special case for first IDAT
		if( !BeginImageDataProcessing() )
		{
			return false;
		}
	}

	return ProcessImageData(file, dataSizeof);
}

bool Png::ProcessImageData(IFile& file, int32 dataSizeof)
{
	if( dataSizeof == 0 )
	{
		// Valid, skip this data to avoid any decoding error
		return true;
	}

	const uint32 compressedSize = dataSizeof;
	if( !m_CompressedBuffer.SetSize(compressedSize) )
	{
		m_nLastError = notEnoughMemory;
		return false;
	}

	uint32 read = file.Read(m_CompressedBuffer.GetPtr(), compressedSize);
	if( read != compressedSize )
	{
		m_nLastError = uncompleteFile;
		return false;
	}
	
	const uint32 availableOut = m_idiCurrent.uncompressedDataSize - m_nOutputOffset;

	uint8* pIn  = m_CompressedBuffer.GetPtr();
	uint8* pOut = m_idiCurrent.pPixels->GetWritePtr() + m_nOutputOffset;
	m_ZStream.SetBuffers(pIn, compressedSize, pOut, availableOut);
	
	int nZRet2 = m_ZStream.Inflate(ZStream::Z_SYNC_FLUSH);
	
	if( nZRet2 == ZStream::Z_OK )
	{
	
	}
	else if( nZRet2 == ZStream::Z_STREAM_END )
	{
	
	}
	else
	{
		m_nLastError = errInflateErr;
		return false;
	}

	uint32 uncompressedSize = availableOut - m_ZStream.GetOutAvailable();
	m_nOutputOffset += uncompressedSize;

	return true;
}

bool Png::Handle_acTL(IFile& file, int32 dataSizeof)
{
	if( IsChunkFound(cf_acTL) )
	{
		m_nLastError = erracTLRepeated;
		return false;
	}

	if( IsChunkFound(cf_IDAT) )
	{
		// acTL must appear before the first IDAT chunk within a valid PNG stream
		m_nLastError = errIDATFoundBeforeacTL;
		return false;
	}
	
	if( dataSizeof < 8 )
	{
		m_nLastError = errNotEnoughDataInAnimCtlChunk;
		return false;	
	}
	
	if( !(file.Read32(m_acTL.frameCount) && file.Read32(m_acTL.loopCount)) )
	{
		m_nLastError = uncompleteFile;
		return false;	
	}
	
	if( m_acTL.frameCount == 0 )
	{
		// Not valid
		m_nLastError = errFrameCountIsZero;
		return false;
	}

	if( m_acTL.frameCount > MAX_INT32 )
	{
		m_nLastError = errFrameCountOutsideRange;
		return false;
	}

	return true;
}

bool Png::Handle_fcTL(IFile& file, int32 dataSizeof)
{
	// Finish processing previous image data for a previous IDAT or previous fdAT.
	// Except if no IDAT or no fdAT was pending
	if( m_nCounter_IDAT > 0 || m_nCounter_fdAT > 0 )
	{
		if( !EndImageDataProcessing() )
		{
			return false;
		}
	}

	////////////////////////////////////////////////////////////
	const int wantedSizeof = k_nChunkDataSize_fcTL;
	if( dataSizeof < wantedSizeof )
	{
		m_nLastError = errNotEnoughDataInFrameCtlChunk;
		return false;	
	}

	PngChunk_fcTL fctl;
	if( file.Read(&fctl, wantedSizeof) != wantedSizeof )
	{
		m_nLastError = uncompleteFile;
		return false;	
	}

	if( file.ShouldSwapBytes() )
	{
		fctl.SwapBytes();
	}

	if( fctl.sequenceNumber != m_currentSequenceNumber )
	{
		m_nLastError = errBadFrameControlSequenceNumber;
		return false;
	}

	if( fctl.delayFracDenominator == 0 )
	{
		// If the denominator is 0, it is to be treated as if it were 100
		fctl.delayFracDenominator = 100;
	}
	m_currentSequenceNumber++; // Incremented for fcTL and fdAT chunks

	ApngFrame* pFrame = new ApngFrame(this);
	if( pFrame == null )
	{
		m_nLastError = notEnoughMemory;
		return false;
	}
	pFrame->m_fctl = fctl;
	if( m_apFrames.Add(pFrame) < 0 )
	{
		m_nLastError = notEnoughMemory;
		return false;
	}

	return true;
}

bool Png::Handle_fdAT(IFile& file, int32 dataSizeof)
{
	if( m_nCounter_IDAT != 0 )
	{
		m_nLastError = errMissingfcTLBeforefdAT;
		return false;
	}

	int32 sequenceNumber;
	if( !file.Read32(sequenceNumber) )
	{
		m_nLastError = uncompleteFile;
		return false;
	}

	// Verify that the sequence number is consistent
	if( sequenceNumber != m_currentSequenceNumber )
	{
		m_nLastError = errBadFrameDataSequenceNumber;
		return false;
	}
	m_currentSequenceNumber++; // Incremented for fcTL and fdAT chunks

	m_nCounter_fdAT++;
	if( m_nCounter_fdAT == 1 )
	{
		// Special case for first fdAT
		if( !BeginImageDataProcessing() )
		{
			return false;
		}
	}
	return ProcessImageData(file, dataSizeof - 4);
}

PixelFormat Png::GetPixelFormat(const PngChunk_IHDR& pngIHDR)
{
	return GetPixelFormat(pngIHDR.colorType, pngIHDR.bitDepth);
}

PixelFormat Png::GetPixelFormat(uint8 colorType, uint8 bitDepth)
{
	// To the bottom : 
	//   default: grey, flags: 0x1=palette used, 0x2=color used, 0x4=alpha used
	// To the left :
	//   Bit depth
	static const PixelFormat pixelFormats[7][6] =
	{
		{ pf1bppGrayScale, pf2bppGrayScale, pf4bppGrayScale, pf8bppGrayScale, pf16bppGrayScale },
		{ pfUnknown,       pfUnknown,       pfUnknown,       pfUnknown,       pfUnknown },
		{ pfUnknown,       pfUnknown,       pfUnknown,       pf24bppRgb,      pf48bppRgb },
		{ pf1bppIndexed,   pf2bppIndexed,   pf4bppIndexed,   pf8bppIndexed,   pfUnknown },
		{ pfUnknown,       pfUnknown,       pfUnknown,       pf16bppGrayScaleAlpha, pf32bppGrayScaleAlpha },
		{ pfUnknown,       pfUnknown,       pfUnknown,       pfUnknown,       pfUnknown },
		{ pfUnknown,       pfUnknown,       pfUnknown,       pf32bppRgba,     pf64bppRgba  }
	};

	const int down = colorType;
	const int left = Math::Log2(bitDepth);
	if( (0 <= down && down <= 6) && (0 <= left && left <= 4) )
	{
		return pixelFormats[down][left];
	}
	return pfUnknown;
}

PixelFormat Png::GetPixelFormat() const
{
	return GetPixelFormat(m_IHDR);
}

bool Png::GetIHDRFeaturesFromPixelFormat(PixelFormat epf, uint8& colorType, uint8& bitsPerComponent)
{
	switch(epf)
	{
	case pfUnknown:
		return false;
		
	case pf1bppGrayScale:      // 1bit per pixel
		colorType = cfGrey;
		bitsPerComponent = 1;
		return true;

	case pf2bppGrayScale:       // 2bits per pixel
		colorType = cfGrey;
		bitsPerComponent = 2;
		return true;

	case pf4bppGrayScale:       // 4bits per pixel
		colorType = cfGrey;
		bitsPerComponent = 4;
		return true;

	case pf8bppGrayScale:       // 8bits per pixel
		colorType = cfGrey;
		bitsPerComponent = 8;
		return true;

	case pf16bppGrayScale:      // 16bits per pixel
		colorType = cfGrey;
		bitsPerComponent = 16;
		return true;

	case pf16bppGrayScaleAlpha:  // GGAA
		colorType = cfAlpha;
		bitsPerComponent = 8;
		return true;

	case pf32bppGrayScaleAlpha: // GGGGAAAA
		colorType = cfAlpha;
		bitsPerComponent = 16;
		return true;

	case pf1bppIndexed:    // Indexed 1bit per pixel
		colorType = cfColor | cfPalette;
		bitsPerComponent = 1;
		return true;

	case pf2bppIndexed:    // Indexed 2bits per pixel
		colorType = cfColor | cfPalette;
		bitsPerComponent = 2;
		return true;

	case pf4bppIndexed:    // Indexed 4bits per pixel
		colorType = cfColor | cfPalette;
		bitsPerComponent = 4;
		return true;

	case pf8bppIndexed:    // Indexed 8bits per pixel
		colorType = cfColor | cfPalette;
		bitsPerComponent = 8;
		return true;

	case pf24bppRgb:        // RRGGBB
		colorType = cfColor;
		bitsPerComponent = 8;
		return true;

	case pf48bppRgb:       // RRRRGGGGBBBB 
		colorType = cfColor;
		bitsPerComponent = 16;
		return true;

	case pf32bppRgba:   // RRGGBBAA
		colorType = cfColor | cfAlpha;
		bitsPerComponent = 8;
		return true;

	case pf32bppBgra:   // BBGGRRAA
		colorType = cfColor | cfAlpha;
		bitsPerComponent = 8;
		return true;

	case pf64bppRgba:  // RRRRGGGGBBBBAAAA	
		colorType = cfColor | cfAlpha;
		bitsPerComponent = 16;
		return true;
	}
	
	return false;
}


const Palette& Png::GetPalette() const
{
	return m_PLTE;
}

const Buffer& Png::GetPixels() const
{
	if( m_pixels.IsEmpty() )
	{
		// APNG with no default image ?
		const int32 frameCount = m_apFrames.GetSize();
		if( frameCount >= 0 )
		{
			// Yes, return the first frame, whose size and offset 
			// must be compatible with what is stored in the IHDR
			return m_apFrames[0]->GetPixels();
		}
	}

	return m_pixels;
}

// Ends uncompression and unfiltering
bool Png::EndImageDataProcessing()
{
	int nZRet = m_ZStream.InflateEnd();
	if( nZRet != ZStream::Z_OK )
	{
		m_nLastError = errInflateErr;
		return false;
	}
	// Reset image data chunk counters
	m_nCounter_IDAT = 0;
	m_nCounter_fdAT = 0;

	// Check that we output an expected byte count
	if( m_nOutputOffset < uint32(m_idiCurrent.uncompressedDataSize) )
	{
		m_nLastError = errImageDataMissing;
		return false;
	}

	bool processingOk = false;
	if( m_IHDR.interlaceMethod == 1 )
	{
		processingOk = UnfilterAndUninterlace();
	}
	else
	{
		// No interlacing, just unfilter
		processingOk = UnfilterBlock(m_idiCurrent.pPixels->GetWritePtr(), 
		                             m_idiCurrent.height, m_idiCurrent.byteWidth, m_nSizeofPixel);
		if( !processingOk )
		{
			m_nLastError = errBadFilterType;  // Unknown filtering type
		}
	}
	if( processingOk )
	{
		// Compute final size of the pixel buffer
		int finalSize = m_idiCurrent.height * m_idiCurrent.byteWidth;
		m_idiCurrent.pPixels->SetSize(finalSize);
	}
	return processingOk;
}

// One of the Adaptative unfiltering method
// a = left, b = top, c = top-left
uint8 Png::PaethPredictor(uint8 a, uint8 b, uint8 c)
{
	const int32 p = a + b - c;  // Initial estimation
	
	// Compute distances
	int32 pa = p - a;
	int32 pb = p - b;
	int32 pc = p - c;
	
	if( pa < 0 )
	{
		pa = -pa;
	}

	if( pb < 0 )
	{
		pb = -pb;
	}

	if( pc < 0 )
	{
		pc = -pc;
	}

	if( pa <= pb && pa <= pc )
	{
		return a;
	}

	if( pb <= pc )
	{
		return b;
	}
	return c;
}

bool Png::UnfilterAndUninterlace()
{
	// Keep the previous data to be used as input in the unfiltering process
	Buffer oldBuffer = *m_idiCurrent.pPixels;
	m_idiCurrent.pPixels->SetSize(0);
	
	// *pPixels is now empty, we reallocate it as it will be used as output in the unfiltering process
	// m_nUncompressedDataSize is modified too
	if( !AllocateImageBuffer(false) )
	{
		return false;
	}

	uint8* const pInterlaced = oldBuffer.GetWritePtr();
	const int32  interlacedSize = oldBuffer.GetSize(); interlacedSize; // Used in debug mode

	uint8* const pResult = m_idiCurrent.pPixels->GetWritePtr();
	const int32  resultSize = m_idiCurrent.uncompressedDataSize;

	static const int32 aStartingRow[7] = { 0, 0, 4, 0, 2, 0, 1 };
	static const int32 aStartingCol[7] = { 0, 4, 0, 2, 0, 1, 0 };
	static const int32 aRowIncrement[7] = { 8, 8, 8, 4, 4, 2, 2 };
	static const int32 aColIncrement[7] = { 8, 8, 4, 4, 2, 2, 1 };

	///////////////////////////////////////////////////////////////////////////
	// Masks arrays for pixel sizes < 8
	static const uint8 aMask4[2] = { 0xf0, 0x0f };
	static const uint8 aShift4[2] = { 4, 0 };

	static const uint8 aMask2[4] = { 0xc0, 0x30, 0x0c, 0x03 };
	static const uint8 aShift2[4] = { 6, 4, 2, 0 };

	static const uint8 aMask1[8] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };
	static const uint8 aShift1[8] = { 7, 6, 5, 4, 3, 2, 1, 0 };
	
	const uint8* paMask = null;
	const uint8* paShift = null;
	int32 dstShift = 0;

	if( m_nSizeofPixelInBits < 8 )
	{
		Memory::Zero(pResult, resultSize);
		if( m_nSizeofPixelInBits == 4 )
		{
			paMask = aMask4;
			paShift = aShift4;
			dstShift = 1;
		}
		else if( m_nSizeofPixelInBits == 2 )
		{
			paMask = aMask2;
			paShift = aShift2;
			dstShift = 2;
		}
		else if( m_nSizeofPixelInBits == 1 )
		{
			paMask = aMask1;
			paShift = aShift1;
			dstShift = 3;
		}
	}
	///////////////////////////////////////////////////////////////////////////
	const int32 width = m_idiCurrent.width;
	const int32 height = m_idiCurrent.height;
	const int32 byteWidth = m_idiCurrent.byteWidth;

	// Current byte position in the interlaced buffer
	int32 srcIndex = 0;
	
	// Sub index for pixel < 8 bits
	int32 srcSubIndex = 0;
	int32 maxSrcSubIndex = (8 / m_nSizeofPixelInBits) - 1;
	
	for(int8 iPass = 0; iPass < 7; ++iPass)
	{
		//////////////////////////////////////////////////////////////////////////////
		// Unfilter the pass-pixel-block
		int32 localRowCount, pixelBytesPerRow;
		GetPassSize(iPass, localRowCount, pixelBytesPerRow, width, height, m_nSizeofPixelInBits);

		if( !UnfilterBlock(pInterlaced + srcIndex, localRowCount, pixelBytesPerRow, m_nSizeofPixel) )
		{
			m_nLastError = errBadFilterType;  // Unknown filtering type
			return false;
		}
		//////////////////////////////////////////////////////////////////////////////
		if( localRowCount == 0 || pixelBytesPerRow == 0 )
		{
			// Empty pass
			continue;
		}

		// Unfilter the pass-pixel-block
		int32 row = aStartingRow[iPass];
		while(row < height)
		{
			int32 col = aStartingCol[iPass];
			if( col >= width )
			{
				// End of row reached
				break;
			}

			// Some data is available for this pass,
			const int32 dstRowOffset = row * byteWidth;
			do
			{
				ASSERT(srcIndex < interlacedSize);
				if( m_nSizeofPixelInBits > 8 )
				{
					const int32 dstOffset = dstRowOffset + col * m_nSizeofPixel;
					for(int i = 0; i < m_nSizeofPixel; ++i)
					{
						pResult[dstOffset + i] = pInterlaced[srcIndex];
						srcIndex++;
					}
				}
				else if( m_nSizeofPixelInBits == 8 )
				{
					pResult[dstRowOffset + col] = pInterlaced[srcIndex];
					srcIndex++;
				}
				else if( m_nSizeofPixelInBits < 8 )
				{
					////////////////////////////////////
					// Extract the index value
					uint8 byte = pInterlaced[srcIndex];
					byte &= paMask[srcSubIndex];
					byte >>= paShift[srcSubIndex];
					////////////////////////////////////
					
					////////////////////////////////////
					int32 dstOffset = dstRowOffset + (col >> dstShift);
					int32 dstSubIndex = col & maxSrcSubIndex;
					byte <<= paShift[dstSubIndex];

					pResult[dstOffset] |= byte;

					srcSubIndex++;

					if( srcSubIndex > maxSrcSubIndex )
					{
						// End of byte
						srcIndex++;
						srcSubIndex = 0;
					}
				}
				
				col += aColIncrement[iPass];
			}
			while(col < width);

			if( srcSubIndex != 0 )
			{
				// Padding
				srcIndex++;
				srcSubIndex = 0; // Reset the sub pixel index
			}

			row += aRowIncrement[iPass];
		}
		// Jump to the next pass bytes (only if pass not empy)
		// (localRowCount == number of filtering bytes for this pass)
		// so we jump over the bytes which gave the sub-filtering info
		srcIndex += localRowCount;
	}

	return true;
}

int32 Png::ComputeInterlacedSize(int32 width, int32 height, int32 sizeofPixelInBits)
{
	int32 totalByteCount = 0;

	for(int8 iPass = 0; iPass < 7; ++iPass)
	{
		int32 rowCount, pixelBytesPerRow;
		GetPassSize(iPass, rowCount, pixelBytesPerRow, width, height, sizeofPixelInBits);

		if( pixelBytesPerRow > 0 )
		{
			// + 1 for each row to count the filtering byte
			totalByteCount += rowCount * (pixelBytesPerRow + 1);
		}
	}
	return totalByteCount;
}

// Gets the size of a block for a given pass
void Png::GetPassSize(int8 iPass, int32& rowCount, int32& pixelBytesPerRow,
					   int32 width, int32 height, int32 sizeofPixelInBits)
{
	ASSERT(0 <= iPass && iPass <= 6);

	static const int32 anAdders[8] =   { 7, 7, 3, 3, 1, 1, 0, 0 };
	static const int32 anDividers[8] = { 8, 8, 8, 4, 4, 2, 2, 1 };

	int32 pixelsPerRow = (width + anAdders[iPass + 1]) / anDividers[iPass + 1];
	int32 pixelBitsPerRow = pixelsPerRow * sizeofPixelInBits;
	
	pixelBytesPerRow = pixelBitsPerRow / 8 + ((pixelBitsPerRow % 8) > 0 ? 1 : 0);

	rowCount = (height + anAdders[iPass]) / anDividers[iPass];
}

// pBlock : intput/ouput
// input format : each row starts with a filter type byte
// Fxxxxxx
// Fxxxxxx
// Fxxxxxx
// output format : linear version of the picture, without filter type byte
// yyyyyy
// yyyyyy
// yyyyyy
// xxx
// pixelBytesPerRow : number of bytes for one row of the picture, NOT counting the filter type byte
// bytesPerPixel : rounded to 1 if the pixel depth is less than 8
bool Png::UnfilterBlock(uint8* pBlock, int32 rowCount, int32 pixelBytesPerRow, int32 bytesPerPixel)
{
	if( pixelBytesPerRow == 0 )
		return true;

	// pRow points on the start of the filtered row
	// pUnfilteredRow points on the start of the unfiltered row
	// Then, for each row, pUnfilteredRow loses one byte in comparison with pRow

	uint8* pRow = pBlock;
	uint8* pUnfilteredRow = pBlock;

	// Create constant data for the loop
	const int32 sizeofPixel = bytesPerPixel;
	const int32 byteWidth = pixelBytesPerRow;

	for(int iRow = 0; iRow < rowCount; ++iRow)
	{
		// Initialize the index (iByte) of the current byte in the row
		int32 iByte = 0;
		
		// Get the filtering method used for this row
		const uint8 method = pRow[0];
				
		// Jump over the filtering method byte
		pRow++;
		
		switch(method)
		{
		case 0:
			// Method "None"
			{
				int counter = byteWidth;
				do
				{
					const uint8 pixel = pRow[0];
					pRow++;
					
					pUnfilteredRow[0] = pixel;
					pUnfilteredRow++;
					
				}
				while(--counter > 0);
			}
			break;
		case 1:
			// Method "Sub"
			for(;iByte < sizeofPixel; iByte++)
			{
				pUnfilteredRow[0] = pRow[0];
				pUnfilteredRow++;
				pRow++;
			}
			for(;iByte < byteWidth; iByte++)
			{
				uint8 sub = pRow[0];
				uint8 left = pUnfilteredRow[0 - sizeofPixel];
				uint8 res = uint8(sub + left);
				pUnfilteredRow[0] = res;
				pUnfilteredRow++;
				pRow++;
			}
			break;
		case 2:
			// Method "Up"
			if( iRow == 0 )
			{
				for(;iByte < byteWidth; iByte++)
				{
					uint8 nUp = pRow[0];
					uint8 above = 0;
					uint8 res = uint8(nUp + above);
					pUnfilteredRow[0] = res;
					pUnfilteredRow++;
					pRow++;
				}
			}
			else
			{
				for(;iByte < byteWidth; iByte++)
				{
					uint8 nUp = pRow[0];
					uint8 above = pUnfilteredRow[0 - byteWidth];
					uint8 res = uint8(nUp + above);
					pUnfilteredRow[0] = res;
					pUnfilteredRow++;
					pRow++;
				}
			}
			break;
		case 3:
			// Method "Average"
			if( iRow == 0 )
			{
				for(; iByte < sizeofPixel; iByte++)
				{
					uint8 average = pRow[0];
					uint8 left = 0;
					uint8 above = 0;
					uint8 res = uint8(average + (left + above) / 2);
					pUnfilteredRow[0] = res;
					pUnfilteredRow++;
					pRow++;
				}
				for(;iByte < byteWidth; iByte++)
				{
					uint8 average = pRow[0];
					uint8 left = pUnfilteredRow[0 - sizeofPixel];
					uint8 above = 0;
					uint8 res = uint8(average + (left + above) / 2);
					pUnfilteredRow[0] = res;
					pUnfilteredRow++;
					pRow++;
				}
			}
			else
			{
				for(; iByte < sizeofPixel; iByte++)
				{
					uint8 average = pRow[0];
					uint8 left = 0;
					uint8 above = pUnfilteredRow[0 - byteWidth];
					uint8 res = uint8(average + (left + above) / 2);
					pUnfilteredRow[0] = res;
					pUnfilteredRow++;
					pRow++;
				}
				for(;iByte < byteWidth; iByte++)
				{
					uint8 average = pRow[0];
					uint8 left = pUnfilteredRow[0 - sizeofPixel];
					uint8 above = pUnfilteredRow[0 - byteWidth];
					uint8 res = uint8(average + (left + above) / 2);
					pUnfilteredRow[0] = res;
					pUnfilteredRow++;
					pRow++;
				}
			}
			break;
		case 4:
			// Method "Paeth"
			if( iRow == 0 )
			{
				for(;iByte < sizeofPixel; iByte++)
				{
					uint8 paeth = pRow[0];
					uint8 left = 0;
					uint8 above = 0;
					uint8 upperLeft = 0;
					uint8 res = uint8(paeth + PaethPredictor(left, above, upperLeft));
					pUnfilteredRow[0] = res;
					pUnfilteredRow++;
					pRow++;
				}
				for(;iByte < byteWidth; iByte++)
				{
					uint8 paeth = pRow[0];
					uint8 left = pUnfilteredRow[0 - sizeofPixel];
					uint8 above = 0;
					uint8 upperLeft = 0;
					uint8 res = uint8(paeth + PaethPredictor(left, above, upperLeft));
					pUnfilteredRow[0] = res;
					pUnfilteredRow++;
					pRow++;
				}
			}
			else
			{
				for(;iByte < sizeofPixel; iByte++)
				{
					uint8 paeth = pRow[0];
					uint8 left = 0;
					uint8 above = pUnfilteredRow[0 - byteWidth];
					uint8 upperLeft = 0;
					uint8 res = uint8(paeth + PaethPredictor(left, above, upperLeft));
					pUnfilteredRow[0] = res;
					pUnfilteredRow++;
					pRow++;
				}
				for(;iByte < byteWidth; iByte++)
				{
					uint8 paeth = pRow[0];
					uint8 left = pUnfilteredRow[0 - sizeofPixel];
					uint8 above = pUnfilteredRow[0 - byteWidth];
					uint8 upperLeft = pUnfilteredRow[0 - sizeofPixel - byteWidth];
					uint8 res = uint8(paeth + PaethPredictor(left, above, upperLeft));
					pUnfilteredRow[0] = res;
					pUnfilteredRow++;
					pRow++;
				}
			}
			break;
		default:
			return false;
		}
	}
	return true;
}

bool Png::Dump(const String& strFileName, const PngDumpSettings& ds)
{
	File file;
	if( !file.Open(strFileName, File::modeWrite) )
	{
		return false;
	}
	return Dump(file, ds);
}

// pBlock points on a buffer which already has room for the sub-filtering byte info given
// at the beginning of each row.
bool Png::FilterBlock(uint8* const pBlock, int32 rowCount, int32 pixelBytesPerRow, int32 bytesPerPixel)
{
	// Add the byte size for the filtering type
	const int32 totalBytesPerRow = pixelBytesPerRow + 1;
	///////////////////////////////////////////////////////////////////////////

	//const int32 filteredBufSize = filteredRowSize * rowCount;

	// Allocate buffers for 5 possible filtered rows
	ByteArray rbFilteredRows;
	if( !rbFilteredRows.SetSize(pixelBytesPerRow * 5) )
	{
		// Not enough memory
		return false;
	}

	uint8* const pFilteredRows = rbFilteredRows.GetPtr();

	uint8* const pDstNone = pFilteredRows;
	uint8* const pDstSub = pDstNone + pixelBytesPerRow;
	uint8* const pDstUp = pDstSub + pixelBytesPerRow;
	uint8* const pDstAverage = pDstUp + pixelBytesPerRow;
	uint8* const pDstPaeth = pDstAverage + pixelBytesPerRow;

	uint8* const apTmpRows[5] = {pDstNone, pDstSub, pDstUp, pDstAverage, pDstPaeth };
	
	////////////////////////////////////////////////////////////////////////
	// The filtering process must be bottom-up
	// TODO : this kills the CPU cache we should sometimes prefetch some data to speed-up the process

	uint8* pRow = pBlock + totalBytesPerRow * rowCount - totalBytesPerRow;

	for(int iRow = rowCount - 1; iRow >= 0; --iRow)
	{
		uint8* pMethod = pRow;

		// Jump over the method byte
		pRow++;

		// Filtering method used
		uint8 method = 0;

		///////////////////////////////////////
		// Try all of the 5 sub-filter types //
		///////////////////////////////////////

		// Filtering = None (0)
		for(int iByte = 0; iByte < pixelBytesPerRow; ++iByte)
		{
			const uint8 byteCurrent = pRow[iByte];
			const uint8 none = byteCurrent;
			
			pDstNone[iByte] = none;
		}

		// Try other types of filtering

		// Filtering = Sub (1)
		for(int iByte = 0; iByte < pixelBytesPerRow; ++iByte)
		{
			uint8 byteLeft;
			if( iByte < bytesPerPixel )
			{
				byteLeft = 0;
			}
			else
			{
				byteLeft = pRow[iByte - bytesPerPixel];
			}

			const uint8 byteCurrent = pRow[iByte];
			const uint8 sub = uint8(byteCurrent - byteLeft);
			
			pDstSub[iByte] = sub;
		}

		// Filtering = Up (2)
		for(int iByte = 0; iByte < pixelBytesPerRow; ++iByte)
		{
			uint8 byteUp;
			if( iRow == 0 )
			{
				byteUp = 0;
			}
			else
			{
				byteUp = pRow[-totalBytesPerRow + iByte];
			}

			const uint8 byteCurrent = pRow[iByte];
			const uint8 nUp = uint8(byteCurrent - byteUp);
			
			pDstUp[iByte] = nUp;
		}

		// Filtering = Average (3)
		for(int iByte = 0; iByte < pixelBytesPerRow; ++iByte)
		{
			uint8 byteUp;
			if( iRow == 0 )
			{
				byteUp = 0;
			}
			else
			{
				byteUp = pRow[-totalBytesPerRow + iByte];
			}

			uint8 byteLeft;
			if( iByte < bytesPerPixel )
			{
				byteLeft = 0;
			}
			else
			{
				byteLeft = pRow[iByte - bytesPerPixel];
			}

			const uint8 byteCurrent = pRow[iByte];
			const uint8 average = uint8(byteCurrent - (byteLeft + byteUp) / 2);
			
			pDstAverage[iByte] = average;
		}

		// Filtering = Paeth (4)
		for(int iByte = 0; iByte < pixelBytesPerRow; ++iByte)
		{
			uint8 byteUp;
			if( iRow == 0 )
			{
				byteUp = 0;
			}
			else
			{
				byteUp = pRow[-totalBytesPerRow + iByte];
			}

			uint8 byteLeft;
			if( iByte < bytesPerPixel )
			{
				byteLeft = 0;
			}
			else
			{
				byteLeft = pRow[iByte - bytesPerPixel];
			}

			uint8 byteUpLeft;
			if( iByte < bytesPerPixel || iRow == 0 )
			{
				byteUpLeft = 0;
			}
			else
			{
				byteUpLeft = pRow[-totalBytesPerRow + iByte - bytesPerPixel];
			}

			const uint8 byteCurrent = pRow[iByte];
			const uint8 paeth = uint8(byteCurrent - PaethPredictor(byteLeft, byteUp, byteUpLeft));
			
			pDstPaeth[iByte] = paeth;
		}

		//////////////////////////////////////////////////////////////////////////////////
		// Choose the best filtered row of the 5

		uint32 smallest = MAX_UINT32;
		method = 0; // Index of the method to use. If all have a sum of 0, choose the method 0

		for(uint8 i = 0; i < 5; ++i)
		{
			uint32 sum = 0;
			for(int iByte = 0; iByte < pixelBytesPerRow; ++iByte)
			{
				uint8 byte = apTmpRows[i][iByte];

				if( byte >= 128 )
				{
					byte = uint8(256 - byte);
				}
			
				sum += byte;
			}

			if( sum < smallest )
			{
				smallest = sum;
				method = i;
			}
		}

		///////////////////////////////////////////////////////////////////////
		// Copy the wanted filtered row to the result buffer
		pMethod[0] = method;
		Memory::Copy(pRow, apTmpRows[method], pixelBytesPerRow);

		pRow -= (1 + totalBytesPerRow);
	}
	////////////////////////////////////////////////////////////////////////
	return true;
}

// pSrc : pointer to a linear buffer of pixels
bool Png::InterlaceAndFilter(uint8* pDst, const uint8* pSrc,
							  const int32 srcWidth, const int32 srcHeight,
							  const int32 srcPixelBytesPerRow,
							  int32 sizeofPixelInBits, bool bDoFiltering)
{
	static const int32 aStartingRow[7] =  { 0, 0, 4, 0, 2, 0, 1 };
	static const int32 aRowIncrement[7] = { 8, 8, 8, 4, 4, 2, 2 };

	static const int32 aStartingCol[7] =  { 0, 4, 0, 2, 0, 1, 0 };
	static const int32 aColIncrement[7] = { 8, 8, 4, 4, 2, 2, 1 };

	///////////////////////////////////////////////////////////////////////////
	// Masks arrays for pixel sizes < 8
	static const uint8 aMask4[2] = { 0xf0, 0x0f };
	static const uint8 aShift4[2] = { 4, 0 };

	static const uint8 aMask2[4] = { 0xc0, 0x30, 0x0c, 0x03 };
	static const uint8 aShift2[4] = { 6, 4, 2, 0 };

	static const uint8 aMask1[8] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };
	static const uint8 aShift1[8] = { 7, 6, 5, 4, 3, 2, 1, 0 };
	
	const uint8* paMask = null;
	const uint8* paShift = null;
	int32 srcShift = 0;

	if( sizeofPixelInBits < 8 )
	{
		if( sizeofPixelInBits == 4 )
		{
			paMask = aMask4;
			paShift = aShift4;
			srcShift = 1;
		}
		else if( sizeofPixelInBits == 2 )
		{
			paMask = aMask2;
			paShift = aShift2;
			srcShift = 2;
		}
		else if( sizeofPixelInBits == 1 )
		{
			paMask = aMask1;
			paShift = aShift1;
			srcShift = 3;
		}
	}

	///////////////////////////////////////////////////////////////////////////
	// The Png file is made of several sub-images, going more and more detailed

	const int32 sizeofPixelInBytes = (sizeofPixelInBits + 7 ) / 8;
	int32 dstIndex = 0; // The byte index in the destination buffer
	
	// Sub index for pixel < 8 bits
	int32 dstSubIndex = 0;
	int32 maxDstSubIndex = (8 / sizeofPixelInBits) - 1;

	for(int8 iPass = 0; iPass < 7; ++iPass)
	{
		//////////////////////////////////////////////////////////////////////////////
		uint8* const pDstPass = pDst + dstIndex;

		int32 localRowCount, localPixelBytesPerRow;
		GetPassSize(iPass, localRowCount, localPixelBytesPerRow, srcWidth, srcHeight, sizeofPixelInBits);
	
		//////////////////////////////////////////////////////////////////////////////
		if( localRowCount == 0 || localPixelBytesPerRow == 0 )
		{
			// Empty pass
			continue;
		}

		if( sizeofPixelInBits < 8 )
		{
			// As for small bit depth we perform binary OR to set the pixels, the destination buffer
			// must be previously filled with 0s
			Memory::Zero(pDstPass, (localPixelBytesPerRow + 1) * localRowCount);
		}

		// Prepare the pass-pixel-block by getting pixels around from the original buffer
		int32 row = aStartingRow[iPass];
		while(row < srcHeight)
		{
			int32 col = aStartingCol[iPass];
			if( col >= srcWidth )
			{
				// End of row reached
				break;
			}

			// Insert an empty byte for the sub-filtering value
			// (We will filter the whole pass buffer later)
			pDst[dstIndex] = 0;
			dstIndex++;

			const int32 srcRowOffset = row * srcPixelBytesPerRow;
			do
			{
				if( sizeofPixelInBits > 8 )
				{
					const int32 srcOffset = srcRowOffset + col * sizeofPixelInBytes;
					for(int i = 0; i < sizeofPixelInBytes; ++i)
					{
						pDst[dstIndex] = pSrc[srcOffset + i];
						dstIndex++;
					}
				}
				else if( sizeofPixelInBits == 8 )
				{
					pDst[dstIndex] = pSrc[srcRowOffset + col];
					dstIndex++;
				}
				else if( sizeofPixelInBits < 8 )
				{
					////////////////////////////////////////////////////
					// Extract the pixel from the source
					int32 srcOffset = srcRowOffset + (col >> srcShift);
					uint8 byte = pSrc[srcOffset];
					int32 srcSubIndex = col & maxDstSubIndex;
					
					byte &= paMask[srcSubIndex];
					byte >>= paShift[srcSubIndex];
										
					// byte now contains the source pixel value
					////////////////////////////////////////////////////

					byte <<= paShift[dstSubIndex];

					pDst[dstIndex] |= byte;

					dstSubIndex++;

					if( dstSubIndex > maxDstSubIndex )
					{
						// End of byte
						dstIndex++;
						dstSubIndex = 0;
					}
				}
				
				col += aColIncrement[iPass];
			}
			while(col < srcWidth);

			if( dstSubIndex != 0 )
			{
				// Padding
				dstIndex++;
				dstSubIndex = 0; // Reset the sub pixel index
			}

			row += aRowIncrement[iPass];
		}
		
		// End of the pass !
		if( bDoFiltering )
		{
			if( !FilterBlock(pDstPass, localRowCount, localPixelBytesPerRow, sizeofPixelInBytes) )
			{
				// Not enough memory
				return false;
			}
		}
	}
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
bool Png::CreateImageData(const uint8* pSrc, int32 width, int32 height, const PngDumpSettings& ds, ByteArray& abImageData)
{
	PixelFormat epf = ds.pixelFormat;

	////////////////////////////////////////////////////////////
	ByteArray rbSwapped;
	if( epf == pf32bppBgra )
	{
		const int32 pixelBytesPerRow = ImageFormat::ComputeByteWidth(ds.pixelFormat, width);
		const int32 srcSize = pixelBytesPerRow * height;

		if( !rbSwapped.SetSize(srcSize) )
		{
			// Not enough memory
			return false;
		}
		uint8* pNewBuffer = rbSwapped.GetPtr();

		// Should swap bytes as PNG does not support BGRA8
		// We convert the pixels into RGBA8 format
		const int32 pixelCount = srcSize / 4;
		for(int iCopy = 0; iCopy < pixelCount; ++iCopy)
		{
			uint8 b = pSrc[4 * iCopy + 0];
			uint8 g = pSrc[4 * iCopy + 1];
			uint8 r = pSrc[4 * iCopy + 2];
			uint8 a = pSrc[4 * iCopy + 3];
			
			pNewBuffer[4 * iCopy + 0] = r;
			pNewBuffer[4 * iCopy + 1] = g;
			pNewBuffer[4 * iCopy + 2] = b;
			pNewBuffer[4 * iCopy + 3] = a;
		}

		pSrc = pNewBuffer;
		epf = pf32bppRgba;
	}
	////////////////////////////////////////////////////////////

	const int32 pixelBytesPerRow = ImageFormat::ComputeByteWidth(epf, width);
	const int32 srcSize = pixelBytesPerRow * height;

	/////////////////////////////////////
	const int32 bitsPerPixel = SizeofPixelInBits(epf);

	int strategy = ZStream::Z_DEFAULT_STRATEGY;
	if( ds.zlibStrategy == PngDumpSettings::zlibStrategyGuess )
	{
		// Guess which strategy we should use
		if( ds.filtering != 0 )
		{
			// Usually achieves better compression with filtered images
			if( bitsPerPixel <= 8 )
			{
				// But for low depths only
				strategy = ZStream::Z_FILTERED;
			}
		}
	}
	else
	{
		if( ds.zlibStrategy == PngDumpSettings::zlibStrategyFilter )
		{
			strategy = ZStream::Z_FILTERED;
		}
	}

	// The buffer that will be given to the zlib routines
	ByteArray abBufferToCompress;

	const int32 bytesPerPixel = (bitsPerPixel + 7 ) / 8;

	// If special sub-filtering is not requested do not try other methods
	bool bDoFiltering = false;
	if( ds.filtering != 0 )
	{
		bDoFiltering = true;
	}
	
	int32 bufferToCompressSize = 0;
	uint8* pBufferToCompress = null;

	if( ds.interlaced )
	{
		bufferToCompressSize = ComputeInterlacedSize(width, height, bitsPerPixel);
		if( !abBufferToCompress.SetSize(bufferToCompressSize) )
		{
			// Not enough memory
			return false;
		}

		pBufferToCompress = abBufferToCompress.GetPtr();
		if( !InterlaceAndFilter(pBufferToCompress, pSrc, width, height, pixelBytesPerRow, bitsPerPixel, bDoFiltering) )
		{
			// Not enough memory
			return false;
		}
	}
	else
	{
		/////////////////////////////////////////////////////////////
		// Create a new buffer from the image with a blank byte at each
		// row beginning (the sub-filtering method)

		bufferToCompressSize = srcSize + height;
		if( !abBufferToCompress.SetSize(bufferToCompressSize) )
		{
			// Not enough memory
			return false;
		}
		pBufferToCompress = abBufferToCompress.GetPtr();

		uint8* pCurNewByte = pBufferToCompress;
		const uint8* pCurSrcByte = pSrc;

		for(int iRow = 0; iRow < height; ++iRow)
		{
			pCurNewByte[0] = 0;
			pCurNewByte++;

			for(int iByte = 0; iByte < pixelBytesPerRow; ++iByte)
			{
				pCurNewByte[0] = pCurSrcByte[0];
				pCurSrcByte++;
				pCurNewByte++;
			}
		}

		////////////////////////////////////////////////////////////////
		// Apply filtering
		if( bDoFiltering )
		{
			FilterBlock(pBufferToCompress, height, pixelBytesPerRow, bytesPerPixel);
		}
	}

	// ZLib documentation about compress() :
	// Upon entry, destLen is the total size of the destination buffer,
	// which must be at least 0.1% larger than sourceLen plus 12 bytes.
	uint32 compressedBufferSize = 12 + bufferToCompressSize + ((bufferToCompressSize + 63) / 64);
	
	if( !abImageData.SetSize(compressedBufferSize) )
	{
		// Not enough memory
		return false;
	}
	uint8* const pCompressedBuffer = abImageData.GetPtr();

	const int32 compressionLevel = ds.zlibCompressionLevel;

	int maxWindowBits = 15;
	int memLevel = 8;

	// Sometimes it improves compression to use settings for low memory platforms,
	// that's why we have this option
	if( ds.zlibWindowBitsAndMem == PngDumpSettings::zlibWindowBitsAndMemLow )
	{
		maxWindowBits = 14;
		memLevel = 6;
	}

	int32 ret = Compress(pCompressedBuffer, &compressedBufferSize,
						pBufferToCompress, bufferToCompressSize,
						compressionLevel, strategy, maxWindowBits, memLevel);

	if( ret != 0 )
	{
		// Error in compression
		return false;
	}

	// Set the final size
	abImageData.SetSize(compressedBufferSize);

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
bool Png::WriteFrameControlChunk(const ApngFrame* pFrame, int32 apngSequenceNumber, ChunkedFile& file)
{
	file.BeginChunkWrite(chunk_fcTL);
	PngChunk_fcTL fctl = pFrame->m_fctl;

	fctl.sequenceNumber = apngSequenceNumber;
	if( file.ShouldSwapBytes() )
	{
		// Swap to big endian
		fctl.SwapBytes();
	}
	if( !file.Write(&fctl, k_nChunkDataSize_fcTL) )
	{
		return false;
	}
	if( !file.EndChunkWrite() )
	{
		return false;
	}
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
bool Png::WriteChunk_bkGD(ChunkedFile& cf, uint8 colorType, const PngChunk_bkGD& content)
{
	if( colorType == 0x03 && content.index >= 0 )
	{
		// Indexed color
		cf.BeginChunkWrite(chunk_bKGD);
		cf.Write8(content.index);
		if( !cf.EndChunkWrite() )
			return false;
	}
	else if( (colorType == 0x00 || colorType == 0x04) && content.grey >= 0 )
	{
		// Grey or Grey+Alpha
		cf.BeginChunkWrite(chunk_bKGD);
		cf.Write16(content.grey);
		if( !cf.EndChunkWrite() )
			return false;
	}
	else if( (colorType == 0x02 || colorType == 0x06) && content.red >= 0 )
	{
		// TrueColor or TrueColor+Alpha
		cf.BeginChunkWrite(chunk_bKGD);
		cf.Write16(content.red);
		cf.Write16(content.green);
		cf.Write16(content.blue);
		if( !cf.EndChunkWrite() )
			return false;
	}
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
bool Png::WriteChunk_tEXt(ChunkedFile& cf, const PngChunk_tEXt& content)
{
	if( content.keyword.IsEmpty() )
	{
		return true;
	}
	ByteArray kwBytes = content.keyword.ToBytes(TextEncoding::Iso8859_1());
	int kwLength = kwBytes.GetSize();
	if( kwLength > 79 )
	{
		kwLength = 79; // Truncate keyword, 79 is the max as per the PNG spec
	}
	ByteArray chunkBytes;
	chunkBytes.Set(kwBytes.GetPtr(), kwLength);
	chunkBytes.Add(0);
	ByteArray dataBytes = content.data.ToBytes(TextEncoding::Iso8859_1());
	chunkBytes.Add(dataBytes.GetPtr(), dataBytes.GetSize());

	cf.BeginChunkWrite(chunk_tEXt);
	cf.Write(chunkBytes.GetPtr(), chunkBytes.GetSize());
	if( !cf.EndChunkWrite() )
	{
		return false;
	}
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Dumps an image or animation in the PNG format.
//
// fileDst: File to write to
// ds:      Dump settings
//
// Returns ture upon success
///////////////////////////////////////////////////////////////////////////////////////////////////
bool Png::Dump(IFile& fileDst, const PngDumpSettings& ds)
{
	PixelFormat epf = ds.pixelFormat;
	const int32 width = ds.width;
	const int32 height = ds.height;
	const Palette& palette = ds.palette;

	if( epf == pf32bppBgra )
	{
		// Accept BGRA pixels, they will be reverted in CreateImageData()
		epf = pf32bppRgba;
	}

	uint8 colorType;
	uint8 bitsPerComponent;
	if( !GetIHDRFeaturesFromPixelFormat(epf, colorType, bitsPerComponent) )
	{
		return false;
	}

	if( (colorType & cfPalette) && palette.m_count == 0 )
	{
		return false;
	}

	/////////////////////////////////////////////////////////////////////////
	// Send the data to the stream
	
	// Start to send the png signature
	if( fileDst.Write(k_PngSignature, 8) != 8 )
	{
		return false;
	}

	// First chunk : the header	
	ChunkedFile file(fileDst);
	
	// 1 = Adam7
	uint8 interlaceMethod = ds.interlaced ? uint8(1) : uint8(0);
	
	file.BeginChunkWrite(chunk_IHDR);
	
	PngChunk_IHDR hi;
	hi.width = width;
	hi.height = height;
	hi.bitDepth = bitsPerComponent;
	hi.colorType = colorType;
	hi.compressionMethod = 0; // 0 = deflate/inflate
	hi.filterMethod = 0;      // 0 = Adaptative
	hi.interlaceMethod = interlaceMethod;
	
	if( file.ShouldSwapBytes() )
	{
		hi.SwapBytes();
	}
	
	const int32 headerSize = k_nChunkDataSize_IHDR;
	if( file.Write(&hi, headerSize) != headerSize )
	{
		return false;
	}

	if( !file.EndChunkWrite() )
	{
		return false;
	}

	if( colorType & cfPalette )
	{
		// Palettized picture
		const int32 colorCount = palette.m_count;

		if( !(0 <= colorCount && colorCount <= 256) )
		{
			// Bad number of colors
			return false;
		}

		/////////////////////////////////////////////////////////////////////////
		// Write the palette colors
		file.BeginChunkWrite(chunk_PLTE);
		
		uint8 aTmpPalette[256 * 3];
		int32 iTmpColor = 0;
		for(int i = 0; i < colorCount; ++i)
		{
			uint8 r, g, b, a;
			palette[i].ToRgba(r, g, b, a);

			aTmpPalette[iTmpColor + 0] = r;
			aTmpPalette[iTmpColor + 1] = g;
			aTmpPalette[iTmpColor + 2] = b;
			iTmpColor += 3;
		}

		file.Write( aTmpPalette, colorCount * 3);

		if( !file.EndChunkWrite() )
			return false;
	
		/////////////////////////////////////////////////////////////////////////
		// Check if we need to write a tRNS chunk
		// We don't if all alphas are 255
		// Note : an external png optimizer should sort the palette colors by their
		// alpha values so the lowest alpha is found first
		
		int32 alphaCount = 0;
		for(int i = colorCount - 1; i >= 0; --i)
		{
			if( palette[i].GetAlpha() != 255 )
			{
				alphaCount = i + 1;
				break;
			}
		}

		if( alphaCount > 0 )
		{
			// tRNS chunk needed
			uint8 aAlphas[256];
			for(int i = 0; i < alphaCount; ++i)
			{
				aAlphas[i] = palette[i].GetAlpha();
			}

			file.BeginChunkWrite(chunk_tRNS);
			
			file.Write(aAlphas, alphaCount);

			if( !file.EndChunkWrite() )
				return false;
		}
	}
	else if( colorType == 0 )
	{
		// Grey
		if( ds.useTransparentColor )
		{
			file.BeginChunkWrite(chunk_tRNS);
			file.Write16(ds.tRNS.grey);
			
			if( !file.EndChunkWrite() )
				return false;
		}
	}
	else if( colorType == 2 )
	{
		// True colors
		if( ds.useTransparentColor )
		{
			file.BeginChunkWrite(chunk_tRNS);
			file.Write16(ds.tRNS.red);
			file.Write16(ds.tRNS.green);
			file.Write16(ds.tRNS.blue);

			if( !file.EndChunkWrite() )
				return false;
		}
	}

	////////////////////////////////////////////////////
	// bKGD chunk
	if( ds.useBackgroundColor )
	{
		if( !WriteChunk_bkGD(file, colorType, ds.bkGD) )
		{
			return false;
		}
	}

	////////////////////////////////////////////////////
	// tEXt chunk
	foreach(ds.textInfos, i)
	{
		if( !WriteChunk_tEXt(file, ds.textInfos[i]) )
		{
			return false;
		}
	}

	////////////////////////////////////////////////////
	// Now it's time for image data
	const uint8* pIdatSrc = ds.pixels.GetReadPtr();

	int32 apngSequenceNumber = 0;
	int32 iFrame = 0;

	int32 frameCount = ds.frames.GetSize();
	if( frameCount > 0 )
	{
		// Aha, an animation, put the acTL chunk first
		file.BeginChunkWrite(chunk_acTL);
		if( !(file.Write32(frameCount) && file.Write32(ds.loopCount)) )
		{
			return false;
		}
		if( !file.EndChunkWrite() )
		{
			return false;
		}

		if( !ds.hasDefaultImage )
		{
			// The IDAT is part of the animation and its pixels are in apFrames[0].
			
			const ApngFrame* pFrame = ds.frames[0];

			// Check that the size matches the IHDR
			bool bOk1 = (pFrame->m_fctl.width == ds.width);
			bool bOk2 = (pFrame->m_fctl.height == ds.height);
			bool bOk3 = (pFrame->m_fctl.offsetX == 0);
			bool bOk4 = (pFrame->m_fctl.offsetY == 0);

			if( !(bOk1 && bOk2 && bOk3 && bOk4) )
			{
				// Nop
				return false;
			}

			// Write a fcTL first.
			if( !WriteFrameControlChunk(pFrame, apngSequenceNumber, file) )
			{
				return false;
			}
			apngSequenceNumber++;

			pIdatSrc = pFrame->GetPixels().GetReadPtr();
			iFrame++;
		}
	}

	/////////////////////////////////////////////////
	// Write the IDAT
	file.BeginChunkWrite(chunk_IDAT);

	ByteArray abImageData;
	if( !CreateImageData(pIdatSrc, width, height, ds, abImageData) )
	{
		return false;
	}
	if( !file.Write(abImageData.GetPtr(), abImageData.GetSize()) )
	{
		return false;
	}
	if( !file.EndChunkWrite() )
	{
		return false;
	}
	/////////////////////////////////////////////////
	
	/////////////////////////////////////////////////
	// Write remaining APNG frames
	for(; iFrame < frameCount; ++iFrame)
	{
		const ApngFrame* pFrame = ds.frames[iFrame];

		/////////////////////////
		// Write a fcTL
		if( !WriteFrameControlChunk(pFrame, apngSequenceNumber, file) )
		{
			return false;
		}
		apngSequenceNumber++;
		/////////////////////////

		/////////////////////////
		// Write the fdAT
		file.BeginChunkWrite(chunk_fdAT);

		if( !file.Write32(apngSequenceNumber) )
		{
			return false;
		}
		apngSequenceNumber++;

		const uint8* pFramePixels = pFrame->GetPixels().GetReadPtr();
		const int32 frameWidth = pFrame->GetWidth();
		const int32 frameHeight = pFrame->GetHeight();

		ByteArray abImageData;
		if( !CreateImageData(pFramePixels, frameWidth, frameHeight, ds, abImageData) )
		{
			return false;
		}
		if( !file.Write(abImageData.GetPtr(), abImageData.GetSize()) )
		{
			return false;
		}
		if( !file.EndChunkWrite() )
		{
			return false;
		}
		/////////////////////////
	}

	/////////////////////////////////////////////////
	// Last chunk : end chunk
	file.BeginChunkWrite(chunk_IEND);
	if( !file.EndChunkWrite() )
	{
		return false;
	}	
	return true;
}

int Png::Compress(uint8* pDest, uint32* pDestLen, const uint8* pSource, uint32 sourceLen, 
				  int level, int strategy, int windowBits, int memLevel)
{
	ASSERT(strategy == ZStream::Z_DEFAULT_STRATEGY || strategy == ZStream::Z_FILTERED);

	ZStream stream;
	stream.SetBuffers(pSource, sourceLen, pDest, *pDestLen);
	
	int err;
	
	uint8 method = ZStream::Z_DEFLATED;

	err = stream.DeflateInit2(level, method, windowBits, memLevel, strategy);
	if( err != ZStream::Z_OK )
		return err;
	
	err = stream.Deflate(ZStream::Z_FINISH);
	if( err != ZStream::Z_STREAM_END )
	{
		stream.DeflateEnd();
		return err == ZStream::Z_OK ? ZStream::Z_BUF_ERROR : err;
	}
	*pDestLen = stream.GetOutTotalRead();
	
	err = stream.DeflateEnd();
	return err;
}

String Png::GetLastErrorString() const
{
	switch(m_nLastError)
	{
	case Png::errNotAPngFile:
		return "Not a PNG file";

	case Png::errBadHeaderSize:
		return "Invalid header size";
		
	case Png::errBadBitDepth:
		return "Invalid bit depth";

	case Png::errBadColorType:
		return "Invalid or unsupported color format";

	case Png::errBadPixelFormat:
		return "Invalid pixel format";

	case Png::errBadCompressionMethod:
		return "Invalid or unsupported compression method";

	case Png::errBadFilterMethod:
		return "Invalid or unsupported filtering method";
		
	case Png::errBadFilterType:
		return "Bad filtering type";

	case Png::errInflateErr:
		return "Decompression error (zlib : " + String::FromAsciiSZ(m_ZStream.GetLastError()) + ")";

	case Png::errBadPicSize:
		return "Invalid image size";
	
	case Png::errNoIENDFound:
		return "No IEND chunk found at the end of the file data (file corrupted)";

	case Png::errBadInterlaceMethod:
		return "Unsupported interlacing method (Adam7 only)";

	case Png::errNotEnoughDataInGammaChunk:
		return "Not enough data in gAMA chunk";

	case Png::errNotEnoughDataInBkChunk:
		return "Not enough data in bKGD chunk";

	case Png::errIDATFoundBeforeacTL:
		return "IDAT found before acTL chunk";

	case Png::errNotEnoughDataInAnimCtlChunk:
		return "Not enough data in acTL chunk";

	case Png::errFrameCountIsZero:
		return "Frame count in acTL chunk is zero";

	case Png::errNotEnoughDataInFrameCtlChunk:
		return "Not enough data in fcTL chunk";
		
	case Png::errBadFrameControlSequenceNumber:
		return "Bad sequence number in fcTL chunk";

	case Png::errBadFrameDataSequenceNumber:
		return "Bad sequence number in fdAT chunk";

	case Png::errIDATNotFound:
		return "No IDAT chunk found";

	case Png::errIDATfcTLNotConsistentWithIHDR:
		return "IDAT's fcTL mismatch with IHDR";

	case Png::erracTLMissing:
		return "acTL chunk missing";

	case Png::erracTLRepeated:
		return "acTL chunk repeated";
	
	case Png::errMissingfcTLBeforefdAT:
		return "fcTL chunk missing before fdAT";

	case Png::errUnexpectedChunk:
		{
			wchar szChunk[6];
			szChunk[0] = (m_nLastUnexpectedChunk >> 24) & 0x00ff;
			szChunk[1] = (m_nLastUnexpectedChunk >> 16) & 0x00ff;
			szChunk[2] = (m_nLastUnexpectedChunk >> 8) & 0x00ff;
			szChunk[3] = (m_nLastUnexpectedChunk) & 0x00ff;
			szChunk[4] = 0;
			return "Unexpected chunk: " + String(szChunk);
		}
		
	case Png::errFrameCountMismatch:
		return "Frame count mismatch with acTL";
	case Png::errFrameCountOutsideRange:
		return "Frame count outside range";
	case Png::errImageDataMissing:
		return "Image data missing";
	case Png::errTextMissingSeparator:
		return "Keyword/data separator missing in tEXt";
	case Png::errTextBadKeywordLength:
		return "Bad tEXt keyword length";
	}

	return ImageFormat::GetLastErrorString();
}

bool Png::IsAnimated() const
{
	// In GIF, everything is a frame, not with PNG, as there is the idea of default image
	if( HasDefaultImage() )
	{
		if( m_apFrames.GetSize() == 0 )
		{
			// No frame at all -> not animated
			return false;
		}
		// One default image + one frame at least
		return true;
	}

	// No default image
	if( m_apFrames.GetSize() == 1 )
	{
		// One single frame -> not animated
		return false;
	}
	// No default image + two frames at least
	return true;
}

bool Png::HasDefaultImage() const
{
	return !m_pixels.IsEmpty();
}

int32 Png::GetFrameCount() const
{
	return m_apFrames.GetSize();
}

const AnimFrame* Png::GetAnimFrame(int index) const
{
	int32 frameCount = m_apFrames.GetSize();
	if( !(0 <= index && index < frameCount) )
	{
		return null;
	}
	return m_apFrames[index];
}

int32 Png::GetLoopCount() const
{
	return m_acTL.loopCount;
}
