///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2011 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#ifndef CHUSTD_PNG_H
#define CHUSTD_PNG_H

#include "ImageFormat.h"
#include "ChunkedFile.h"

#include "ZStream.h"
#include "PtrArray.h"
#include "Buffer.h"

namespace chustd {

///////////////////////////////////////////////////////////
struct PngChunk_IHDR
{
	int32  width;
	int32  height;
	uint8  bitDepth;          // Bit count for each RGB component or for each index
	uint8  colorType;         // Color flags: default: grey, flags: 0x1=palette, 0x2=color, 0x4=alpha
	uint8  compressionMethod; // 0=deflate/inflate (zlib)
	uint8  filterMethod;      // 0=Adaptative: 5 types=(none,sub,up,average,paeth)
	uint8  interlaceMethod;   // 0=no interlace, 1=Adam7

	PngChunk_IHDR() { Clear(); }
	void Clear()
	{
		chustd::Memory::Zero(this, sizeof(PngChunk_IHDR));
	}
	void SwapBytes()
	{
		width = IFile::Swap32(width);
		height = IFile::Swap32(height);
	}
};

///////////////////////////////////////////////////////////
struct PngChunk_tRNS
{
	uint16  grey;
	uint16  red;
	uint16  green;
	uint16  blue;

	PngChunk_tRNS() { Clear(); }

	void Clear()
	{
		grey = 0;
		red = 0;
		green = 0;
		blue = 0;
	}
};

///////////////////////////////////////////////////////////
struct PngChunk_bkGD
{
	uint8   index;
	uint16  grey;
	uint16  red;
	uint16  green;
	uint16  blue;

	PngChunk_bkGD() { Clear(); }

	void Clear()
	{
		index = 0;
		grey = 0;
		red = 0;
		green = 0;
		blue = 0;
	}
};

///////////////////////////////////////////////////////////
struct PngChunk_tEXt
{
	String keyword;
	String data;
};

///////////////////////////////////////////////////////////
struct PngChunk_fcTL
{
	int32 sequenceNumber;
	int32 width;
	int32 height;
	int32 offsetX;
	int32 offsetY;
	uint16 delayFracNumerator;
	uint16 delayFracDenominator;
	uint8  disposal; // What to do before rendering the next frame
	uint8  blending;

	PngChunk_fcTL() { Clear(); }
	void Clear()
	{
		sequenceNumber = 0;
		width = 0;
		height = 0;
		offsetX = 0;
		offsetY = 0;
		delayFracNumerator = 0;
		delayFracDenominator = 100;
		disposal = 0;
		blending = 0;
	}
	void SwapBytes()
	{
		sequenceNumber = IFile::Swap32(sequenceNumber);
		width = IFile::Swap32(width);
		height = IFile::Swap32(height);
		offsetX = IFile::Swap32(offsetX);
		offsetY = IFile::Swap32(offsetY);
		delayFracNumerator = IFile::Swap16(delayFracNumerator);
		delayFracDenominator = IFile::Swap16(delayFracDenominator);
	}
};

class ApngFrame;

///////////////////////////////////////////////////////////////////////////////
struct PngDumpSettings
{
public:
	Buffer         pixels;
	Palette        palette;
	int32          width;
	int32          height;
	PixelFormat    pixelFormat;

	enum ZLibOption
	{ 
		zlibStrategyGuess = 0x00,   // Default
		zlibStrategyDefault = 0x01,
		zlibStrategyFilter = 0x02,
	
		zlibWindowBitsAndMemHigh = 0x00, // Default
		zlibWindowBitsAndMemLow = 0x10   // Can sometimes improve compression
	};

	uint8       zlibCompressionLevel; // [1..9], default = 6
	uint8       zlibStrategy;
	uint8       zlibWindowBitsAndMem;
	
	bool          useTransparentColor; // Set to true to use the following members
	PngChunk_tRNS tRNS;

	bool          useBackgroundColor;
	PngChunk_bkGD bkGD;
	
	Array<PngChunk_tEXt> textInfos;

	// 0 = Adaptative filtering with "None" set for each scanline
	// 1 = Adaptative filtering with various sub-method set for each scanline
	uint8       filtering;    // Default = 1
	bool        interlaced;    // true to save an interlaced png

	////////////////////////
	// Animation

	// true  = Use pixels in m_pBuffer to create the default image.
	// false = m_pBuffer can be null, the IDAT will be part of the animation, 
	//         but there must be one frame at least in m_apFrames.
	bool   hasDefaultImage;
	uint32 loopCount;
	PtrArray<ApngFrame> frames;

	PngDumpSettings()
	{
		width = 0;
		height = 0;
		pixelFormat = pfUnknown;
		zlibCompressionLevel = 6;
		zlibStrategy = zlibStrategyGuess;
		zlibWindowBitsAndMem = zlibWindowBitsAndMemHigh;
		
		useTransparentColor = false;
		useBackgroundColor = false;

		filtering = 1;
		interlaced = false;

		hasDefaultImage = false;
		loopCount = 0;
	}
};

//////////////////////////////////////////////////////////////////////////////////////
// This class is used to read and write PNG pictures
class Png : public ImageFormat
{
public:
	// Possible errors for m_nLastError
	enum LastError
	{
		errNotAPngFile = firstErrEnumDerived, // Not a png file (mismatch in header)
		errBadHeaderSize,      // Invalid header size
		errBadBitDepth,        // Invalid bit depth (should be 1, 2, 4, 8 or 16)
		errBadColorType,       // Invalid or not supported color format
		errBadPixelFormat,     // Invalid combination of bit depth and color type
		errBadCompressionMethod,
		errBadFilterMethod,    // Invalid or unsupported filtering method
		errBadFilterType,      // Bad filtering type
		errInflateErr,         // Decompression error (zlib)
		errBadInterlaceMethod, // Unsupported interlacing method (Adam7 only)
		errBadPicSize,         // Invalid image size
		errNoIENDFound,        // No IEND chunk found at the end of the stream
		errNotEnoughDataInGammaChunk,
		errNotEnoughDataInBkChunk,
		errIDATFoundBeforeacTL,
		errNotEnoughDataInAnimCtlChunk,
		errFrameCountIsZero,
		errNotEnoughDataInFrameCtlChunk,
		errBadFrameControlSequenceNumber,
		errBadFrameDataSequenceNumber,
		errIDATNotFound,
		errIDATfcTLNotConsistentWithIHDR,
		erracTLMissing,
		erracTLRepeated,
		errMissingfcTLBeforefdAT,
		errUnexpectedChunk,
		errFrameCountMismatch,
		errFrameCountOutsideRange,
		errImageDataMissing,
		errTextMissingSeparator,
		errTextBadKeywordLength
	};

	// Chunk names
	enum
	{
		chunk_IHDR = 'IHDR', // Header
		chunk_PLTE = 'PLTE', // Palette
		chunk_IDAT = 'IDAT', // Pixels
		chunk_IEND = 'IEND', // End of file
		chunk_gAMA = 'gAMA', // Gamma
		chunk_tRNS = 'tRNS', // Transparency information
		chunk_bKGD = 'bKGD', // Background color
		chunk_tEXt = 'tEXt', // Textual information
		chunk_acTL = 'acTL', // Animation control
		chunk_fcTL = 'fcTL', // Frame control
		chunk_fdAT = 'fdAT'  // Frame data
	};

public:
	/////////////////////////////////////////////////
	virtual bool LoadFromFile(IFile& file);
	virtual PixelFormat GetPixelFormat() const;
	virtual const Palette& GetPalette() const;
	virtual const Buffer& GetPixels() const;
	/////////////////////////////////////////////////

	virtual bool IsSimpleTransparencyUsed() const { return IsChunkFound(cf_tRNS); }
	virtual uint16 GetGreyTransIndex() const { return m_tRNS.grey; }
	virtual void GetTransIndexes(uint16& red, uint16& green, uint16& blue) const
	{ red = m_tRNS.red; green = m_tRNS.green; blue = m_tRNS.blue; }

	virtual String GetLastErrorString() const;

	/////////////////////////////////////////////////////////////////////////////////////
	static int Read_IHDR(IFile& file, int32 dataSizeof, PngChunk_IHDR& out);
	static PixelFormat GetPixelFormat(const PngChunk_IHDR& pngIHDR);
	static PixelFormat GetPixelFormat(uint8 colorType, uint8 bitDepth);

	/////////////////////////////////////////////////////////////////////////////////////
	static bool Dump(const String& strFileName, const PngDumpSettings& ds);
	static bool Dump(IFile& file, const PngDumpSettings& ds);
	static bool WriteChunk_bkGD(ChunkedFile& cf, uint8 colorType, const PngChunk_bkGD& content);
	static bool WriteChunk_tEXt(ChunkedFile& cf, const PngChunk_tEXt& content);

	/////////////////////////////////////////////////////////////////////////////////////
	bool   HasBackgroundColor() const { return IsChunkFound(cf_bKGD); }
	uint8  GetBkIndex() const { return m_bkGD.index; }
	uint16 GetBkGrey() const  { return m_bkGD.grey; }
	uint16 GetBkRed() const   { return m_bkGD.red; }
	uint16 GetBkGreen() const { return m_bkGD.green; }
	uint16 GetBkBlue() const  { return m_bkGD.blue; }

	bool   HasGamma() const { return IsChunkFound(cf_gAMA); }
	uint32 GetGamma() const { return m_nGamma; }

	bool   HasText() const { return m_tEXts.GetSize() > 0; }
	const Array<PngChunk_tEXt>& GetTexts() const { return m_tEXts; }

	virtual bool IsInterlaced() const { return m_IHDR.interlaceMethod != 0; }

	static bool IsPng(IFile& file);
	static bool WriteSignature(IFile& file);

	/////////////////////////////////////////////////////////////////////////////////////
	virtual bool  IsAnimated() const;
	virtual bool  HasDefaultImage() const;
	virtual int32 GetFrameCount() const;
	virtual const AnimFrame* GetAnimFrame(int index) const;
	virtual int32 GetLoopCount() const;
	/////////////////////////////////////////////////////////////////////////////////////

	Png();
	virtual ~Png();
	
private:
	// To manage expected chunks in the stream
	enum ChunkFlag
	{
		cf_IHDR = 0x0001,
		cf_PLTE = 0x0002,
		cf_IDAT = 0x0004,
		cf_IEND = 0x0008,
		cf_gAMA = 0x0010,
		cf_tRNS = 0x0020,
		cf_bKGD = 0x0040,
		cf_tEXt = 0x0080,
		cf_acTL = 0x0100,
		cf_fcTL = 0x0200,
		cf_fdAT = 0x0400,
		
		cf_ALL = 0xffffffff
	};
	uint32 m_nExpectedChunks;
	uint32 m_nLastUnexpectedChunk;

	uint32 m_foundChunks; // Bit flags of chunks found during decoding

	// colorType flags
	enum { cfGrey = 0, cfPalette = 1, cfColor = 2, cfAlpha = 4 };
	//////////////////////////////////////////////////////////
	bool   m_bChunkFound_IHDR;
	PngChunk_IHDR m_IHDR;
	
	Palette m_PLTE;  // Palette information

	int32 m_nSizeofPixel;       // Pixel size in bytes; if nbbpp < 8 => 1 byte
	int32 m_nSizeofPixelInBits; // Pixel size in bits
	
	/////////////////////////////////////////////////////////////////////////////////
	// Information needed to process image data
	struct ImageDataInfo
	{
		Buffer* pPixels;
		int32 width;
		int32 height;
		int32 byteWidth; // Size of one row not counting the filter byte
		int32 uncompressedDataSize;

		void Clear();
	};
	ImageDataInfo m_idiCurrent; // Current image data information: default image or current anim frame

	/////////////////////////////////////////////////////////////////////////////////
	PngChunk_tRNS m_tRNS;
	/////////////////////////////////////////////////////////////////////////////////

	/////////////////////////////////////////////////////////////////////////////////
	uint32 m_nGamma;       // Gamma * 100000
	/////////////////////////////////////////////////////////////////////////////////

	/////////////////////////////////////////////////////////////////////////////////
	PngChunk_bkGD m_bkGD;
	/////////////////////////////////////////////////////////////////////////////////

	/////////////////////////////////////////////////////////////////////////////////
	Array<PngChunk_tEXt> m_tEXts; // Non null size if found
	/////////////////////////////////////////////////////////////////////////////////

	int32     m_nCounter_IDAT;
	int32     m_nCounter_fdAT;

	ZStream   m_ZStream;
	uint32    m_nOutputOffset;
	ByteArray m_CompressedBuffer;

	struct AnimationControl
	{
		int32 frameCount;
		int32 loopCount;
	};
	AnimationControl m_acTL;

	PtrArray<ApngFrame> m_apFrames;
	int32 m_currentSequenceNumber;

private:
	void Initialize();

	static inline uint8 PaethPredictor(uint8 a, uint8 b, uint8 c);

	bool Handle_IDAT(IFile& file, int32 dataSizeof);
	bool Handle_PLTE(IFile& file, int32 dataSizeof);
	bool Handle_IHDR(IFile& file, int32 dataSizeof);
	bool Handle_gAMA(IFile& file, int32 dataSizeof);
	bool Handle_tRNS(IFile& file, int32 dataSizeof);
	bool Handle_bKGD(IFile& file, int32 dataSizeof);
	bool Handle_tEXt(IFile& file, int32 dataSizeof);
	bool Handle_acTL(IFile& file, int32 dataSizeof);
	bool Handle_fcTL(IFile& file, int32 dataSizeof);
	bool Handle_fdAT(IFile& file, int32 dataSizeof);

	bool AllocateImageBuffer(bool bInterlacedBuffer);
	bool BeginImageDataProcessing();
	bool ProcessImageData(IFile& file, int32 dataSizeof);
	bool EndImageDataProcessing();
	
	bool UnfilterAndUninterlace();
	
	static bool UnfilterBlock(uint8* pBlock, int32 rowCount, int32 pixelBytesPerRow, int32 bytesPerPixel);
	
	static void GetPassSize(int8 iPass, int32& rowCount, int32& pixelBytesPerRow,
		int32 width, int32 height, int32 sizeofPixelInBits);
	
	static int32 ComputeInterlacedSize(int32 width, int32 height, int32 sizeofPixelInBits);
	static bool GetIHDRFeaturesFromPixelFormat(PixelFormat epf, uint8& colorType, uint8& bitDepthPerComponent);
	static bool FilterBlock(uint8* const pBlock, int32 rowCount, int32 pixelBytesPerRow, int32 bytesPerPixel);

	static bool InterlaceAndFilter(uint8* pDst, const uint8* pSrc,
		const int32 srcWidth, const int32 srcHeight, const int32 srcPixelBytesPerRow,
		int32 sizeofPixelInBits, bool bDoFiltering);

	// ZLib compression tuned for PNG
	static int Compress(uint8* pDest, uint32* pDestLen, const uint8* pSource, uint32 sourceLen, 
		int level, int strategy, int windowBits, int memLevel);

	static bool CreateImageData(const uint8* pSrc, int32 width, int32 height, 
		const PngDumpSettings& ds, ByteArray& abImageData);

	static bool WriteFrameControlChunk(const ApngFrame* pFrame, int32 apngSequenceNumber, ChunkedFile& file);

	bool IsChunkFound(ChunkFlag cf) const { return (m_foundChunks & cf) ? true : false; }
	static bool IsValidBitDepth(uint8 bitDepth);
	static bool IsValidColorType(uint8 colorType);
};

///////////////////////////////////////////////////////////////////////////////
class ApngFrame : public AnimFrame
{
public:
	PngChunk_fcTL m_fctl;
	Buffer        m_pixels;

public:
	ApngFrame(Png* pPng) : AnimFrame(pPng) {}

	virtual int32 GetWidth() const { return m_fctl.width; }
	virtual int32 GetHeight() const { return m_fctl.height; }

	virtual int32 GetOffsetX() const { return m_fctl.offsetX; }
	virtual int32 GetOffsetY() const { return m_fctl.offsetY; }

	virtual int32 GetDelayFracNumerator() const { return m_fctl.delayFracNumerator; }
	virtual int32 GetDelayFracDenominator() const { return m_fctl.delayFracDenominator; }

	virtual Disposal GetDisposal() const { return Disposal(m_fctl.disposal); }
	virtual Blending GetBlending() const { return Blending(m_fctl.blending); }

	virtual const Buffer& GetPixels() const { return m_pixels; }
};

} // namespace chustd;

#endif // ndef CHUSTD_PNG_H
