///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2011 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Socket.h"
#include "String.h"
#include "IFile.h"
#include "StringBuilder.h"
#include "FormatType.h"

using namespace chustd;

//////////////////////////////////////////////////////////////////////
int Socket::ms_nSocketStartupCount = 0;
bool Socket::ms_bNetworkWorking = true;
//////////////////////////////////////////////////////////////////////

Socket::Socket()
{
#ifdef WIN32
	if( ms_nSocketStartupCount == 0 )
	{
		WSADATA wsaData;
		int ret = WSAStartup(MAKEWORD( 2, 2 ), &wsaData);
		ms_bNetworkWorking = ( ret == 0);
	}

	ms_nSocketStartupCount++;
#elif CHUOS
#endif

	m_nIDSocket = INVALID_SOCKET;
	m_nType = typeUnknown;
}

Socket::~Socket()
{
	Close();
	
	ms_nSocketStartupCount--;
	if( ms_nSocketStartupCount <= 0 )
	{
#ifdef WIN32
		WSACleanup();
#endif
	}
}

bool Socket::Open(Type eType)
{
	if( !ms_bNetworkWorking )
		return false;

	if( m_nType != typeUnknown )
	{
		// Socket already open, close it before we continue
		Close();
	}

#ifdef WIN32
	// Socket creation
	int nBkSocketType = 0;
	if( eType == typeTcp )
	{
		nBkSocketType = SOCK_STREAM;
	}
	else if( eType == typeUdp )
	{
		nBkSocketType = SOCK_DGRAM;
	}
	else
	{
		return false;
	}

	const SOCKET_ID nIDSocket = socket(AF_INET, nBkSocketType, 0);
	if( nIDSocket == INVALID_SOCKET )
		return false;

	m_nIDSocket = int64(nIDSocket);
	m_nType = uint8(eType);
	return true;

#elif CHUOS
	return false;
#endif
	
}

bool Socket::IsOpen() const
{
	return m_nType != typeUnknown;
}

void Socket::Close()
{
	if( m_nIDSocket != INVALID_SOCKET )
	{
#ifdef WIN32
		closesocket( SOCKET_ID(m_nIDSocket)); // Windows
#elif CHUOS

#else
		close(m_nIDSocket); // Other
#endif
	}

	m_nIDSocket = INVALID_SOCKET;
	m_nType = typeUnknown;
}


int32 Socket::Receive(uint8* pBuffer, int32 bufferSize, bool bBlockUntilSizeReached)
{
	if( !IsOpen() )
		return 0;

#ifdef WIN32
	if( !bBlockUntilSizeReached )
	{
		return recv( SOCKET_ID(m_nIDSocket), (char*)pBuffer, bufferSize, 0);
	}

	int32 received = 0; // Number of bytes received
	do
	{
		const int ret = recv(SOCKET_ID(m_nIDSocket), (char*)pBuffer + received, bufferSize - received, 0);
		if( ret <= 0 )
		{
			break;
		}
		received += ret;
	}while(received < bufferSize);
	return received;

#elif CHUOS
	return 0;
#endif

}

int32 Socket::Send(const uint8* pBuffer, int32 bufferSize, bool bBlockUntilSizeReached)
{
	if( !IsOpen() )
		return 0;

#ifdef WIN32
	if( !bBlockUntilSizeReached )
	{
		return send(SOCKET_ID(m_nIDSocket), (const char*)pBuffer, bufferSize, 0);
	}

	int32 sent = 0; // Number of bytes sent
	do
	{
		int ret = send(SOCKET_ID(m_nIDSocket), (const char*)(pBuffer + sent), bufferSize - sent, 0);
		if( ret <= 0 )
		{
			if( bBlockUntilSizeReached && IsSendBufferFlooded() )
			{
				// We should retry sending
				bool bOk = BlockUntilSendPossible();
				if( !bOk )
				{
					// An error occured while waiting for sending authorization
					return 0;
				}
			}
			else
			{
				// Real error
				return 0;
			}
		}
		else
		{
			sent += ret;
		}
	}while( sent < bufferSize);
	
	return sent;
#elif CHUOS
	return 0;
#endif
}

// Returns true if a send operation could not be completed because of a too small TCP/IP send buffer
bool Socket::IsSendBufferFlooded() const
{
#ifdef WIN32
	int ret = WSAGetLastError();

	if( ret == WSAEWOULDBLOCK )
	{
		// Yep, send buffer is too small, we should retry sending
		// Note : this error can occurs on some Windows configuration even if the socket
		// has not been set for asynchronous write operations (WSAAsyncSelect)
		return true;
	}
#elif CHUOS
	return false;

#else
	if( errno == EWOULDBLOCK )
	{
		// Non blocking socket
		return true;
	}
#endif

	// Other error, does not concern send buffer flooding
	return false;
}

bool Socket::BlockUntilSendPossible()
{
	ASSERT(m_nIDSocket != INVALID_SOCKET);

#ifdef WIN32
	fd_set fdsWrite;

	FD_ZERO(&fdsWrite); // Clear the fd_set
	
	// int on Unix, uint on Windows, so warning here
	SOCKET_ID socketId = SOCKET_ID(m_nIDSocket);
	FD_SET(socketId, &fdsWrite); // Add the socket ID to the socket set
 
	int ret = select(0, NULL, &fdsWrite, NULL, NULL);
	if( ret <= 0 )
		return false;
#elif CHUOS
#endif
	return true;
}

int64 Socket::GetSocketID() const
{
	return m_nIDSocket;
}

bool Socket::StringToIpAddress(const String& strIpAddress, uint32& nIpAddress)
{
	const int32 stringLength = strIpAddress.GetLength();
	const wchar* psz = strIpAddress.GetBuffer();

	/////////////////////////////////////////////////////////////////////
	// First pass, count dot and store dots positions
	int32 aDotIndex[3];
	int32 dotCount = 0;

	for(int i = 0; i < stringLength; i++)
	{
		wchar c = psz[i];
		if( c == '.' )
		{
			if( dotCount == 3 )
			{
				// Too many dots
				return false;
			}
			aDotIndex[dotCount] = i;
			dotCount++;
		}
	}

	if( dotCount != 3 )
	{
		// Not enough dots
		return false;
	}
	/////////////////////////////////////////////////////////////////////
	const int32 length0 = aDotIndex[0];
	if( length0 > 3 || length0 <= 0 )
		return false;

	const int32 length1 = (aDotIndex[1] - aDotIndex[0]) - 1;
	if( length1 > 3 || length1 <= 0 )
		return false;

	const int32 length2 = (aDotIndex[2] - aDotIndex[1]) - 1;
	if( length2 > 3 || length2 <= 0 )
		return false;
	
	const int32 length3 = (stringLength - aDotIndex[2]) - 1;
	if( length3 > 3 || length3 <= 0 )
		return false;

	int32 aPieceLengths[4] = { length0, length1, length2, length3 };

	uint32 result = 0; // IP Address, to be copied into nIpAddress later

	for(int iPiece = 0; iPiece < 4; iPiece++)
	{
		const int32 pieceLength = aPieceLengths[iPiece];

		uint32 val = 0;
		for(int iDigit = 0; iDigit < pieceLength; iDigit++)
		{
			const uint16 c = psz[iDigit];
			if( '0' <= c && c <= '9' )
			{
				const int32 digit = c - '0';
				
				val *= 10;
				val += digit;
			}
			else
			{
				// The character is not a digit
				return false;
			}
		}
		
		if( val > 255 )
		{
			// Each piece is 8 bits so > 255 is not acceptable
			return false;
		}

		// Copy Piece value
		result <<= 8;
		result |= val;

		// Jump over last piece + dot
		psz += pieceLength + 1;
	}
	
	nIpAddress = result;
	return true;
}

String Socket::IpAddressToString(uint32 nIpAddress)
{
	uint8 nA = uint8( (nIpAddress >> 24) & 0x00ff);
	uint8 nB = uint8( (nIpAddress >> 16) & 0x00ff);
	uint8 nC = uint8( (nIpAddress >> 8) & 0x00ff);
	uint8 nD = uint8( (nIpAddress) & 0x00ff);

	StringBuilder sb;

	wchar szBuffer[12];

	FormatInt32(nA, szBuffer);
	sb += szBuffer;
	sb += '.';

	FormatInt32(nB, szBuffer);
	sb += szBuffer;
	sb += '.';

	FormatInt32(nC, szBuffer);
	sb += szBuffer;
	sb += '.';

	FormatInt32(nD, szBuffer);
	sb += szBuffer;

	return sb.ToString();
}
