///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2011 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#ifndef CHUSTD_ZSTREAM_H
#define CHUSTD_ZSTREAM_H

namespace chustd
{

class ZStream
{
public:
	// Flush lvalues
	enum
	{
		Z_NO_FLUSH      = 0,
		Z_PARTIAL_FLUSH = 1, // will be removed, use Z_SYNC_FLUSH instead
		Z_SYNC_FLUSH    = 2,
		Z_FULL_FLUSH    = 3,
		Z_FINISH        = 4,
		Z_BLOCK         = 5
	};


	// Return codes for the compression/decompression functions. Negative
	// values are errors, positive values are used for special but normal events.
 	enum
	{
		Z_OK            = 0,
		Z_STREAM_END    = 1,
		Z_NEED_DICT     = 2,
		Z_ERRNO         = (-1),
		Z_STREAM_ERROR  = (-2),
		Z_DATA_ERROR    = (-3),
		Z_MEM_ERROR     = (-4),
		Z_BUF_ERROR     = (-5),
		Z_VERSION_ERROR = (-6)
	};

	// Compression method
	enum
	{
		Z_DEFLATED = 8
	};

	// Compression strategy
	enum
	{
		Z_FILTERED           = 1,
		Z_HUFFMAN_ONLY       = 2,
		Z_RLE                = 3,
		Z_FIXED              = 4,
		Z_DEFAULT_STRATEGY   = 0
	};

public:
	ZStream();
	~ZStream();

	void SetBuffers(const uint8* pInNext, uint32 nInAvailable, uint8* pOutNext, uint32 outAvailable);
	uint32 GetOutAvailable() const;
	uint32 GetOutTotalRead() const;

	///////////////////////
	// Compression
	int DeflateInit(int level); // Default compression is for level = 6
	int DeflateInit2(int level, uint8 method, int windowBits, int memLevel, int strategy);

	int Deflate(int flush);
	int DeflateParams(int newLevel, int strategy);
	int DeflateEnd();

	int DeflateSetDictionary(const uint8* pDictionary, uint32 dictLength);
	int DeflateReset();
	
	static uint32 DeflateBound(uint32 sourceLength);

	///////////////////////
	// Uncompression
	int InflateInit();
	int InflateInit2(int windowBits);

	int Inflate(int flushType);
	int InflateEnd();

	int InflateSetDictionary(const uint8* pDictionary, uint32 dictLength);
	int InflateSync();
	int InflateReset();
	int InflateSyncPoint();

	// Compresses the source buffer into the destination buffer. The level
	// parameter has the same meaning as in deflateInit.  sourceLen is the byte
	// length of the source buffer. Upon entry, destLen is the total size of the
	// destination buffer, which must be at least 0.1% larger than sourceLen plus
	// 12 bytes. Upon exit, destLen is the actual size of the compressed buffer.

	// compress2 returns Z_OK if success, Z_MEM_ERROR if there was not enough
	// memory, Z_BUF_ERROR if there was not enough room in the output buffer,
	// Z_STREAM_ERROR if the level parameter is invalid.
	static int Compress(uint8* pDest, uint32* pDestLen, const uint8* pSource, uint32 sourceLen, int level);

	void DebugDump();
	
	const char* GetLastError() const;

private:
	class ZStreamImpl* m_pImpl;
};

} // namespace chustd

#endif // ndef CHUSTD_ZSTREAM_H
