
#include "stdafx.h"

class TestImg;

class TestFrame : public ApngFrame
{
public:
	bool m_ownPalette;
	Palette m_palette;

	TestFrame(TestImg* pOwner) : ApngFrame((Png*)pOwner)
	{
		m_ownPalette = false;
	}

	virtual const Palette& GetPalette() const
	{
		if( m_ownPalette )
		{
			return m_palette;
		}
		return ApngFrame::GetPalette();
	}

};

class TestImg : public ImageFormat
{
public:
	PixelFormat m_pf;
	Palette m_palette;
	Array<TestFrame> m_animFrames;
	bool m_hasDefaultImage;
	int32 m_loopCount;

public:
	TestImg()
	{
		m_hasDefaultImage = false;
		m_loopCount = 0;
	}
	void SetWidth(int width) { m_nWidth = width; }
	void SetHeight(int height) { m_nHeight = height; }
	void SetPixels(const uint8* pPixels, int size) { m_pixels.Assign(pPixels, size); }

	virtual PixelFormat GetPixelFormat() const { return m_pf; }
	const Palette& GetPalette() const { return m_palette; }
	bool LoadFromFile(IFile& file) { return false; }

	virtual bool  IsAnimated() const { return m_animFrames.GetSize() > 0; }
	virtual bool  HasDefaultImage() const { return m_hasDefaultImage; }
	virtual int32 GetFrameCount() const { return m_animFrames.GetSize(); }
	virtual const AnimFrame* GetAnimFrame(int32 nIndex) const { return &m_animFrames[nIndex]; }
	virtual int32 GetLoopCount() const { return m_loopCount; } // 0 = infinite
};

// Test the case with a GIF which frames are smaller than the drawing area
TEST(POEngine, OptimizeAnimated_BigTransparentFrame)
{
	TestImg img;
	img.m_pf = pf8bppIndexed;
	img.SetWidth(7); // 2 + 3 + 2
	img.SetHeight(7);

	// The palette does provide a transparent color, this one should be used
	img.m_palette.m_count = 2;
	img.m_palette[0].SetRgba(255, 255, 255, 0);
	img.m_palette[1].SetRgba(100, 200, 100, 255);

	TestFrame frame0(&img);
	frame0.m_fctl.offsetX = 2;
	frame0.m_fctl.offsetY = 2;
	frame0.m_fctl.width = 3;
	frame0.m_fctl.height = 3;
	frame0.m_fctl.delayFracNumerator = 100;
	frame0.m_pixels.SetSize(3 * 3);
	frame0.m_pixels.Fill(1); // Fully opaque
	img.m_animFrames.Add(frame0);

	POEngine engine;
	PngDumpSettings ds;
	static const String resultFilePath = "result.png";
	Png png;

	File::Delete(resultFilePath);
	ASSERT_TRUE( engine.OptimizeAnimated(img, ds, resultFilePath) );

	// We should get a APNG image with the same palette
	// A default image of 7x7, containing the first 3x3 frame, with transparent color (index 0) around
	ASSERT_TRUE( png.Load(resultFilePath) );

	// One single frame : no animation information at all
	ASSERT_TRUE( !png.IsAnimated() );
	ASSERT_TRUE( png.HasDefaultImage() );
	ASSERT_EQ( 7, png.GetWidth() );
	ASSERT_EQ( 7, png.GetHeight() );
	ASSERT_TRUE( png.GetPixelFormat() == pf1bppIndexed );
	ASSERT_TRUE( png.GetPalette().m_count == 2 );

	static const uint8 frame0Bytes[] = {
		BYTE_FROM_BITS(0, 0, 0, 0, 0, 0, 0,  0),
		BYTE_FROM_BITS(0, 0, 0, 0, 0, 0, 0,  0),
		BYTE_FROM_BITS(0, 0, 1, 1, 1, 0, 0,  0),
		BYTE_FROM_BITS(0, 0, 1, 1, 1, 0, 0,  0),
		BYTE_FROM_BITS(0, 0, 1, 1, 1, 0, 0,  0),
		BYTE_FROM_BITS(0, 0, 0, 0, 0, 0, 0,  0),
		BYTE_FROM_BITS(0, 0, 0, 0, 0, 0, 0,  0)
	};
	const uint8* pLoaded = png.GetPixels().GetReadPtr();
	ASSERT_TRUE( Memory::Equals(pLoaded, frame0Bytes, sizeof(frame0Bytes)) );

	/////////////////////////////////////////////////////////
	// Now add a 2nd and frame and redump
	img.m_animFrames[0].m_fctl.disposal = AnimFrame::dispClearToTransBlack;

	TestFrame frame1(&img);
	frame1.m_fctl.offsetX = 3;
	frame1.m_fctl.offsetY = 3;
	frame1.m_fctl.width = 1;
	frame1.m_fctl.height = 1;
	frame1.m_fctl.delayFracNumerator = 100;
	frame1.m_pixels.SetSize(1 * 1);
	frame1.m_pixels.Fill(1); // Fully opaque
	img.m_animFrames.Add(frame1);

	File::Delete(resultFilePath);
	ASSERT_TRUE( engine.OptimizeAnimated(img, ds, resultFilePath) );

	// We should get a APNG image with the same palette
	// A default image of 7x7, containing the first 3x3 frame, with transparent color (index 0) around
	ASSERT_TRUE( png.Load(resultFilePath) );

	// One single frame : no animation information at all
	ASSERT_TRUE( png.IsAnimated() );
	ASSERT_TRUE( !png.HasDefaultImage() );
	ASSERT_TRUE( png.GetWidth() == 7 );
	ASSERT_TRUE( png.GetHeight() == 7 );
	ASSERT_TRUE( png.GetFrameCount() == 2 );
	ASSERT_TRUE( png.GetPixelFormat() == pf1bppIndexed );

	static const uint8 frame1Bytes[] = {
		BYTE_FROM_BITS(1, 0, 0, 0, 0, 0, 0,  0)
	};
	const uint8* pFramePixels = png.GetAnimFrame(0)->GetPixels().GetReadPtr();
	ASSERT_TRUE( Memory::Equals(pFramePixels, frame0Bytes, sizeof(frame0Bytes)) );
	pFramePixels = png.GetAnimFrame(1)->GetPixels().GetReadPtr();
	ASSERT_TRUE( Memory::Equals(pFramePixels, frame1Bytes, sizeof(frame1Bytes)) );

}

///////////////////////////////////////////////////////////////////////////////////////////////////
TEST(POEngine, OptimizeAnimated_BigTransparentFrame_MultiPalette)
{
	TestImg img;
	img.m_pf = pf8bppIndexed;
	img.SetWidth(7); // 2 + 3 + 2
	img.SetHeight(7);

	TestFrame frame0(&img);
	frame0.m_fctl.offsetX = 2;
	frame0.m_fctl.offsetY = 2;
	frame0.m_fctl.width = 3;
	frame0.m_fctl.height = 3;
	frame0.m_fctl.delayFracNumerator = 100;
	frame0.m_pixels.SetSize(3 * 3);
	frame0.m_pixels.Fill(1); // Fully opaque
	frame0.m_ownPalette = true;
	frame0.m_palette.m_count = 2;
	frame0.m_palette[0].SetRgba(255, 255, 255, 0);
	frame0.m_palette[1].SetRgba(100, 200, 100, 255);
	img.m_animFrames.Add(frame0);

	img.m_animFrames[0].m_fctl.disposal = AnimFrame::dispClearToTransBlack;

	TestFrame frame1(&img);
	frame1.m_fctl.offsetX = 3;
	frame1.m_fctl.offsetY = 3;
	frame1.m_fctl.width = 1;
	frame1.m_fctl.height = 1;
	frame1.m_fctl.delayFracNumerator = 100;
	frame1.m_pixels.SetSize(1 * 1);
	frame1.m_pixels.Fill(1); // Fully opaque
	frame1.m_ownPalette = true;
	frame1.m_palette.m_count = 2;
	frame1.m_palette[0].SetRgba(255, 255, 255, 0);
	frame1.m_palette[1].SetRgba(200, 100, 100, 255);
	img.m_animFrames.Add(frame1);

	//////////////////////////////////////////////////////////////////
	// Dump
	POEngine engine;
	PngDumpSettings ds;
	static const String resultFilePath = "result.png";
	Png png;

	File::Delete(resultFilePath);
	ASSERT_TRUE( engine.OptimizeAnimated(img, ds, resultFilePath) );

	// We should get a APNG image with the same palette
	// A default image of 7x7, containing the first 3x3 frame, with transparent color (index 0) around
	ASSERT_TRUE( png.Load(resultFilePath) );

	// One single frame : no animation information at all
	ASSERT_TRUE( png.IsAnimated() );
	ASSERT_TRUE( !png.HasDefaultImage() );
	ASSERT_TRUE( png.GetWidth() == 7 );
	ASSERT_TRUE( png.GetHeight() == 7 );
	ASSERT_TRUE( png.GetFrameCount() == 2 );
	ASSERT_TRUE( png.GetPixelFormat() == pf2bppIndexed );
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// After the merge of all frame palettes, the new palette exceeds 256 colors. true colors conversion needed
void Helper_OptimizeAnimated_MultiPalette_Saturated(int w, int h, PixelFormat expectedPixelFormat)
{
	TestImg img;
	img.m_pf = pf8bppIndexed;
	img.SetWidth(w);
	img.SetHeight(h);

	TestFrame frame0(&img);
	frame0.m_fctl.offsetX = 0;
	frame0.m_fctl.offsetY = 0;
	frame0.m_fctl.width = 16;
	frame0.m_fctl.height = 16;
	frame0.m_fctl.delayFracNumerator = 100;
	frame0.m_pixels.SetSize(16 * 16);
	frame0.m_ownPalette = true;
	frame0.m_palette.m_count = 256;
	uint8* pRaw = frame0.m_pixels.GetWritePtr();
	for(int i = 0; i < 256; ++i)
	{
		pRaw[i] = i;
		frame0.m_palette[i].SetRgba(i, 255, 255, 255);
	}
	img.m_animFrames.Add(frame0);

	img.m_animFrames[0].m_fctl.disposal = AnimFrame::dispClearToTransBlack;

	TestFrame frame1(&img);
	frame1.m_fctl.offsetX = 0;
	frame1.m_fctl.offsetY = 0;
	frame1.m_fctl.width = 16;
	frame1.m_fctl.height = 16;
	frame1.m_fctl.delayFracNumerator = 100;
	frame1.m_pixels.SetSize(16 * 16);
	frame1.m_ownPalette = true;
	frame1.m_palette.m_count = 256;
	pRaw = frame1.m_pixels.GetWritePtr();
	for(int i = 0; i < 256; ++i)
	{
		pRaw[i] = i;
		frame1.m_palette[i].SetRgba(255, i, 255, 255);
	}
	img.m_animFrames.Add(frame1);

	//////////////////////////////////////////////////////////////////
	// Dump
	POEngine engine;
	PngDumpSettings ds;
	static const String resultFilePath = "result.png";
	Png png;

	File::Delete(resultFilePath);
	ASSERT_TRUE( engine.OptimizeAnimated(img, ds, resultFilePath) );

	// We should get a APNG image with the same palette
	// A default image of 7x7, containing the first 3x3 frame, with transparent color (index 0) around
	ASSERT_TRUE( png.Load(resultFilePath) );

	// One single frame : no animation information at all
	ASSERT_TRUE( png.IsAnimated() );
	ASSERT_TRUE( !png.HasDefaultImage() );
	ASSERT_TRUE( png.GetWidth() == w );
	ASSERT_TRUE( png.GetHeight() == h );
	ASSERT_TRUE( png.GetFrameCount() == 2 );
	ASSERT_TRUE( png.GetPixelFormat() == expectedPixelFormat );
}

TEST(POEngine, OptimizeAnimated_MultiPalette_Saturated)
{
	Helper_OptimizeAnimated_MultiPalette_Saturated(16, 16, pf24bppRgb);
}

TEST(POEngine, OptimizeAnimated_BigTransparentFrame_MultiPalette_Saturated)
{
	Helper_OptimizeAnimated_MultiPalette_Saturated(20, 20, pf32bppRgba);
}

// Create a default image and two frames, all of them use 3 colors out of 4
// We should obtain a PNG with a palette of 3 colors only
TEST(POEngine, OptimizeAnimated_UnusedColors)
{
	TestImg img;
	img.m_pf = pf8bppIndexed;
	img.SetWidth(7); // 2 + 3 + 2
	img.SetHeight(7);

	// Do not use frame palettes, as local palettes (GIF) is mutually exclusive with a default image (PNG)

	TestFrame frame0(&img);
	frame0.m_fctl.offsetX = 2;
	frame0.m_fctl.offsetY = 2;
	frame0.m_fctl.width = 3;
	frame0.m_fctl.height = 3;
	frame0.m_fctl.delayFracNumerator = 100;
	frame0.m_pixels.SetSize(3 * 3);
	frame0.m_pixels.Fill(1); // Fully opaque
	frame0.m_ownPalette = false;
	img.m_animFrames.Add(frame0);

	img.m_animFrames[0].m_fctl.disposal = AnimFrame::dispClearToTransBlack;

	TestFrame frame1(&img);
	frame1.m_fctl.offsetX = 3;
	frame1.m_fctl.offsetY = 3;
	frame1.m_fctl.width = 1;
	frame1.m_fctl.height = 1;
	frame1.m_fctl.delayFracNumerator = 100;
	frame1.m_pixels.SetSize(1 * 1);
	frame1.m_pixels.Fill(3); // Fully opaque
	frame1.m_ownPalette = false;
	img.m_animFrames.Add(frame1);

	//////////////////////////////////////////////////////////////////
	// Create a default image and fill the common palette
	static const uint8 defaultBytes[] = {
		1, 1, 1, 1, 1, 1, 1,
		1, 0, 0, 0, 0, 0, 1,
		1, 0, 3, 3, 3, 0, 1,
		1, 0, 3, 3, 3, 0, 1,
		1, 0, 3, 3, 3, 0, 1,
		1, 0, 0, 0, 0, 0, 1,
		1, 1, 1, 1, 1, 1, 1,
	};
	img.m_hasDefaultImage = true;
	img.SetPixels(defaultBytes, sizeof(defaultBytes));
	img.m_palette.m_count = 4;
	img.m_palette[0].SetRgba(255, 255, 255, 0);
	img.m_palette[1].SetRgba(200, 100, 50, 255);
	img.m_palette[2].SetRgba(88, 88, 88, 255); // Unused color
	img.m_palette[3].SetRgba(200, 100, 100, 255);

	//////////////////////////////////////////////////////////////////
	// Dump
	POEngine engine;
	PngDumpSettings ds;
	static const String resultFilePath = "result.png";
	Png png;

	File::Delete(resultFilePath);
	ASSERT_TRUE( engine.OptimizeAnimated(img, ds, resultFilePath) );
	ASSERT_TRUE( png.Load(resultFilePath) );
	ASSERT_TRUE( png.IsAnimated() );
	ASSERT_TRUE( png.HasDefaultImage() );
	ASSERT_TRUE( png.GetWidth() == 7 );
	ASSERT_TRUE( png.GetHeight() == 7 );
	ASSERT_TRUE( png.GetFrameCount() == 2 );
	ASSERT_TRUE( png.GetPixelFormat() == pf2bppIndexed );
	ASSERT_EQ( png.GetPalette().m_count, 3 );
}

// Create a default image and two frames, all of them use 3 colors out of 4
// We should obtain a PNG with a palette of 3 colors only
TEST(POEngine, OptimizeAnimated_Loops)
{
	TestImg img;
	img.m_pf = pf8bppIndexed;
	img.SetWidth(4);
	img.SetHeight(4);

	TestFrame frame0(&img);
	frame0.m_fctl.width = 4;
	frame0.m_fctl.height = 4;
	frame0.m_fctl.delayFracNumerator = 200;
	frame0.m_pixels.SetSize(16);
	frame0.m_pixels.Fill(0);
	img.m_animFrames.Add(frame0);

	img.m_animFrames[0].m_fctl.disposal = AnimFrame::dispClearToTransBlack;

	TestFrame frame1(&img);
	frame1.m_fctl.width = 4;
	frame1.m_fctl.height = 4;
	frame1.m_fctl.delayFracNumerator = 200;
	frame1.m_pixels.SetSize(16);
	frame1.m_pixels.Fill(1);
	img.m_animFrames.Add(frame1);

	img.m_palette.m_count = 2;
	img.m_palette[0].SetRgb(255, 0, 0);
	img.m_palette[1].SetRgb(0, 255, 0);
	
	//////////////////////////////////////////////////////////////////
	// Dump
	POEngine engine;
	PngDumpSettings ds;
	static const String resultFilePath = "result.png";
	Png png;

	File::Delete(resultFilePath);
	ASSERT_TRUE( engine.OptimizeAnimated(img, ds, resultFilePath) );
	ASSERT_TRUE( png.Load(resultFilePath) );
	ASSERT_TRUE( png.IsAnimated() );
	ASSERT_TRUE( png.GetLoopCount() == 0 ); // 0 = infinite

	img.m_loopCount = 1;

	File::Delete(resultFilePath);
	ASSERT_TRUE( engine.OptimizeAnimated(img, ds, resultFilePath) );
	ASSERT_TRUE( png.Load(resultFilePath) );
	ASSERT_TRUE( png.IsAnimated() );
	ASSERT_TRUE( png.GetLoopCount() == 1 );

	img.m_loopCount = 347;

	File::Delete(resultFilePath);
	ASSERT_TRUE( engine.OptimizeAnimated(img, ds, resultFilePath) );
	ASSERT_TRUE( png.Load(resultFilePath) );
	ASSERT_TRUE( png.IsAnimated() );
	ASSERT_TRUE( png.GetLoopCount() == 347 );
}

int main(int argc, char** argv)
{
	::testing::InitGoogleTest(&argc, argv);
	return RUN_ALL_TESTS();
}
